﻿using System;
using System.Drawing;
using System.ComponentModel;
using System.Windows.Forms;

namespace WobbrockLib.Controls
{
	/// <summary>
	/// Provides a common base class for our custom numeric up down controls, encapsulating common functionality
	/// and exposing common events.
	/// </summary>
	public abstract partial class AbsNumericUpDown : NumericUpDown
	{
		#region Events

		/// <summary>
		/// Event that callers can subscribe to when the up or down buttons on this numeric up down
		/// control are clicked.
		/// </summary>
		[Browsable(true)]
		[Category("Action")]
		[Description("Occurs when the up or down button is clicked on this control.")]
		public event EventHandler<UpDownClickEventArgs> UpDownButtonClick;

		/// <summary>
		/// Expose the MouseEnter event to clients for this control.
		/// </summary>
		[Browsable(true)]
		[Category("Mouse")]
		[Description("Occurs when the mouse enters the visible part of the control.")]
		public new event EventHandler MouseEnter;

		/// <summary>
		/// Expose the MouseHover event to clients for this control.
		/// </summary>
		[Browsable(true)]
		[Category("Mouse")]
		[Description("Occurs when the mouse remains stationary inside of the control for an amount of time.")]
		public new event EventHandler MouseHover;

		/// <summary>
		/// Expose the MouseLeave event to clients for this control.
		/// </summary>
		[Browsable(true)]
		[Category("Mouse")]
		[Description("Occurs when the mouse leaves the visible part of the control.")]
		public new event EventHandler MouseLeave;

		/// <summary>
		/// Expose the MouseMove event to clients for this control.
		/// </summary>
		[Browsable(true)]
		[Category("Mouse")]
		[Description("Occurs when the mouse pointer is moved over the component.")]
		public new event EventHandler MouseMove;

		#endregion

		#region Constructor 

		/// <summary>
		/// Default constructor. 
		/// </summary>
		protected AbsNumericUpDown()
			: this(null)
		{
			// do nothing
		}

		/// <summary>
		/// Constructor with container.
		/// </summary>
		/// <param name="container">The container for this control.</param>
		protected AbsNumericUpDown(IContainer container)
		{
			container?.Add(this);
			InitializeComponent();
			tmrHover.Enabled = false;
			tmrHover.Interval = SystemInformation.MouseHoverTime;

			// register for enter, hover, leave, and move events for the 
			// textbox portion of the control. unfortunately, these events
			// do not fire reliably for Controls[0], the updown button
			// portion of the control, so we don't expose them. also, we  
			// could subscribe to hover but prefer to allow a custom hover
			// time property.
			TextBox tb = (TextBox) Controls[1];
			tb.MouseEnter += TextBox_MouseEnter;
			tb.MouseLeave += TextBox_MouseLeave;
			tb.MouseMove += TextBox_MouseMove;
		}

		#endregion

		#region Numeric Value

		/// <summary>
		/// Gets the text currently displayed in this widget. This text includes formatting,
		/// such as a comma or decimal place and all displayed significant digits.
		/// </summary>
		[Browsable(true)]
		[Category("Appearance")]
		[Description("The text visible in the textbox portion of the control.")]
		[RefreshProperties(RefreshProperties.Repaint)]
		public override string Text
		{
			get
			{
				// the second embedded control is the text box
				return this.Controls[1].Text;
			}
		}

		/// <summary>
		/// For some reason, the text box component of this control does not update its Text property
		/// even when this OnValueChanged method is called. So clients handling the ValueChanged event who 
		/// access the Value property and the Text property will see them differ. Thus, before calling the 
		/// base method, force the Text property to be updated. Then clients will see Value and Text correctly 
		/// correspond.
		/// </summary>
		/// <param name="e">The arguments for this event.</param>
		protected override void OnValueChanged(EventArgs e)
		{
			this.Controls[1].Text = this.Value.ToString();
			((TextBox) this.Controls[1]).Select(this.Controls[1].Text.Length, 0); // place text cursor at value's end
			base.OnValueChanged(e);
		}

		#endregion

		#region UpDown Button Clicks

		/// <summary>
		/// Override for when the Up button has been clicked on this numeric up down control.
		/// </summary>
		public override void UpButton()
		{
			decimal pre = this.Value;
			base.UpButton();
			decimal post = this.Value;
			UpDownClickEventArgs e = new UpDownClickEventArgs(UpDownButton.Up, pre, post);
			UpDownButtonClick?.Invoke(this, e);
		}

		/// <summary>
		/// Override for when the Down button has been clicked on this numeric up down control.
		/// </summary>
		public override void DownButton()
		{
			decimal pre = this.Value;
			base.DownButton();
			decimal post = this.Value;
			UpDownClickEventArgs e = new UpDownClickEventArgs(UpDownButton.Down, pre, post);
			UpDownButtonClick?.Invoke(this, e);
		}

		#endregion

		#region TextBox Mouse

		/// <summary>
		/// Gets or sets the mouse hover time interval.
		/// </summary>
		[Browsable(true)]
		[Category("Mouse")]
		[Description("The mouse hover time, in milliseconds, for the MouseHover event.")]
		[DefaultValue(400)]
		[RefreshProperties(RefreshProperties.None)]
		public int MouseHoverTime
		{
			get
			{
				return tmrHover.Interval;
			}
			set
			{
				if (value > 0)
				{
					tmrHover.Interval = value;
				}
			}
		}

		/// <summary>
		/// Handler for when the mouse enters the textbox portion of the control.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void TextBox_MouseEnter(object sender, EventArgs e)
		{
			tmrHover.Start();
			MouseEnter?.Invoke(this, e);
		}

		/// <summary>
		/// Handler for when the hover timer fires.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void tmrHover_Tick(object sender, EventArgs e)
		{
			tmrHover.Stop();
			MouseHover?.Invoke(this, e);
		}

		/// <summary>
		/// Handler for when the mouse leaves the textbox portion of the control.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void TextBox_MouseLeave(object sender, EventArgs e)
		{
			tmrHover.Stop();
			MouseLeave?.Invoke(this, e);
		}

		/// <summary>
		/// Handler for when the mouse moves over the textbox portion of the control.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void TextBox_MouseMove(object sender, MouseEventArgs e)
		{
			tmrHover.Stop();
			tmrHover.Start();
			MouseMove?.Invoke(this, e);
		}

		#endregion
	}
}
