﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Drawing;
using System.Windows.Forms;

namespace WobbrockLib.Controls
{
	/// <summary>
	/// This numeric updown control accepts the pressing of Enter and Esc keys without playing an audible 
	/// ding, which happens when typing these keys into a regular NumericUpDown widget unless the Form 
	/// that it is on has an AcceptButton or CancelButton set, respectively, in which case, that button
	/// is triggered, perhaps unwantedly. To avoid the ding, hacks like adding a disabled invisible
	/// no-op button to a Form were sometimes used. This widget prevents the need for such antics. It
	/// also ensures that a new value can be committed by typing Enter even when an AcceptButton is 
	/// specified for the Form, and that an old value can be restored by typing Esc, even when a CancelButton
	/// is specified for the Form, all without playing a ding.
	/// </summary>
	[ToolboxBitmap(typeof(System.Windows.Forms.NumericUpDown))]
	public partial class EnterEscNumericUpDown : AbsNumericUpDown
	{
		private decimal _savedValue; // for restoring when Esc is pressed

		/// <summary>
		/// Default constructor.
		/// </summary>
		public EnterEscNumericUpDown()
			: this(null)
		{
			// do nothing
		}

		/// <summary>
		/// Constructs a new instance of a EnterEscNumericUpDown to be put in the given container.
		/// </summary>
		/// <param name="container">The container in which to put this widget.</param>
		public EnterEscNumericUpDown(IContainer container)
		{
			container?.Add(this);
			InitializeComponent();
			_savedValue = this.Value;
		}

		/// <summary>
		/// Processes any command keys typed into this numeric text box. These keys include Enter and Escape.
		/// When Enter is pressed, cause the current text to be validated and "committed," which the default
		/// NumericUpDown widget does not do. When Escape is pressed, cause the current text to revert back
		/// to the previously "committed" value, whatever that is. In both cases, set the text cursor to the end
		/// of the text and indicate we've handled the keys, preventing any dings or need for Form AcceptButton
		/// or CancelButton fields to be set to avoid them.
		/// </summary>
		/// <param name="msg">The Windows message coming into this function from this widget's message queue.</param>
		/// <param name="keyData">The data indicating which keys were pressed.</param>
		/// <returns>True if we have handled this message; false otherwise.</returns>
		/// <remarks>One downside of this widget is that Enter will no longer work to trigger a dialog's OK button,
		/// and Escape will no longer work to trigger a dialog's Cancel button. That's because we're returning
		/// true when Enter or Escape have been pressed, which avoids the "ding," but also avoids triggering
		/// a dialog's Accept or Cancel buttons. A tradeoff.</remarks>
		protected override bool ProcessCmdKey(ref Message msg, Keys keyData)
		{
			bool retval;
			if (keyData == Keys.Enter)
			{
				this.ValidateEditText();
				((TextBox) this.Controls[1]).Select(this.Controls[1].Text.Length, 0); // place text cursor at value's end
				retval = true; // handled, no ding!
			}
			else if (keyData == Keys.Escape)
			{
				this.Controls[1].Text = _savedValue.ToString(); // revert to last committed value
				((TextBox) this.Controls[1]).Select(this.Controls[1].Text.Length, 0); // place text cursor at value's end
				retval = true; // handled, no ding!
			}
			else // all other keys
			{
				retval = base.ProcessCmdKey(ref msg, keyData);
			}
			return retval;
		}

		/// <summary>
		/// When the value of this widget changes, i.e., when a numeric value is "committed," then save the value
		/// so that we can revert back to it when Escape is pressed. We set it after the base call so that if a
		/// client handles the event and changes this widget's value, we'll have the most up-to-date "committed"
		/// value saved.
		/// </summary>
		/// <param name="e">The arguments for this event.</param>
		protected override void OnValueChanged(EventArgs e)
		{
			base.OnValueChanged(e);
			_savedValue = this.Value;
		}
	}
}
