﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Text;
using System.Windows.Forms;

namespace WobbrockLib.Controls
{
    /// <summary>
    /// Widget for a numeric updown control that can be made to only show even numbers or 
	/// odd numbers.
    /// </summary>
    [ToolboxBitmap(typeof(System.Windows.Forms.NumericUpDown))]
    public partial class EvenOddNumericUpDown : AbsNumericUpDown
    {
        private decimal _lastValue;
        private bool _preventReentry;
        private bool _isEven;

		/// <summary>
		/// Constructs a new NumericUpDown control that specifically takes either only
		/// even integers or only odd integers.
		/// </summary>
		public EvenOddNumericUpDown() 
			: this(null)
		{
			// do nothing
		}

        /// <summary>
        /// Constructs a new NumericUpDown control that specifically takes either only
        /// even integers or only odd integers in the given container.
        /// </summary>
        /// <param name="container">The container in which to construct the up/down control.</param>
        public EvenOddNumericUpDown(IContainer container)
        {
            container?.Add(this);
            InitializeComponent();
            base.Increment = 2;
			_preventReentry = false;
			_isEven = true;
            _lastValue = this.Value;
        }

		/// <summary>
		/// The increment must be an even number so that the value stays even (or odd) with each
		/// increment.
		/// </summary>
		[Browsable(true)]
		[Category("Data")]
		[Description("Gets or sets the increment for this control. This value must be an even integer.")]
		[DefaultValue(2)]
		[RefreshProperties(RefreshProperties.Repaint)]
        public new decimal Increment
        {
            get
			{
				return base.Increment;
			}
			set
			{
				if ((value % 2) == 0)
				{
					base.Increment = value;
				}
			}
        }

		/// <summary>
		/// Gets or sets a value indicating whether or not this numeric up down control
		/// displays even integers or odd integers only.
		/// </summary>
		[Browsable(true)]
		[Category("Data")]
		[Description("Gets or sets whether this control shows even (or odd) numbers only.")]
		[DefaultValue(true)]
		[RefreshProperties(RefreshProperties.Repaint)]
        public bool Even
        {
            get
			{
				return _isEven;
			}
            set
			{
				_isEven = value;
				if (
					(value && (this.Value % 2) != 0) ||		// we want even #'s only but are showing an odd #
					(!value && (this.Value % 2) == 0)		// we want odd #'s only but are showing an even #
					)
				{
					OnValueChanged(EventArgs.Empty);
				}
			}
        }

		/// <summary>
		/// This override occurs after the value has been changed. We check if we're handling
		/// only even (or odd) numbers, and ensure that we increment (or decrement) the
		/// new value accordingly.
		/// </summary>
		/// <param name="e">The arguments for this event.</param>
		protected override void OnValueChanged(EventArgs e)
        {
            if (_preventReentry)
            {
                _preventReentry = false;
                return;
            }

            if (
				(_isEven && (this.Value % 2) != 0) ||	// we allow even #'s but are showing an odd #
				(!_isEven && (this.Value % 2) == 0)		// we allow odd #'s but are showing an even #
				)
            {
                _preventReentry = true; // avoids re-entry when setting new value
				if (this.Value > _lastValue)
					this.Value = _lastValue + this.Increment;
				else if (this.Value < _lastValue)
					this.Value = _lastValue - this.Increment;
            }
            _lastValue = this.Value; // update
			base.OnValueChanged(e); // raise the event only after updating the value
        }
    }
}
