﻿using System;
using System.Drawing;
using System.Collections.Generic;
using System.Windows.Forms;
using System.ComponentModel;
using System.Diagnostics;
using WobbrockLib.Windows;

namespace WobbrockLib.Controls
{
	/// <summary>
	/// Implements a control that has three checkboxes for Ctrl, Shift, and Alt, and a drop-down
	/// list that contains an enumeration of keys that can be selected. Allows the user to select
	/// a key shortcut from these combined widgets. Also has a reset button to clear the control.
	/// And allows a user to set the key value in the drop-down list by typing the key on the
	/// keyboard.
	/// </summary>
	[ToolboxBitmap(typeof(System.Windows.Forms.UserControl))]
	public partial class KeyShortcut : UserControl
	{
		#region Fields 

		public const int MaxShortcutKeys = 4; // three modifiers (Ctrl, Shift, Alt) plus a base key
		private const int TabKeyIndex = 1; // index of tab in the key listing

		private Keys _shortcut;
		private bool _isLoading;
		private bool _includeTab;
		private bool _tabHoldFocus;
		private bool _shiftTabPrevent;

		#endregion

		#region Events

		/// <summary>
		/// Event that fires when the shortcut has changed. The event arguments are the
		/// new key shortcut that the widget contains. This event is only fired when 
		/// the user changes the values of any widget in the shortcut UI. It is not fired
		/// when a caller programmatically changes the shortcut value using the Shortcut 
		/// property setter or the Reset method.
		/// </summary>
		public event EventHandler<Keys> ShortcutChanged;

		#endregion

		#region Constructor

		/// <summary>
		/// Default constructor with no key shortcut to load in.
		/// </summary>
		public KeyShortcut()
			: this(Keys.None)
		{
			// do nothing
		}

		/// <summary>
		/// Constructor that is supplied with the given key shortcut to load in.
		/// </summary>
		/// <param name="k">The shortcut to load into this widget.</param>
		public KeyShortcut(Keys k)
		{
			InitializeComponent();
			InitializeKeys();
			_shortcut = k;
			_isLoading = false;
			_includeTab = true;
			_tabHoldFocus = true;
			_shiftTabPrevent = false;
		}

		#endregion

		#region Properties

		/// <summary>
		/// Gets or sets the shortcut value for this widget. This is the main property for the widget.
		/// The ShortcutChanged event will not fire if the shortcut is programmatically set with this
		/// property. That event fires only when the shortcut is changed via the user interface.
		/// </summary>
		/// <remarks>If an attempt is made to set the shortcut value to a key that is not represented
		/// within this shortcut, then the widget is cleared and set to Keys.None.</remarks>
		[Category("Value")]
		[Description("Gets or sets the shortcut key value represented by this control.")]
		[DefaultValue(Keys.None)]
		[RefreshProperties(RefreshProperties.All)]
		public Keys Shortcut
		{
			get
			{
				return _shortcut; // this field should always be in sync with the widget
			}
			set
			{
				if (value != _shortcut) // if different
				{
					if (LoadShortcut(value))
						_shortcut = value; // update if able to load
					else
						this.Reset(); // otherwise, clear widget
				}
			}
		}

		/// <summary>
		/// Gets the current shortcut represented by this widget in the form of an array of Win32
		/// virtual key codes. If there is no shortcut set, this property returns an array of 
		/// length zero; it does not return null.
		/// </summary>
		[Category("Value")]
		[Description("Gets the shortcut as an array of Win32 virtual key codes.")]
		public Win32.VK[] ShortcutVirtualKeys
		{
			get
			{
				Keys keys = _shortcut; // copy

				// extract any modifiers
				bool ctrl = ((keys & Keys.Control) != 0);
				bool alt = ((keys & Keys.Alt) != 0);
				bool shift = ((keys & Keys.Shift) != 0);

				// strip the modifiers
				keys &= ~Keys.Control;
				keys &= ~Keys.Alt;
				keys &= ~Keys.Shift;

				// build a list of keys to test
				List<Win32.VK> vks = new List<Win32.VK>(4);
				if (ctrl)
					vks.Add(Win32.VK.CONTROL);
				if (alt)
					vks.Add(Win32.VK.MENU);
				if (shift)
					vks.Add(Win32.VK.SHIFT);

				// set the remaining base key, if any
				if (keys != Keys.None)
				{
					Win32.VK vkBase = (Win32.VK) keys;
					vks.Add(vkBase);
				}

				return vks.ToArray();
			}
		}

		/// <summary>
		/// Gets the modifier part of the shortcut, if any. If no modifier keys are present,
		/// then returns Keys.None.
		/// </summary>
		[Category("Value")]
		[Description("Gets any modifier keys represented by this control.")]
		[DefaultValue(Keys.None)]
		public new Keys ModifierKeys
		{
			get
			{
				Keys k = Keys.None;
				k |= (_shortcut & Keys.Control);
				k |= (_shortcut & Keys.Shift);
				k |= (_shortcut & Keys.Alt);
				return k;
			}
		}

		/// <summary>
		/// Gets a boolean indicating whether this shortcut has modifiers or not.
		/// </summary>
		[Category("Value")]
		[Description("True if any modifier keys are represented by this control; false otherwise.")]
		[DefaultValue(false)]
		public bool HasModifierKeys
		{
			get { return (this.ModifierKeys != Keys.None); }
		}

		/// <summary>
		/// Gets the non-modifier base key part of this shortcut, if any. If no base keys are present,
		/// then returns Keys.None.
		/// </summary>
		[Category("Value")]
		[Description("Gets the base key represented by this control. The base key is the key unmodified by Ctrl, Shift, or Alt.")]
		[DefaultValue(Keys.None)]
		public Keys BaseKey
		{
			get
			{
				Keys k = _shortcut; // copy
				k &= ~Keys.Control; // strip
				k &= ~Keys.Shift; // strip
				k &= ~Keys.Alt; // strip
				return k;
			}
		}

		/// <summary>
		/// Gets a boolean indicating whether this shortcut has a non-modifier base key or not.
		/// </summary>
		[Category("Value")]
		[Description("True if a base key is represented by this control; false otherwise.")]
		[DefaultValue(false)]
		public bool HasBaseKey
		{
			get { return (this.BaseKey != Keys.None); }
		}

		/// <summary>
		/// Gets or sets the style of the drop down box on the widget. Determines whether 
		/// a user can type into the drop down box or can only select from its pre-determined
		/// list.
		/// </summary>
		[Category("Appearance")]
		[Description("Determines the style of the drop down widget within this control.")]
		[DefaultValue(ComboBoxStyle.DropDown)]
		public ComboBoxStyle DropDownStyle
		{
			get { return cboKeys.DropDownStyle; }
			set { cboKeys.DropDownStyle = value; }
		}

		/// <summary>
		/// Gets or sets whether the Tab key will be available as a shortcut. Some applications
		/// may wish to remove Tab as an option because of its role in moving the focus among
		/// widgets.
		/// </summary>
		[Category("Behavior")]
		[Description("Determines whether the Tab key is available as a shortcut.")]
		[DefaultValue(true)]
		public bool IncludeTab
		{
			get
			{
				return _includeTab;
			}
			set
			{
				if (value && !_includeTab) // add Tab in
				{
					Debug.Assert((Keys) cboKeys.Items[TabKeyIndex] != Keys.Tab);
					cboKeys.Items.Insert(TabKeyIndex, Keys.Tab);
				}
				else if (!value && _includeTab) // pull Tab out
				{
					Debug.Assert((Keys) cboKeys.Items[TabKeyIndex] == Keys.Tab);
					cboKeys.Items.RemoveAt(TabKeyIndex);
				}
				_includeTab = value;
			}
		}

		#endregion

		#region String Conversion

		/// <summary>
		/// Converts this key shortcut into a string representation. This is the preferred method
		/// for turning the shortcut represented by this control into a string. A different result
		/// might be obtained by using this.Shortcut.ToString, because that amounts to a Key.ToString
		/// call, bypassing the KeysConverter class.
		/// </summary>
		/// <returns>This key shortcut as a string representation.</returns>
		public override string ToString()
		{
			return ToString(_shortcut); // call into static fn
		}

		/// <summary>
		/// Converts the given Keys value into a string representation. This is the preferred function
		/// for turning a Keys value into a string. A different result might be obtained by using k.ToString, 
		/// because that amounts to a Key.ToString call, which bypasses the KeysConverter class.
		/// </summary>
		/// <param name="k">The given Keys value.</param>
		/// <returns>A string representation of the given Keys value.</returns>
		public static string ToString(Keys k)
		{
			if (k != Keys.None)
			{
				KeysConverter kc = new KeysConverter();
				return kc.ConvertToString(k);
			}
			else return String.Empty;
		}

		/// <summary>
		/// Converts the given string representation of a key shortcut into a Keys value.
		/// </summary>
		/// <param name="s">The string representation of a Keys value.</param>
		/// <returns>The key representation of the given string.</returns>
		public static Keys FromString(string s)
		{
			if (s != String.Empty)
			{
				KeysConverter kc = new KeysConverter();
				try
				{
					return (Keys) kc.ConvertFromString(s);
				}
				catch (ArgumentException aex)
				{
					Console.WriteLine(aex);
					return Keys.None;
				}
			}
			else return Keys.None;
		}

		#endregion

		#region Form Events

		/// <summary>
		/// Initially loads the shortcut value given in the constructor into this control. If
		/// the given Keys value cannot be represented, then the control is set to Keys.None,
		/// which means all widgets are unset.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void KeyShortcut_Load(object sender, EventArgs e)
		{
			cmdClear.Height = cboKeys.Height;

			if (!LoadShortcut(_shortcut))
				this.Reset(); // clear our widget if key can't be represented
		}

		/// <summary>
		/// When the form is resized, keep the reset buttom aligned with the combo box key
		/// selection button.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void KeyShortcut_SizeChanged(object sender, EventArgs e)
		{
			cmdClear.Top = cboKeys.Top;
			cmdClear.Height = cboKeys.Height;
		}

		/// <summary>
		/// Override the form's handler for processing command keys. We focus on keys that
		/// are toggles, often have an LED associated with them, and thus alter the user's
		/// keyboard state. These are caps lock, number lock, and scroll lock. When any of
		/// these are pressed, we want to prevent them. However, we cannot programmatically 
		/// set their state. Instead, we have to simulate sending the same keypress again 
		/// to undo the user's keypress. But that starts an infinite cycle of keyboard messages
		/// and this function being called. To avoid this problem, we set a flag preventing
		/// re-entry. It only works because the SendWait function blocks until the key message
		/// has been processed, meaning re-entry is attempted, denied, and the function returns
		/// true, indicating the synthesized keypress was handled. Thus, these toggle keys are
		/// undone.
		/// </summary>
		/// <param name="msg">The Windows message that was send to WndProc and shunted here.</param>
		/// <param name="keyData">The extracted key data for the command keypress.</param>
		/// <returns>True if we have handled the key; false otherwise.</returns>
		/// <see cref="https://docs.microsoft.com/en-us/dotnet/api/system.windows.forms.sendkeys?view=netframework-4.8"/>
		//protected override bool ProcessCmdKey(ref Message msg, Keys keyData)
		//{
		//	if (_preventCmdKey)
		//		return true; // handled synthesized keypress
		//
		//	if (cboKeys.Focused)
		//	{
		//		strip any modifiers first
		//		keyData &= ~Keys.Control;
		//		keyData &= ~Keys.Shift;
		//		keyData &= ~Keys.Alt;
		//
		//		_preventCmdKey = true; // prevent re-entry for the SendWait call
		//		switch (keyData)
		//		{
		//			case Keys.Capital:
		//				SendKeys.SendWait("{CAPSLOCK}");
		//				break;
		//			case Keys.NumLock:
		//				SendKeys.SendWait("{NUMLOCK}");
		//				break;
		//			case Keys.Scroll:
		//				SendKeys.SendWait("{SCROLLLOCK}");
		//				break;
		//		}
		//		_preventCmdKey = false;
		//	}
		//	return base.ProcessCmdKey(ref msg, keyData);
		//}

		#endregion

		#region Widget Event Handlers

		/// <summary>
		/// Click handler for the Clear button that clears the state of the widget.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		/// <remarks>Instead of setting the property, we could also call this.Reset.</remarks>
		private void cmdClear_Click(object sender, EventArgs e)
		{
			if (_shortcut != Keys.None)
			{
				this.Shortcut = Keys.None; // use the property setter
				ShortcutChanged?.Invoke(this, Keys.None);
			}
		}

		/// <summary>
		/// When ANY of the checkboxes change, this event is fired and causes us to possibly 
		/// update the shortcut value. We might, in turn, fire an event as well that the 
		/// shortcut has changed.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void CheckChanged(object sender, EventArgs e)
		{
			// if we're just loading in values programmatically, abort this event handler
			if (_isLoading)
				return;

			// read the state of the widget and if it is different than before, fire a changed event
			Keys k = ReadShortcut();
			if (k != _shortcut) // if different
			{
				_shortcut = k; // update
				ShortcutChanged?.Invoke(this, k);
			}
		}

		/// <summary>
		/// When the combo box holding all of the key values changes, this event is fired and
		/// causes us to possibly update the shortcut value. We might in turn fire an event as 
		/// well that the shortcut has changed.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void cboKeys_SelectedIndexChanged(object sender, EventArgs e)
        {
			// if we're just loading in the selection programmatically, abort this event handler
			if (_isLoading)
				return;

			// disable and uncheck checkboxes if the base key happens to be a modifier key (Ctrl, Shift, or Alt)
			if (cboKeys.SelectedIndex != -1)
				HandleModifierAsBaseKey((Keys) cboKeys.SelectedItem);			

			// read the state of the widget and if it is different than before, fire a changed event
			Keys k = ReadShortcut();
			if (k != _shortcut) // if different
			{
				_shortcut = k; // update
				ShortcutChanged?.Invoke(this, k);
			}
		}

		/// <summary>
		/// Keep the text cursor caret at the end of the combo box text.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void cboKeys_MouseClick(object sender, MouseEventArgs e)
		{
			if (e.Button == MouseButtons.Left)
			{
				cboKeys.Select(cboKeys.Text.Length, 0);
			}
		}

		/// <summary>
		/// Enable the entire drop down widget to drop down when it is clicked upon.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void cboKeys_MouseDown(object sender, MouseEventArgs e)
		{
			this.Cursor = Cursors.Arrow;
			cboKeys.DroppedDown = true;
		}

		#endregion

		#region Typing a Key to Set the Base Key and Handling Tab and Shift+Tab

		/// <summary>
		/// When the combo box has the focus within the KeyShortcut widget, it allows the
		/// user to type a key and have the base key set in the combo box. Some keys are 
		/// just a single letter, e.g., "A", while others are a whole word, like "Escape".
		/// If the user types, e.g., the "Esc" key, then "Escape" should show in the combo
		/// box. Keys that are eligible for setting this way must appear in the combo box's
		/// list of items (see the InitializeKeys helper function).
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		/// <remarks>All keys are set on key down except for Shift as a base key. Why? Because 
		/// the user might type Shift+Tab to move the focus backward in the Tab order, and we 
		/// don't want this action to cause Shift to be set as the base key prematurely. So
		/// Shift is set as a base key only in cboKeys_KeyUp.
		/// </remarks>
		private void cboKeys_KeyDown(object sender, KeyEventArgs e)
		{
			// we only set Shift as a base key on KeyUp, because 
			// it could be a part of Shift+Tab to move the focus.
			if (e.KeyCode == Keys.ShiftKey)
				return;

			// strip any modifiers that might have been held down
			Keys k = e.KeyCode;
			k &= ~Keys.Control;
			k &= ~Keys.Shift;
			k &= ~Keys.Alt;

			// convert the pressed base key's code to a string
			string s = ToString(k);

			// if we have a sensible string representation, see if it is in our drop down list.
			// if so, select it and place its text in our combo box.
			if (s != String.Empty && cboKeys.FindStringExact(s) != -1)
			{
				cboKeys.SelectedItem = s;
				cboKeys.Text = s;
			}

			// set the text cursor to be at the end of the combo box text.
			cboKeys.DroppedDown = false;
			cboKeys.Select(cboKeys.Text.Length, 0);

			// we've programmatically dealt with the key.
			e.SuppressKeyPress = true;
			e.Handled = true;
		}

		/// <summary>
		/// When Shift is being set as a base key, we do so not on key down, like for all
		/// other keys, but here in key up, because Shift could have been pressed as part
		/// of a Shift+Tab action. However, the focus could arrive at the cboKeys widget
		/// after Shift+Tab was pressed when the focus was on the Reset button, which would
		/// cause Shift to go up on cboKeys, even though it was still part of a Shift+Tab 
		/// action. The prevent flag here prevents this occurrence from setting the Shift
		/// key as a base key in the combo box.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void cboKeys_KeyUp(object sender, KeyEventArgs e)
		{
			if (e.KeyCode == Keys.ShiftKey)
			{
				if (_shiftTabPrevent)
				{
					_shiftTabPrevent = false;
				}
				else
				{
					string s = ToString(e.KeyCode);
					if (s != String.Empty && cboKeys.FindStringExact(s) != -1)
					{
						cboKeys.SelectedItem = s;
						cboKeys.Text = s;
					}
					cboKeys.DroppedDown = false;
					cboKeys.Select(cboKeys.Text.Length, 0);
				}
				e.SuppressKeyPress = true;
				e.Handled = true;
			}
		}

		/// <summary>
		/// If Shift+Tab is pressed to move the focus backwards from the Reset button to 
		/// the combo box, then the Shift key will go up on the combo box. We don't want to
		/// set the Shift key as the pressed key on the combo box at this moment, so to prevent
		/// that, we set a flag here indicating that this has happened. Then, cboKeys_KeyUp
		/// can look for this flag and avoid setting the Shift key as the pressed key on a 
		/// Shift+Tab action.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void cmdClear_PreviewKeyDown(object sender, PreviewKeyDownEventArgs e)
		{
			if (e.KeyCode == Keys.Tab && e.Shift)
			{
				_shiftTabPrevent = true;
			}
		}

		/// <summary>
		/// The Tab key does not generate a key down event in cboKeys, so if the IncludeTab
		/// property is true, and we want users to be able to press Tab to set the base key,
		/// we have to capture it here and then send it manually along to the KeyDown handler.
		/// Furthermore, if Tab can be pressed to set the key, then Tab shouldn't move the
		/// focus to the next widget in such a case. But we also don't want cboKeys to become
		/// a "focus trap," so if Tab is already set as the base key, then pressing it again
		/// will move the focus forward. We set a flag here to indicate whether pressing Tab
		/// should hold the focus where it is, or allow it to move on. It is allowed to move
		/// on if we either don't include Tab as a base key in the first place, or if Tab
		/// is already set as the base key. The influence of this flag occurs in the form's
		/// ProcessTabKey override, where the flag is checked to determine whether the focus
		/// is allowed to move on or not.
		/// </summary>
		/// <param name="sender">The sender of this event.</param>
		/// <param name="e">The arguments for this event.</param>
		private void cboKeys_PreviewKeyDown(object sender, PreviewKeyDownEventArgs e)
		{
			if (e.KeyCode == Keys.Tab)
			{
				// set a flag to hold the focus in the combo box when Tab is pressed but only 
				// if Tab was not already the selection there.
				Keys prev = (cboKeys.SelectedIndex != -1) ? (Keys) cboKeys.SelectedItem : Keys.None;
				_tabHoldFocus = (_includeTab && prev != Keys.Tab);
				cboKeys_KeyDown(sender, new KeyEventArgs(Keys.Tab)); // patch thru
			}
		}

		/// <summary>
		/// This override handles the pressing of the Tab key to move the focus around among
		/// the widgets on the form. We check the status of a flag that was possibly set in 
		/// cboKeys_PreviewKeyDown, where we might hold the focus on cboKeys when Tab is pressed
		/// to set the base key as Tab. Or, we might allow the focus to move on, if either Tab
		/// cannot be set as a base key (i.e., the IncludeTab property is false), or if Tab has
		/// already been set as the base key.
		/// </summary>
		/// <param name="forward">Whether to move the focus forward or backward.</param>
		/// <returns>True if we are to hold the focus where it is; false if the focus should
		/// be allowed to move to the next widget in the Tab order.</returns>
		protected override bool ProcessTabKey(bool forward)
		{
			if (cboKeys.Focused) // combo box has focus
				return _tabHoldFocus ? true : base.ProcessTabKey(forward); // see cboKeys_PreviewKeyDown
			else
				return base.ProcessTabKey(forward); // Mod checkboxes or the Reset button have focus
		}

		#endregion

		#region Public Methods

		/// <summary>
		/// Reset the state of this widget. Note that no ShortcutChanged event is fired for
		/// programmaticaly resetting the widget. That event is fired only for changes that
		/// occur in the user interface, i.e., via clicking the modifier checkboxes, selecting
		/// from the combo box, or clicking the reset button.
		/// </summary>
		public void Reset()
		{
			LoadShortcut(Keys.None);
			_shortcut = Keys.None;
		}

		/// <summary>
		/// Determines whether the given key is a valid shortcut representable by this control or not.
		/// Note that shortcuts only consisting of modifier keys but no base key are not considered 
		/// valid shortcuts. Also, the base key given must be one of the keys present in the dropdown
		/// list of this control for it to be a valid shortcut.
		/// </summary>
		/// <param name="k">The key to test.</param>
		/// <returns>True if the given key is a valid shortcut representable by this control; false otherwise.</returns>
		public bool IsValidShortuct(Keys k)
		{
			// strip out any modifier keys first
			k &= ~Keys.Control;
			k &= ~Keys.Shift;
			k &= ~Keys.Alt;

			// convert the key to a string and see if it matches in the dropdown list
			string s = ToString(k); // NOT k.ToString!
			return (cboKeys.FindStringExact(s) != -1);
		}

		#endregion

		#region Private Helpers

		/// <summary>
		/// Loads up the given shortcut key into the widget's settings.
		/// </summary>
		/// <param name="k">The key to load into the widget.</param>
		/// <returns>True if the shortcut can be represented in the widget successfully; false otherwise.</returns>
		private bool LoadShortcut(Keys k)
		{
			bool success = true;

			// set this flag to prevent widget-changed events from firing 
			// when programmatically loading the shortcut into its widgets.
			_isLoading = true;

			// clear the three modifier checkboxes
			chkCtrl.Checked = false;
			chkShift.Checked = false;
			chkAlt.Checked = false;

			// now set the relevant checkboxes
			if ((k & Keys.Control) != 0)
			{
				chkCtrl.Checked = true;
				k &= ~Keys.Control; // strip out
			}
			if ((k & Keys.Shift) != 0)
			{
				chkShift.Checked = true;
				k &= ~Keys.Shift; // strip out
			}
			if ((k & Keys.Alt) != 0)
			{
				chkAlt.Checked = true;
				k &= ~Keys.Alt; // strip out
			}

			// disable and uncheck checkboxes if a base key happens 
			// to itself be a modifier key
			HandleModifierAsBaseKey(k);

			// now set the dropdown combo box
			if (k != Keys.None)
			{
				if (this.IsValidShortuct(k))
					cboKeys.SelectedItem = k; // set what's left
				else
					success = false;
			}
			else // clear the combo box for Keys.None
			{
				cboKeys.SelectedIndex = -1; // clear combo box
			}

			_isLoading = false; // unset
			return success;
		}

		/// <summary>
		/// When the combo box is set to a base key, check to see if that base key is, itself,
		/// a modifier key (Ctrl, Shift, or Alt). Such keys can serve as base keys when they
		/// are pressed as any other base key would be. But if they are used as a base key, then
		/// we must gray out their corresponding modifier key checkbox. We must also uncheck that
		/// checkbox if it was previously checked.
		/// </summary>
		/// <param name="baseKey">The base key for setting in the combo box.</param>
		/// <remarks>
		/// <para>
		/// It is assumed that any modifier keys have already been stripped out of the given base key.
		/// </para>
		/// <para>
		/// Note that Keys.Control, Keys.Shift, and Keys.Alt are modifier keys. As base keys, these
		/// are encoded as Keys.ControlKey, Keys.ShiftKey, and Keys.Menu, which all have left and 
		/// right variants as well.
		/// </para>
		/// </remarks>
		private void HandleModifierAsBaseKey(Keys baseKey)
		{
			chkCtrl.Enabled = (baseKey != Keys.ControlKey);
			chkShift.Enabled = (baseKey != Keys.ShiftKey);
			chkAlt.Enabled = (baseKey != Keys.Menu);

			if (baseKey == Keys.ControlKey)
				chkCtrl.Checked = false;
			if (baseKey == Keys.ShiftKey)
				chkShift.Checked = false;
			if (baseKey == Keys.Menu)
				chkAlt.Checked = false;
		}

		/// <summary>
		/// Reads in the current shortcut based on the widget's current UI state. 
		/// </summary>
		/// <returns>The key representing the widget's current state.</returns>
		private Keys ReadShortcut()
        {
			Keys k = Keys.None;
			if (chkCtrl.Checked)
				k |= Keys.Control;
			if (chkShift.Checked)
				k |= Keys.Shift;
			if (chkAlt.Checked)
				k |= Keys.Alt;
			if (cboKeys.SelectedIndex != -1)
				k |= (Keys) cboKeys.SelectedItem;
			return k;
		}

		/// <summary>
		/// Helper function that initializes the drop down list with key values corresponding 
		/// to keys on the keyboard. Any key that is not represented in this list can not be set 
		/// as a shortcut using this control.
		/// </summary>
		/// <remarks>
		/// <para>
		/// The commented-out line would enumerate all of the Keys enumerated type 
		/// values, but some of these are undesirable for listing in this control. So we
		/// manually place a subset of these values into the control.
		/// </para>
		/// <para>
		/// The string representations that result for each of the added keys usually, but
		/// not always, match the keys value. For example, Keys.Insert renders as "Ins".
		/// These strings are NOT the result of a Keys.ToString call. Rather, they are
		/// the KeysConverter.ConvertToString equiavlents. See the static ToString function 
		/// in this class, and its use in the IsValidShortcut method.
		/// </para>
		/// </remarks>
		private void InitializeKeys()
		{
			_isLoading = true;
			//cboKeys.DataSource = Keys.GetNames(typeof(Keys));
			cboKeys.Items.AddRange(new object[] {
				Keys.Escape,
				Keys.Tab,			// Ensure Tab's index is the TabKeyIndex const value
				Keys.Capital,
				Keys.ShiftKey,
				Keys.ControlKey,
				Keys.Menu,			// Alt
				Keys.Space,
				Keys.Back,
				Keys.Enter,
				Keys.Home,
				Keys.End,
				Keys.Insert,
				Keys.Delete,
				Keys.Left,
				Keys.Up,
				Keys.Right,
				Keys.Down,
				Keys.PageUp,
				Keys.Next,			// PageDown
				Keys.Pause,			// Break
				Keys.Scroll,		// Scroll Lock
				Keys.D0,
				Keys.D1,
				Keys.D2,
				Keys.D3,
				Keys.D4,
				Keys.D5,
				Keys.D6,
				Keys.D7,
				Keys.D8,
				Keys.D9,
				Keys.A,
				Keys.B,
				Keys.C,
				Keys.D,
				Keys.E,
				Keys.F,
				Keys.G,
				Keys.H,
				Keys.I,
				Keys.J,
				Keys.K,
				Keys.L,
				Keys.M,
				Keys.N,
				Keys.O,
				Keys.P,
				Keys.Q,
				Keys.R,
				Keys.S,
				Keys.T,
				Keys.U,
				Keys.V,
				Keys.W,
				Keys.X,
				Keys.Y,
				Keys.Z,
				Keys.NumLock,
				Keys.NumPad0,
				Keys.NumPad1,
				Keys.NumPad2,
				Keys.NumPad3,
				Keys.NumPad4,
				Keys.NumPad5,
				Keys.NumPad6,
				Keys.NumPad7,
				Keys.NumPad8,
				Keys.NumPad9,
				Keys.Multiply,
				Keys.Add,
				Keys.Separator,
				Keys.Subtract,
				Keys.Decimal,
				Keys.Divide,
				Keys.F1,
				Keys.F2,
				Keys.F3,
				Keys.F4,
				Keys.F5,
				Keys.F6,
				Keys.F7,
				Keys.F8,
				Keys.F9,
				Keys.F10,
				Keys.F11,
				Keys.F12,
				Keys.F13,
				Keys.F14,
				Keys.F15,
				Keys.F16,
				Keys.F17,
				Keys.F18,
				Keys.F19,
				Keys.F20,
				Keys.F21,
				Keys.F22,
				Keys.F23,
				Keys.F24,
				Keys.OemClear,			// Fn?
				Keys.Oemtilde,			// `
				Keys.OemMinus,			// -
				Keys.Oemplus,			// +
				Keys.OemOpenBrackets,	// [
				Keys.Oem6,				// ]
				Keys.OemBackslash,		// \
				Keys.Oem5,				// \
				Keys.Oem1,				// ;
				Keys.Oem7,				// '
				Keys.Oemcomma,			// ,
				Keys.OemPeriod,			// .
				Keys.OemQuestion		// /
			});
			_isLoading = false;
		}

		#endregion

	}
}
