﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Diagnostics;
using System.IO;
using Microsoft.VisualBasic;
using WobbrockLib.Extensions;
using WobbrockLib.Types;

namespace WobbrockLib.Controls
{
    /// <summary>
    /// 
    /// </summary>
    [ToolboxBitmap(typeof(System.Windows.Forms.Control))]
    public partial class Plot : Control
    {
        #region Inner Class: Series

        /// <summary>
        /// This class represents a series of data points that can be plotted with a <b>Plot</b> instance.
        /// </summary>
        public class Series
        {
            public static readonly Series Empty;
            public string Name;
            public Color PointColor;
            public Color LineColor;
            public bool DrawPoints;
            public bool DrawLine;
            private List<PointR> _points;

            /// <summary>
            /// Constructs a <i>Series</i> instance, which is a class that contains a series
            /// of data points that can be plotted on a <b>Plot</b> widget.
            /// </summary>
            /// <param name="name">The name of the series.</param>
            /// <param name="points">The color of the points to draw, if drawn.</param>
            /// <param name="line">The color of the line to draw, if drawn.</param>
            /// <param name="drawPoints">If <b>true</b>, points will be drawn; if <b>false</b>, they will not be drawn.</param>
            /// <param name="drawLine">If <b>true</b>, a line connecting each point will be drawn; if <b>false</b>, these lines will not be drawn.</param>
            public Series(string name, Color points, Color line, bool drawPoints, bool drawLine)
            {
                Name = name;
                PointColor = points;
                LineColor = line;
                DrawPoints = drawPoints;
                DrawLine = drawLine;
                _points = new List<PointR>();
            }

            /// <summary>
            /// Indexer which gets the data point at the given index in this series.
            /// </summary>
            /// <param name="index">The 0-based index of the point to get.</param>
            /// <returns>The point at the given index, or <b>PointR.Empty</b> if <i>index</i>
            /// is out of range.</returns>
            public PointR this[int index]
            {
                get
                {
                    if (0 <= index && index < _points.Count)
                    {
                        return _points[index];
                    }
                    else
                    {
                        return PointR.Empty;
                    }
                }
            }

            /// <summary>
            /// Adds the supplied points to this data series for plotting.
            /// </summary>
            /// <param name="points">A <b>List&lt;PointR&gt;</b> instance containing points to add.</param>
            public void AddPoints(List<PointR> points)
            {
                foreach (PointR p in points)
                {
                    _points.Add(p);
                }
            }

            /// <summary>
            /// Adds the supplied points to this data series for plotting.
            /// </summary>
            /// <param name="points">A <b>PointR[]</b> instance containing points to add.</param>
            public void AddPoints(PointR[] points)
            {
                for (int i = 0; i < points.Length; i++)
                {
                    _points.Add(points[i]);
                }
            }

            /// <summary>
            /// Adds the supplied <i>p</i> point to this data series.
            /// </summary>
            /// <param name="p">The point to add.</param>
            public void AddPoint(PointR p)
            {
                _points.Add(p);
            }

            /// <summary>
            /// Adds the supplied <i>(x,y)</i> point to this data series.
            /// </summary>
            /// <param name="x">The x-coordinate of the point to add.</param>
            /// <param name="y">The y-coordinate of the point to add.</param>
            public void AddPoint(double x, double y)
            {
                _points.Add(new PointR(x, y));
            }

            /// <summary>
            /// Removes the data point at the given <i>index</i> from this series.
            /// </summary>
            /// <param name="index">The 0-based index of the point to remove.</param>
            /// <returns>The removed point, or <b>PointR.Empty</b> if <i>index</i> is
            /// out of range.</returns>
            public PointR RemovePoint(int index)
            {
                PointR p = PointR.Empty;
                if (0 <= index && index < _points.Count)
                {
                    p = _points[index];
                    _points.RemoveAt(index);
                }
                return p;
            }

            /// <summary>
            /// Gets the number of points in this data series.
            /// </summary>
            public int NumPoints
            {
                get
                {
                    return (_points != null) ? _points.Count : 0;
                }
            }

            /// <summary>
            /// Gets a reference to a list of all the points in this data series.
            /// </summary>
            public List<PointR> Points
            {
                get
                {
                    return _points;
                }
            }

            /// <summary>
            /// Gets the domain (x-values) and range (y-values) of this series; that is, the minimum 
            /// and maximum x- and y-values. These are returned in a <b>System.Drawing.RectangleR</b>
            /// whose <i>Left</i> and <i>Right</i> properties indicate <i>(minX, maxX)</i>, and whose
            /// <i>Top</i> and <i>Bottom</i> properties indicate <i>(minY, maxY)</i>, respectively.
            /// </summary>
            public RectangleR DomainRange
            {
                get { return GeotrigEx.FindBoundingBox(_points); }
            }
        }

        #endregion

        #region Fields

        private const int DefaultPlotMargins = 40;
        private const string DefaultPlotMarginsString = "40";
        private const int DefaultTicks = 10;
        private const int DefaultDecimals = 1;

        private List<Series> _series; // we can plot multiple series
        private RectangleR _plotRect; // we draw within this rect on our control
        private RectangleR _domainRange; // the domain and range of the data being plotted -- for tick marks and labels

        private Padding _plotMargin; // spacing around outside of plot; determines _plotRect

        private string _xName; // name of xAxis
        private string _yName; // name of yAxis

        private int _xTicks; // number of tick marks on X-axis
        private int _yTicks; // number of tick marks on Y-axis

        private int _xDecimals; // number of decimals to round to for x-axis labels
        private int _yDecimals; // number of decimals to round to for y-axis labels

        private bool _xUseLabels; // if true, x labels are provided by client 
        private bool _yUseLabels; // if true, y labels are provided by client

        private string[] _xLabels; // custom x labels provided by client
        private string[] _yLabels; // custom y labels provided by client

        private bool _showLegend; // whether or not to show a legend for the series

        #endregion

        #region Constructor

        /// <summary>
        /// 
        /// </summary>
        public Plot()
			: this(null)
        {
            // do nothing
        }

		/// <summary>
		/// 
		/// </summary>
		/// <param name="container"></param>
		public Plot(IContainer container)
		{
			container?.Add(this);
			InitializeComponent();
			MyInitialization();
		}

		/// <summary>
		/// 
		/// </summary>
		private void MyInitialization()
		{
			this.SetStyle(ControlStyles.AllPaintingInWmPaint | ControlStyles.UserPaint | ControlStyles.OptimizedDoubleBuffer, true);

			this.BackColor = Color.White;
			_series = new List<Series>();

			this.Text = "Plot";
			_plotMargin = new Padding(DefaultPlotMargins);
			_xName = "X-axis";
			_yName = "Y-axis";
			_xTicks = DefaultTicks;
			_yTicks = DefaultTicks;
			_xDecimals = DefaultDecimals;
			_yDecimals = DefaultDecimals;
			_xUseLabels = false;
			_yUseLabels = false;
			_xLabels = null;
			_yLabels = null;
			_showLegend = false;
		}

        #endregion

        #region Properties

        /// <summary>
        /// Gets or sets the title of the plot, which is the same as the plot's Text property.
        /// </summary>
        [Category("Appearance")]
        [Description("The title of the plot. This is also the plot's Text property.")]
        [DefaultValue("Plot")]
        [RefreshProperties(RefreshProperties.Repaint)]
        public string Title
        {
            get { return this.Text; }
            set { this.Text = value; }
        }

        /// <summary>
        /// Gets or sets the margin spacing of the actual plot inside the Plot control.
        /// </summary>
        [Category("Appearance")]
        [Description("The margin spacing of the actual plot inside the Plot control.")]
        [DefaultValue(typeof(Padding), DefaultPlotMarginsString)]
        [RefreshProperties(RefreshProperties.Repaint)]
        public Padding PlotMargin
        {
            get { return _plotMargin; }
            set
            {
                _plotMargin = value;
                OnResize(EventArgs.Empty);
            }
        }

        /// <summary>
        /// Gets or sets the name of the X-axis in the plot control.
        /// </summary>
        [Category("Appearance")]
        [Description("The label for the X-axis of the plot.")]
        [DefaultValue("X-axis")]
        [RefreshProperties(RefreshProperties.Repaint)]
        public string XAxisName
        {
            get { return _xName; }
            set { _xName = value; }
        }

        /// <summary>
        /// Gets or sets the name of the Y-axis in the plot control.
        /// </summary>
        [Category("Appearance")]
        [Description("The label for the Y-axis of the plot.")]
        [DefaultValue("Y-axis")]
        [RefreshProperties(RefreshProperties.Repaint)]
        public string YAxisName
        {
            get { return _yName; }
            set { _yName = value; }
        }

        /// <summary>
        /// Gets or sets the number of tick marks along the X-axis. These will be
        /// automatically labeled by the plot control depending on the domain
        /// of the series data, or the client can provide custom labels.
        /// </summary>
        [Category("Appearance")]
        [Description("The number of labeled tick marks along the X-axis of the plot. The origin does not count as a tick.")]
        [DefaultValue(DefaultTicks)]
        [RefreshProperties(RefreshProperties.Repaint)]
        public int XAxisTicks
        {
            get { return _xTicks; }
            set
            {
                if (1 <= value && value <= 100)
                    _xTicks = value;
            }
        }

        /// <summary>
        /// Gets or sets the number of tick marks along the Y-axis. These will be
        /// automatically labeled by the plot control depending on the range of
        /// the series data, or the client can provide custom labels.
        /// </summary>
        [Category("Appearance")]
        [Description("The number of labeled tick marks along the Y-axis of the plot. The origin does not count as a tick.")]
        [DefaultValue(DefaultTicks)]
        [RefreshProperties(RefreshProperties.Repaint)]
        public int YAxisTicks
        {
            get { return _yTicks; }
            set
            {
                if (1 <= value && value <= 100)
                    _yTicks = value;
            }
        }

        /// <summary>
        /// Gets or sets the number of decimals to round the axis values to at 
        /// each tick mark.
        /// </summary>
        [Category("Appearance")]
        [Description("The number of significant digits in the X-axis labels.")]
        [DefaultValue(DefaultDecimals)]
        [RefreshProperties(RefreshProperties.Repaint)]
        public int XAxisDecimals
        {
            get { return _xDecimals; }
            set
            {
                if (0 <= value && value <= 5)
                    _xDecimals = value;
            }
        }

        /// <summary>
        /// Gets or sets the number of decimals to round the axis values to at 
        /// each tick mark.
        /// </summary>
        [Category("Appearance")]
        [Description("The number of significant digits in the Y-axis labels.")]
        [DefaultValue(DefaultDecimals)]
        [RefreshProperties(RefreshProperties.Repaint)]
        public int YAxisDecimals
        {
            get { return _yDecimals; }
            set
            {
                if (0 <= value && value <= 5)
                    _yDecimals = value;
            }
        }

        /// <summary>
        /// Gets or sets whether the X-axis will use custom labels provided by the client.
        /// </summary>
        [Category("Appearance")]
        [Description("Indicates whether the X-axis will use custom labels provided by the client.")]
        [DefaultValue(false)]
        [RefreshProperties(RefreshProperties.Repaint)]
        public bool XUseCustomLabels
        {
            get { return _xUseLabels; }
            set { _xUseLabels = value; }
        }

        /// <summary>
        /// Gets or sets whether the Y-axis will use custom labels provided by the client.
        /// </summary>
        [Category("Appearance")]
        [Description("Indicates whether the Y-axis will use custom labels provided by the client.")]
        [DefaultValue(false)]
        [RefreshProperties(RefreshProperties.Repaint)]
        public bool YUseCustomLabels
        {
            get { return _yUseLabels; }
            set { _yUseLabels = value; }
        }

        /// <summary>
        /// Gets or sets the custom labels for the X-axis.
        /// </summary>
        [Category("Appearance")]
        [Description("Provides custom labels for the X-axis.")]
        [DefaultValue(null)]
        [RefreshProperties(RefreshProperties.Repaint)]
        public string[] XAxisLabels
        {
            get { return _xLabels; }
            set
            {
                if (value != null)
                {
                    _xLabels = new string[value.Length];
                    Array.Copy(value, _xLabels, value.Length); // deep copy
                }
                else _xLabels = null;
            }
        }

        /// <summary>
        /// Gets or sets the custom labels for the Y-axis.
        /// </summary>
        [Category("Appearance")]
        [Description("Provides custom labels for the Y-axis.")]
        [DefaultValue(null)]
        [RefreshProperties(RefreshProperties.Repaint)]
        public string[] YAxisLabels
        {
            get { return _yLabels; }
            set
            {
                if (value != null)
                {
                    _yLabels = new string[value.Length];
                    Array.Copy(value, _yLabels, value.Length); // deep copy
                }
                else _yLabels = null;
            }
        }

        /// <summary>
        /// Show or hide the legend for the plot.
        /// </summary>
        [Category("Appearance")]
        [Description("Indicates whether a legend should be shown.")]
        [DefaultValue(false)]
        [RefreshProperties(RefreshProperties.Repaint)]
        public bool Legend
        {
            get { return _showLegend; }
            set { _showLegend = value; }
        }

        #endregion

        #region Context Menu

        /// <summary>
        /// When a right-click comes in, dynamically set up a context menu with commands allowing us to
        /// affect the plot and the series plotted on it, if any. We do not install this menu as the
        /// actual context menu of the control, but simply show it after assembling it manually.
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        private void Plot_MouseClick(object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Right)
            {
                ContextMenu menu = new ContextMenu();
                menu.MenuItems.Add("Set Plot Properties...", new EventHandler(mniSetPlot_Click));
                menu.MenuItems.Add("Set X-Axis Properties...", new EventHandler(mniSetXAxis_Click));
                menu.MenuItems.Add("Set Y-Axis Propertes...", new EventHandler(mniSetYAxis_Click));
                menu.MenuItems.Add("Copy Image to Clipboard", new EventHandler(mniCopy_Click));
                menu.MenuItems.Add("Export All", new EventHandler(mniExportAll_Click));
                if (_series.Count > 0)
                    menu.MenuItems.Add("-"); // separator

                foreach (Series s in _series)
                {
                    MenuItem mni = new MenuItem(s.Name);
                    mni.Tag = s; // store a series reference in the Tag field for retrieval in the event handlers!
                    mni.MenuItems.Add("Change Name...", new EventHandler(mniChangeName_Click));
                    mni.MenuItems.Add("Toggle Points", new EventHandler(mniTogglePoints_Click));
                    mni.MenuItems.Add("Toggle Line", new EventHandler(mniToggleLine_Click));
                    mni.MenuItems.Add("Color Points", new EventHandler(mniColorPoints_Click));
                    mni.MenuItems.Add("Color Line", new EventHandler(mniColorLine_Click));
                    mni.MenuItems.Add("Export", new EventHandler(mniExportSeries_Click));
                    menu.MenuItems.Add(mni);
                }

                // show the just-assembled context menu at the click location
                menu.Show(this, new Point(e.X, e.Y));
            }
        }

        /// <summary>
        /// Allows the user to set the title and margins of the plot with a Visual Basic input box.
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        /// <remarks>This handler is dynamically installed when the context menu is created at runtime.</remarks>
        private void mniSetPlot_Click(object sender, EventArgs e)
        {
            string s = Interaction.InputBox("Enter a new title for this plot.", "Plot Title", this.Title, -1, -1);
            if (s != String.Empty)
            {
                this.Title = s;
            }

            s = Interaction.InputBox("Enter the LEFT margin in pixels.", "Left Margin", _plotMargin.Left.ToString(), -1, -1);
            if (s != String.Empty)
            {
                int n;
                if (int.TryParse(s, out n))
                {
                    _plotMargin.Left = n;
                }
            }

            s = Interaction.InputBox("Enter the TOP margin in pixels.", "Top Margin", _plotMargin.Top.ToString(), -1, -1);
            if (s != String.Empty)
            {
                int n;
                if (int.TryParse(s, out n))
                {
                    _plotMargin.Top = n;
                }
            }

            s = Interaction.InputBox("Enter the RIGHT margin in pixels.", "Right Margin", _plotMargin.Right.ToString(), -1, -1);
            if (s != String.Empty)
            {
                int n;
                if (int.TryParse(s, out n))
                {
                    _plotMargin.Right = n;
                }
            }

            s = Interaction.InputBox("Enter the BOTTOM margin in pixels.", "Bottom Margin", _plotMargin.Bottom.ToString(), -1, -1);
            if (s != String.Empty)
            {
                int n;
                if (int.TryParse(s, out n))
                {
                    _plotMargin.Bottom = n;
                }
            }

            DialogResult yesno = MessageBox.Show("Do you want to show a legend for this plot?", "Legend?", MessageBoxButtons.YesNo);
            _showLegend = (yesno == DialogResult.Yes);

            OnResize(EventArgs.Empty);
        }

        /// <summary>
        /// Allows the user to set the properties of the X-axis with a Visual Basic input box. The properties are
        /// its label, number of ticks, and number of decimal places for each tick value.
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        /// <remarks>This handler is dynamically installed when the context menu is created at runtime.</remarks>
        private void mniSetXAxis_Click(object sender, EventArgs e)
        {
            string s = Interaction.InputBox("Enter a new label for the X-axis.", "X-Axis Label", this.XAxisName, -1, -1);
            if (s != String.Empty)
            {
                this.XAxisName = s;
            }

            s = Interaction.InputBox("Enter the number of tick marks for the X-axis (1-100).", "X-Axis Ticks", this.XAxisTicks.ToString(), -1, -1);
            if (s != String.Empty)
            {
                int n;
                if (int.TryParse(s, out n))
                {
                    this.XAxisTicks = n;
                }
            }

            s = Interaction.InputBox("Enter the number of decimal places at each X-axis tick (0-5).", "X-Axis Decimals", this.XAxisDecimals.ToString(), -1, -1);
            if (s != String.Empty)
            {
                int n;
                if (int.TryParse(s, out n))
                {
                    this.XAxisDecimals = n;
                }
            }

            Invalidate();
        }

        /// <summary>
        /// Allows the user to set the properties of the Y-axis with a Visual Basic input box. The properties are
        /// its label, number of ticks, and number of decimal places for each tick value.
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        /// <remarks>This handler is dynamically installed when the context menu is created at runtime.</remarks>
        private void mniSetYAxis_Click(object sender, EventArgs e)
        {
            string s = Interaction.InputBox("Enter a new label for the Y-axis.", "Y-Axis Label", this.YAxisName, -1, -1);
            if (s != String.Empty)
            {
                this.YAxisName = s;
            }

            s = Interaction.InputBox("Enter the number of tick marks for the Y-axis (1-100).", "Y-Axis Ticks", this.YAxisTicks.ToString(), -1, -1);
            if (s != String.Empty)
            {
                int n;
                if (int.TryParse(s, out n))
                {
                    this.YAxisTicks = n;
                }
            }

            s = Interaction.InputBox("Enter the number of decimal places at each Y-axis tick (0-5).", "Y-Axis Decimals", this.YAxisDecimals.ToString(), -1, -1);
            if (s != String.Empty)
            {
                int n;
                if (int.TryParse(s, out n))
                {
                    this.YAxisDecimals = n;
                }
            }

            Invalidate();
        }

        /// <summary>
        /// Copies the plot image to the clipboard.
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        /// <remarks>This handler is dynamically installed when the context menu is created at runtime.</remarks>
        private void mniCopy_Click(object sender, EventArgs e)
        {
            Bitmap bmp = new Bitmap(this.Width, this.Height);
            Graphics g = Graphics.FromImage(bmp);
            g.Clear(Color.White);

            // don't blt directly from the screen, but draw directly on this bitmap we've made.
            PaintEventArgs pe = new PaintEventArgs(g, ClientRectangle);
            this.OnPaint(pe);
            
            Clipboard.SetImage(bmp);
            
            bmp.Dispose();
            g.Dispose();
            pe.Dispose();
        }

        /// <summary>
        /// Allows the user to change the name of the series who is the parent menu of this menu item.
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        /// <remarks>This handler is dynamically installed when the context menu is created at runtime.</remarks>
        private void mniChangeName_Click(object sender, EventArgs e)
        {
            MenuItem item = (MenuItem) sender;
            MenuItem parent = (MenuItem) item.Parent;
            Series s = (Series) parent.Tag;

            string newName = Interaction.InputBox("Enter a new name for this series.", "Series Name", s.Name, -1, -1);
            if (newName != String.Empty)
                s.Name = newName;
            if (_showLegend)
                Invalidate();
        }

        /// <summary>
        /// Toggles on/off the points drawn for the series who is the parent menu of this menu item.
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        /// <remarks>This handler is dynamically installed when the context menu is created at runtime.</remarks>
        private void mniTogglePoints_Click(object sender, EventArgs e)
        {
            MenuItem item = (MenuItem) sender;
            MenuItem parent = (MenuItem) item.Parent;
            Series s = (Series) parent.Tag;

            s.DrawPoints = !s.DrawPoints;
            Invalidate();
        }

        /// <summary>
        /// Toggles on/off the line drawn for the series who is the parent menu of this menu item.
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        /// <remarks>This handler is dynamically installed when the context menu is created at runtime.</remarks>
        private void mniToggleLine_Click(object sender, EventArgs e)
        {
            MenuItem item = (MenuItem) sender;
            MenuItem parent = (MenuItem) item.Parent;
            Series s = (Series) parent.Tag;

            s.DrawLine = !s.DrawLine;
            Invalidate();
        }

        /// <summary>
        /// Sets the color of the points for the series who is the parent menu of this menu item. This handler
        /// invokes a ColorDialog to do so.
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        /// <remarks>This handler is dynamically installed when the context menu is created at runtime.</remarks>
        private void mniColorPoints_Click(object sender, EventArgs e)
        {
            MenuItem item = (MenuItem) sender;
            MenuItem parent = (MenuItem) item.Parent;
            Series s = (Series) parent.Tag;

            ColorDialog dlg = new ColorDialog();
            dlg.FullOpen = true;
            dlg.Color = s.PointColor;
            if (dlg.ShowDialog(this) == DialogResult.OK)
            {
                s.PointColor = dlg.Color;
                Invalidate();
            }
            dlg.Dispose();
        }

        /// <summary>
        /// Sets the color of the line for the series who is the parent menu of this menu item. This handler
        /// invokes a ColorDialog to do so.
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        /// <remarks>This handler is dynamically installed when the context menu is created at runtime.</remarks>
        private void mniColorLine_Click(object sender, EventArgs e)
        {
            MenuItem item = (MenuItem) sender;
            MenuItem parent = (MenuItem) item.Parent;
            Series s = (Series) parent.Tag;

            ColorDialog dlg = new ColorDialog();
            dlg.FullOpen = true;
            dlg.Color = s.LineColor;
            if (dlg.ShowDialog(this) == DialogResult.OK)
            {
                s.LineColor = dlg.Color;
                Invalidate();
            }
            dlg.Dispose();
        }

        /// <summary>
        /// Exports all the data for the plot at once in a CSV format.
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        private void mniExportAll_Click(object sender, EventArgs e)
        {
            string title = this.Title;
            char[] invalids = Path.GetInvalidFileNameChars();
            for (int i = 0; i < invalids.Length; i++)
                title = title.Replace(invalids[i], ' ');
            string file = String.Format("{0}\\{1}__{2}.csv", Directory.GetCurrentDirectory(), title, Environment.TickCount);

            StreamWriter writer = null;
            try
            {
                writer = new StreamWriter(file, false, Encoding.ASCII);
                writer.WriteLine("Title,Series,PointColor,LineColor,ShowPoints,ShowLine,Index,X,Y");
                foreach (Plot.Series s in _series)
                {
                    for (int i = 0; i < s.Points.Count; i++)
                    {
                        writer.WriteLine("{0},{1},{2},{3},{4},{5},{6},{7},{8}",
                            title.Replace(',', '_'),
                            s.Name.Replace(',', '_'),
                            s.PointColor.Name,
                            s.LineColor.Name,
                            s.DrawPoints ? 1 : 0,
                            s.DrawLine ? 1 : 0,
                            i,
                            s[i].X,
                            s[i].Y);
                    }
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex);
            }
            finally
            {
                if (writer != null)
                    writer.Close();
            }
            Process.Start("notepad.exe", file);
        }

        /// <summary>
        /// Exports the points for one series who is the parent menu of this menu item. The exporting is done
        /// to a text file written to the current directory. This allows a user to quickly re-plot the 
        /// series in an application like Microsoft Excel.
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        /// <remarks>This handler is dynamically installed when the context menu is created at runtime.</remarks>
        private void mniExportSeries_Click(object sender, EventArgs e)
        {
            MenuItem item = (MenuItem) sender;
            MenuItem parent = (MenuItem) item.Parent;
            Series s = (Series) parent.Tag;

            string title = this.Title;
            string name = s.Name;
            char[] invalids = Path.GetInvalidFileNameChars();
            for (int i = 0; i < invalids.Length; i++)
            {
                title = title.Replace(invalids[i], ' ');
                name = name.Replace(invalids[i], ' ');
            }
            string file = String.Format("{0}\\{1}_{2}__{3}.csv", Directory.GetCurrentDirectory(), title, name, Environment.TickCount);
            
            StreamWriter writer = null;
            try
            {
                writer = new StreamWriter(file, false, Encoding.ASCII);
                writer.WriteLine("Title,Series,PointColor,LineColor,ShowPoints,ShowLine,Index,X,Y");
                for (int i = 0; i < s.Points.Count; i++)
                {
                    writer.WriteLine("{0},{1},{2},{3},{4},{5},{6},{7},{8}",
                        title.Replace(',', '_'),
                        name.Replace(',', '_'),
                        s.PointColor.Name,
                        s.LineColor.Name,
                        s.DrawPoints ? 1 : 0,
                        s.DrawLine ? 1 : 0,
                        i,
                        s[i].X,
                        s[i].Y);
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex);
            }
            finally
            {
                if (writer != null)
                    writer.Close();
            }
            Process.Start("notepad.exe", file);
        }

        #endregion

        #region Context Menu API

        /// <summary>
        /// Programmatically triggers the functionality for the Set Plot Properties context menu item.
        /// </summary>
        public void SetPlotProperties()
        {
            mniSetPlot_Click(null, EventArgs.Empty);
        }

        /// <summary>
        /// Programmatically triggers the functionality for the Set X Axis Properties context menu item.
        /// </summary>
        public void SetXAxisProperties()
        {
            mniSetXAxis_Click(null, EventArgs.Empty);
        }

        /// <summary>
        /// Programmatically triggers the functionality for the Set Y Axis Properties context menu item.
        /// </summary>
        public void SetYAxisProperties()
        {
            mniSetYAxis_Click(null, EventArgs.Empty);
        }

        /// <summary>
        /// Programmatically triggers the functionality for the Copy Image to Clipboard context menu item.
        /// </summary>
        public void CopyImageToClipboard()
        {
            mniCopy_Click(null, EventArgs.Empty);
        }

        /// <summary>
        /// Programmatically triggers the functionality for the Export All context menu item.
        /// </summary>
        public void ExportAll()
        {
            mniExportAll_Click(null, EventArgs.Empty);
        }

        /// <summary>
        /// Programmatically triggers the functionality for the Toggle Points context menu item
        /// for a given series.
        /// </summary>
        /// <param name="name">The name of the series.</param>
        /// <returns>True if the series is found; false otherwise.</returns>
        public bool TogglePoints(string name)
        {
            bool found = false;
            for (int i = 0; i < _series.Count; i++)
            {
                if (_series[i].Name == name)
                {
                    MenuItem item = new MenuItem();
                    MenuItem parent = new MenuItem();
                    parent.MenuItems.Add(item);
                    parent.Tag = _series[i];
                    mniTogglePoints_Click(item, EventArgs.Empty);
                    found = true;
                    break;
                }
            }
            return found;
        }

        /// <summary>
        /// Programmatically triggers the functionality for the Toggle Line context menu item
        /// for a given series.
        /// </summary>
        /// <param name="name">The name of the series.</param>
        /// <returns>True if the series is found; false otherwise.</returns>
        public bool ToggleLine(string name)
        {
            bool found = false;
            for (int i = 0; i < _series.Count; i++)
            {
                if (_series[i].Name == name)
                {
                    MenuItem item = new MenuItem();
                    MenuItem parent = new MenuItem();
                    parent.MenuItems.Add(item);
                    parent.Tag = _series[i];
                    mniToggleLine_Click(item, EventArgs.Empty);
                    found = true;
                    break;
                }
            }
            return found;
        }

        /// <summary>
        /// Programmatically triggers the functionality for the Color Points context menu item
        /// for a given series.
        /// </summary>
        /// <param name="name">The name of the series.</param>
        /// <returns>True if the series is found; false otherwise.</returns>
        public bool ColorPoints(string name)
        {
            bool found = false;
            for (int i = 0; i < _series.Count; i++)
            {
                if (_series[i].Name == name)
                {
                    MenuItem item = new MenuItem();
                    MenuItem parent = new MenuItem();
                    parent.MenuItems.Add(item);
                    parent.Tag = _series[i];
                    mniColorPoints_Click(item, EventArgs.Empty);
                    found = true;
                    break;
                }
            }
            return found;
        }

        /// <summary>
        /// Programmatically triggers the functionality for the Color Line context menu item
        /// for a given series.
        /// </summary>
        /// <param name="name">The name of the series.</param>
        /// <returns>True if the series is found; false otherwise.</returns>
        public bool ColorLine(string name)
        {
            bool found = false;
            for (int i = 0; i < _series.Count; i++)
            {
                if (_series[i].Name == name)
                {
                    MenuItem item = new MenuItem();
                    MenuItem parent = new MenuItem();
                    parent.MenuItems.Add(item);
                    parent.Tag = _series[i];
                    mniColorLine_Click(item, EventArgs.Empty);
                    found = true;
                    break;
                }
            }
            return found;
        }

        /// <summary>
        /// Programmatically triggers the functionality for the Export context menu item
        /// for a given series.
        /// </summary>
        /// <param name="name">The name of the series.</param>
        /// <returns>True if the series is found; false otherwise.</returns>
        public bool Export(string name)
        {
            bool found = false;
            for (int i = 0; i < _series.Count; i++)
            {
                if (_series[i].Name == name)
                {
                    MenuItem item = new MenuItem();
                    MenuItem parent = new MenuItem();
                    parent.MenuItems.Add(item);
                    parent.Tag = _series[i];
                    mniExportSeries_Click(item, EventArgs.Empty);
                    found = true;
                    break;
                }
            }
            return found;
        }

        #endregion

        #region Data API

        /// <summary>
        /// Adds a series, which is a set of data points, to this plot control. The
        /// plot control is capable of plotting an unlimited number of series.
        /// </summary>
        /// <param name="s">The series to add.</param>
        public void AddSeries(Series s)
        {
            _series.Add(s);
            UpdateDomainRange();
        }

        /// <summary>
        /// Inserts a series at the index given.
        /// </summary>
        /// <param name="index">The 0-based index at which to insert the series.</param>
        /// <param name="s">The series to insert.</param>
        public void InsertSeries(int index, Series s)
        {
            _series.Insert(index, s);
            UpdateDomainRange();
        }

        /// <summary>
        /// Removes the series at the given index.
        /// </summary>
        /// <param name="index">The 0-based index of the series to remove.</param>
        /// <returns>The removed series, or <b>null</b> if <i>index</i> is out of range.</returns>
        public Series RemoveSeries(int index)
        {
            Series s = null;
            if (0 <= index && index < _series.Count)
            {
                s = _series[index];
                _series.RemoveAt(index);
                UpdateDomainRange();
            }
            return s;
        }

        /// <summary>
        /// Removes the series with the given name, if any. If more than one series with
        /// this name exists, then the first such series (i.e., the one with the lower index
        /// number, which was added first) is removed.
        /// </summary>
        /// <param name="name">The case-senitive name of the series to remove.</param>
        /// <returns>The removed series.</returns>
        public Series RemoveSeries(string name)
        {
            Series s = null;
            for (int i = 0; i < _series.Count; i++)
            {
                if (_series[i].Name == name)
                {
                    s = _series[i];
                    _series.RemoveAt(i);
                    UpdateDomainRange();
                    break;
                }
            }
            return s;
        }

        /// <summary>
        /// Clears all series from this plot control.
        /// </summary>
        public void ClearSeries()
        {
            _series.Clear();
            UpdateDomainRange();
        }

        /// <summary>
        /// Gets the number of series that have been added to this plot control.
        /// </summary>
        public int NumSeries
        {
            get
            {
                return _series.Count;
            }
        }

        /// <summary>
        /// Gets the total number of points in all series added to this plot control.
        /// </summary>
        public int NumPoints
        {
            get
            {
                int n = 0;
                for (int i = 0; i < _series.Count; i++)
                {
                    n += _series[i].NumPoints;
                }
                return n;
            }
        }

        /// <summary>
        /// Gets the series from the plot at the given index.
        /// </summary>
        /// <param name="index">The 0-based index of the series to get.</param>
        /// <returns>The series at the given index.</returns>
        public Series this[int index]
        {
            get
            {
                if (0 <= index && index < _series.Count)
                {
                    return _series[index];
                }
                return null;
            }
        }

        /// <summary>
        /// Gets the series from the plot at the given index.
        /// </summary>
        /// <param name="index">The 0-based index of the series to get.</param>
        /// <returns>The series at the given index.</returns>
        public Series GetSeries(int index)
        {
            return this[index];
        }

        /// <summary>
        /// Gets the series with the given name, if any. If more than one series with
        /// this name exists, then the first such series (i.e., the one with the lower index
        /// number, which was added first) is retrieved.
        /// </summary>
        /// <param name="name">The case-senitive name of the series to get.</param>
        /// <returns>The series with the given name, or null if none.</returns>
        public Series GetSeries(string name)
        {
            Series s = null;
            for (int i = 0; i < _series.Count; i++)
            {
                if (_series[i].Name == name)
                {
                    s = _series[i];
                    break;
                }
            }
            return s;
        }

        /// <summary>
        /// This method updates the plot control's overall domain (x-values) and
        /// range (y-values). This information is used to set the automatic scaling and labels
        /// for the X- and Y-axes.
        /// </summary>
        /// <remarks>This method should be called after any series data is changed within this
        /// plot control. This is done automatically when a new series is added or an existing
        /// series is removed. However, if points are added to a series instance after the
        /// series has already been added to the plot control, then the client who added these
        /// points must call this method in order for the plot control to update its autoscaled
        /// axes.</remarks>
        public void UpdateDomainRange()
        {
            if (_series.Count == 0)
            {
                _domainRange = RectangleR.Empty;
            }
            else
            {
                double minX = double.MaxValue;
                double maxX = double.MinValue;
                double minY = double.MaxValue;
                double maxY = double.MinValue;

                foreach (Series s in _series)
                {
                    RectangleR r = s.DomainRange;
                    if (r.Left < minX)
                        minX = r.Left;
                    if (r.Right > maxX)
                        maxX = r.Right;
                    if (r.Top < minY)
                        minY = r.Top;
                    if (r.Bottom > maxY)
                        maxY = r.Bottom;
                }

                _domainRange = new RectangleR(minX, minY, maxX - minX, maxY - minY);
            }
        }

        /// <summary>
        /// Gets the minimum value currently shown on the x-axis.
        /// </summary>
        public double XMin
        {
            get
            {
                double min = double.MaxValue;
                for (int i = 0; i < _series.Count; i++)
                {
                    for (int j = 0; j < _series[i].NumPoints; j++)
                    {
                        if (_series[i][j].X < min)
                            min = _series[i][j].X;
                    }
                }
                return min;
            }
        }

        /// <summary>
        /// Gets the maximum value currently shown on the x-axis.
        /// </summary>
        public double XMax
        {
            get
            {
                double max = double.MinValue;
                for (int i = 0; i < _series.Count; i++)
                {
                    for (int j = 0; j < _series[i].NumPoints; j++)
                    {
                        if (_series[i][j].X > max)
                            max = _series[i][j].X;
                    }
                }
                return max;
            }
        }

        /// <summary>
        /// Gets the minimum value currently shown on the y-axis.
        /// </summary>
        public double YMin
        {
            get
            {
                double min = double.MaxValue;
                for (int i = 0; i < _series.Count; i++)
                {
                    for (int j = 0; j < _series[i].NumPoints; j++)
                    {
                        if (_series[i][j].Y < min)
                            min = _series[i][j].Y;
                    }
                }
                return min;
            }
        }

        /// <summary>
        /// Gets the maximum value currently shown on the y-axis.
        /// </summary>
        public double YMax
        {
            get
            {
                double max = double.MinValue;
                for (int i = 0; i < _series.Count; i++)
                {
                    for (int j = 0; j < _series[i].NumPoints; j++)
                    {
                        if (_series[i][j].Y > max)
                            max = _series[i][j].Y;
                    }
                }
                return max;
            }
        }

        #endregion

        #region Protected and Private Methods

        /// <summary>
        /// The resize method for this control. It keeps the <i>_plotRect</i>, the area in which the 
        /// actual plot is drawn, sized proportional to the overall size of the plot control.
        /// </summary>
        /// <param name="e">The arguments for this event.</param>
        protected override void OnResize(EventArgs e)
        {
            base.OnResize(e);
            _plotRect = new RectangleR(
                _plotMargin.Left,
                _plotMargin.Top,
                this.Width - _plotMargin.Left - _plotMargin.Right,
                this.Height - _plotMargin.Top - _plotMargin.Bottom
             );
            Invalidate();
        }

        /// <summary>
        /// The paint method for this control. Different aspects of the plot are drawn in turn according
        /// to the parameters of the plot control and the domain and range of the data series within it.
        /// </summary>
        /// <param name="pe">The arguments for this event.</param>
        protected override void OnPaint(PaintEventArgs pe)
        {
            base.OnPaint(pe);
            RectangleF grect = (RectangleF) _plotRect; // cast for drawing

            //
            // first draw the axes -- the left and bottom sides of the drawing rect
            //
            pe.Graphics.DrawLine(Pens.Black, grect.Left, grect.Bottom, grect.Right, grect.Bottom);
            pe.Graphics.DrawLine(Pens.Black, grect.Left, grect.Bottom, grect.Left, grect.Top);

            //
            // draw the tick marks and labels along the X-axis
            //
            for (int i = 0; i <= _xTicks; i++)
            {
                float px = grect.Left + i * (grect.Width / _xTicks);
                pe.Graphics.DrawLine(Pens.Black, px, grect.Bottom - 5f, px, grect.Bottom + 5f);
                if (_xUseLabels)
                {
                    if (_xLabels != null && i < _xLabels.Length)
                    {
                        SizeF sz = pe.Graphics.MeasureString(_xLabels[i], this.Font);
                        pe.Graphics.DrawString(_xLabels[i], this.Font, Brushes.Black, px - sz.Width / 2f, grect.Bottom + sz.Height / 2f);
                    }
                }
                else if (_series.Count > 0)
                {
                    double x = Math.Round(_domainRange.Left + i * (_domainRange.Width / _xTicks), _xDecimals); // make label
                    SizeF sz = pe.Graphics.MeasureString(x.ToString(), this.Font);
                    pe.Graphics.DrawString(x.ToString(), this.Font, Brushes.Black, px - sz.Width / 2f, grect.Bottom + sz.Height / 2f);
                }
            }

            //
            // draw the tick marks and labels along the Y-axis
            //
            for (int i = 0; i <= _yTicks; i++)
            {
                float py = grect.Bottom - i * (grect.Height / _yTicks);
                pe.Graphics.DrawLine(Pens.Black, grect.Left - 5f, py, grect.Left + 5f, py);
                if (_yUseLabels)
                {
                    if (_yLabels != null && i < _yLabels.Length)
                    {
                        SizeF sz = pe.Graphics.MeasureString(_yLabels[i], this.Font);
                        pe.Graphics.DrawString(_yLabels[i], this.Font, Brushes.Black, grect.Left - 5f - sz.Width, py - sz.Height / 2f);
                    }
                }
                else if (_series.Count > 0)
                {
                    double y = Math.Round(_domainRange.Top + i * (_domainRange.Height / _yTicks), _yDecimals); // make label
                    SizeF sz = pe.Graphics.MeasureString(y.ToString(), this.Font);
                    pe.Graphics.DrawString(y.ToString(), this.Font, Brushes.Black, grect.Left - 5f - sz.Width, py - sz.Height / 2f);
                }
            }

            //
            // draw the plot title
            //
            Font fTitle = new Font(this.Font, FontStyle.Bold);
            SizeF szTitle = pe.Graphics.MeasureString(this.Text, fTitle);
            pe.Graphics.DrawString(this.Text, fTitle, Brushes.Black, this.Width / 2f - szTitle.Width / 2f, 2f);
            fTitle.Dispose();

            //
            // draw X-axis label
            //
            Font fX = new Font(this.Font, FontStyle.Italic);
            SizeF szX = pe.Graphics.MeasureString(_xName, fX);
            pe.Graphics.DrawString(_xName, fX, Brushes.Black, this.Width / 2f - szX.Width / 2f, this.Height - szX.Height - 7f);
            fX.Dispose();

            //
            // draw Y-axis label
            //
            Font fY = new Font(this.Font, FontStyle.Italic);
            SizeF szY = pe.Graphics.MeasureString(_yName, fY);
            pe.Graphics.DrawString(_yName, fY, Brushes.Black, 1f, 2f);
            fY.Dispose();

            //
            // now draw the series' lines (added first is on top)
            //
            for (int i = _series.Count - 1; i >= 0; i--)
            {
                Series s = _series[i];
                if (s.DrawLine)
                {
                    Pen pen = new Pen(s.LineColor, 2f);
                    for (int j = 1; j < s.NumPoints; j++)
                    {
                        PointF p = Logical2Device(s[j]);
                        PointF p0 = Logical2Device(s[j - 1]);
                        pe.Graphics.DrawLine(pen, p0, p);
                    }
                    pen.Dispose();
                }
            }

            //
            // now draw the series' points (atop the lines)
            //
            for (int i = _series.Count - 1; i >= 0; i--)
            {
                Series s = _series[i];
                if (s.DrawPoints)
                {
                    Brush brush = new SolidBrush(s.PointColor);
                    for (int j = 0; j < s.NumPoints; j++)
                    {
                        PointF p = Logical2Device(s[j]);
                        pe.Graphics.FillEllipse(brush, p.X - 3f, p.Y - 3f, 6f, 6f);
                    }
                    brush.Dispose();
                }
            }

            //
            // now draw the legend
            //
            if (_showLegend)
            {
                SizeF szMax = SizeF.Empty;
                foreach (Series s in _series) // measure the widest series label
                {
                    SizeF sz = pe.Graphics.MeasureString(s.Name, this.Font);
                    if (sz.Width > szMax.Width)
                        szMax = sz;
                }
                for (int i = 0; i < _series.Count; i++)
                {
                    Series s = _series[i];
                    Brush brush = new SolidBrush(s.PointColor);
                    Pen pen = new Pen(s.LineColor, 2f);

                    SizeF sz = pe.Graphics.MeasureString(s.Name, this.Font);
                    pe.Graphics.DrawString(s.Name, 
                        this.Font, 
                        Brushes.Black, 
                        this.Width - sz.Width,                          // series' name x
                        sz.Height * i + 2f);                            // series' name y

                    if (s.DrawLine)
                    {
                        pe.Graphics.DrawLine(pen, 
                            this.Width - szMax.Width - 16f,             // line x starting point
                            sz.Height * i + sz.Height * 0.5f + 1f,      // line y starting point
                            this.Width - szMax.Width,                   // line x ending point
                            sz.Height * i + sz.Height * 0.5f + 1f);     // line y ending point
                    }
                    if (s.DrawPoints)
                    {
                        pe.Graphics.FillEllipse(brush, 
                            this.Width - szMax.Width - 11f,             // left edge of ellipse
                            sz.Height * i + sz.Height * 0.5f - 3f,      // top edge of ellipse
                            6f,                                         // width
                            6f);                                        // height
                    }

                    pen.Dispose();
                    brush.Dispose();
                }
            }
        }

		/// <summary>
		/// Usually, clients should add data series to this Plot and allow the Plot to draw its
		/// data. However, there may be rare instances where clients want to handle the Paint
		/// event for this Plot control and do some painting of their own. This method enables
		/// a logical point to be displayed with the proper device plotting coordinates.
		/// </summary>
		/// <param name="point">The data point to map.</param>
		/// <returns>The equivalent plotting point to display.</returns>
		/// <remarks>
		/// Besides simply mapping the point mathematically, care must be taken to flip the Y-axis 
		/// since computer coordinates proceed downward for increasing +y values, but traditional 
		/// top-right-quadrant plot coordinates proceed UPWARD for +y.
		/// </remarks>
		public PointF Logical2Device(PointR point)
        {
            // plotting will fail if there is a zero-width
            // or zero-height domain (x) or range (y) of
            // the data. we don't want to change the actual
            // domain range data, but we do want to affect
            // how points are chosen for display, so we
            // sanity-check this here.
            RectangleR dr = _domainRange;
            if (dr.Width == 0.0)
                dr.Width = 1.0;
            if (dr.Height == 0.0)
                dr.Height = 1.0;

            // map the point from the domain-range to the plot display rect
            PointR gP = GeotrigEx.MapPoint(point, dr, _plotRect);
            gP.Y = _plotRect.Bottom - (gP.Y - _plotRect.Top); // flip Y-axis for drawing
            return (PointF) gP; // cast for drawing
        }

        #endregion

    }
}
