﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Media;

namespace WobbrockLib.Controls
{
    /// <summary>
    /// Allows for the creation of a custom message box that utilizes the same icons and sounds
    /// as the built-in Windows MessageBox class. In addition, the custom message box allows
    /// the caller to position it on the screen and to avoid playing the usual accompanying sounds.
    /// Up to four lines of text can be shown in the message for the box.
    /// </summary>
    [ToolboxBitmap(typeof(System.Windows.Forms.MessageBox))]
    public partial class PositionableMessageBox : Form
    {
        private Point _pos;
        private bool _centered;
        private MessageBoxButtons _buttons;
        private MessageBoxIcon _icon;
        private bool _playSound;

        /// <summary>
        /// This constructor is protected. Use the Show method to display a PositionableMessageBox.
        /// </summary>
        /// <param name="position">The screen coordinates of where this message box should be shown. Whether it signifies the center point or top-left point is governed by the <i>centered</i> parameter.</param>
        /// <param name="centered">If true, <i>location</i> indicates the center point of the message box. If false, <i>location</i> indicates the top-left point of the message box.</param>
        /// <param name="text">The text to show on the message.</param>
        /// <param name="caption">The caption to show in the message box's title bar.</param>
        /// <param name="buttons">MessageBoxButtons.OK, MessageBoxButtons.OKCancel, and MessageBoxButtons.YesNo are currently supported.</param>
        /// <param name="icon">The MessageBoxIcon that determines what icon, if any, is shown, and what sound is played.</param>
        /// <param name="playSound">If true, the sound accompanying <i>icon</i> is played. If false, no sound will be played.</param>
        protected PositionableMessageBox(Point position, bool centered, string text, string caption, MessageBoxButtons buttons, MessageBoxIcon icon, bool playSound)
        {
            InitializeComponent();

            this.Text = caption; // set the window title now
            lblMsg.Text = text; // label will autosize
            this.Width = Math.Max(this.Width, picIcon.Width / 2 + lblMsg.Right); // then size the form to show label
            this.Height = Math.Max(this.Height, lblMsg.Bottom + pnlShading.Bottom + 10); // then size the form to show label

            // these we'll handle in the Form's load event after sizing and dimensions are set
            _pos = position;
            _centered = centered;
            _buttons = buttons;
            _icon = icon;
            _playSound = playSound;
        }

        /// <summary>
        /// Displays a modal custom message box.
        /// </summary>
        /// <param name="owner">The owner of this form. Can be null.</param>
        /// <param name="position">The screen coordinates of where this message box should be shown. Whether it signifies the center point or top-left point is governed by the <i>centered</i> parameter.</param>
        /// <param name="centered">If true, <i>location</i> indicates the center point of the message box. If false, <i>location</i> indicates the top-left point of the message box.</param>
        /// <param name="text">The text to show on the message.</param>
        /// <param name="caption">The caption to show in the message box's title bar.</param>
        /// <param name="buttons">MessageBoxButtons.OK, MessageBoxButtons.OKCancel, and MessageBoxButtons.YesNo are currently supported.</param>
        /// <param name="icon">The MessageBoxIcon that determines what icon, if any, is shown, and what sound is played.</param>
        /// <param name="playSound">If true, the sound accompanying <i>icon</i> is played. If false, no sound will be played.</param>
        /// <returns>A DialogResult indicating whether OK or Cancel, if available, was selected.</returns>
        /// <see cref="http://msdn.microsoft.com/en-us/library/ms645505(VS.85).aspx">http://msdn.microsoft.com/en-us/library/ms645505(VS.85).aspx</see>
        public static DialogResult Show(IWin32Window owner, Point position, bool centered, string text, string caption, MessageBoxButtons buttons, MessageBoxIcon icon, bool playSound)
        {
            PositionableMessageBox This = new PositionableMessageBox(position, centered, text, caption, buttons, icon, playSound);
            return This.ShowDialog(owner); // blocking modal
        }

        /// <summary>
        /// Private event handler for loading the message box.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void PositionableMessageBox_Load(object sender, EventArgs e)
        {
			// set our unadjusted form's ideal location first
            this.Left = _centered ? _pos.X - this.Width / 2 : _pos.X;
            this.Top = _centered ? _pos.Y - this.Height / 2 : _pos.Y;

			// then constrain our message box to appear entirely on-screen
			Point pt = this.Location;
			Rectangle bounds = Screen.GetWorkingArea(pt);
			this.Left = Math.Min(pt.X + this.Width, bounds.Right) - this.Width;
			this.Top = Math.Min(pt.Y + this.Height, bounds.Bottom) - this.Height;
            
            // if only the OK button is to be shown, move it atop the hidden Cancel button
            if (_buttons == MessageBoxButtons.OK)
            {
                cmdCancel.Visible = false;
                cmdCancel.Enabled = false;
                cmdCancel.DialogResult = DialogResult.None;
				cmdOK.Left = cmdCancel.Left;
            }
            else if (_buttons == MessageBoxButtons.YesNo)
            {
                cmdOK.Text = "Yes";
                cmdCancel.Text = "No";
                cmdOK.DialogResult = DialogResult.Yes;
                cmdCancel.DialogResult = DialogResult.No;
            }

            // set the icon and possibly play the sound
            DoIconAndSound();
        }

        /// <summary>
        /// Private helper method for setting the icons and playing the associated sound.
        /// </summary>
        private void DoIconAndSound()
        {
            if (_icon == MessageBoxIcon.Hand || _icon == MessageBoxIcon.Stop || _icon == MessageBoxIcon.Error)
            {
                picIcon.Image = imglstIcons.Images[0];
                if (_playSound)
                    SystemSounds.Hand.Play();
            }
            else if (_icon == MessageBoxIcon.Question)
            {
                picIcon.Image = imglstIcons.Images[1];
                if (_playSound)
                    SystemSounds.Question.Play();
            }
            else if (_icon == MessageBoxIcon.Exclamation || _icon == MessageBoxIcon.Warning)
            {
                picIcon.Image = imglstIcons.Images[2];
                if (_playSound)
                    SystemSounds.Exclamation.Play();
            }
            else if (_icon == MessageBoxIcon.Asterisk || _icon == MessageBoxIcon.Information)
            {
                picIcon.Image = imglstIcons.Images[3];
                if (_playSound)
                    SystemSounds.Asterisk.Play();
            }
            else if (_playSound)
			{
				SystemSounds.Beep.Play();
            }
        }
    }
}
