﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using WobbrockLib.Extensions;
using WobbrockLib.Optimization;

namespace WobbrockLib.Data
{
    //#pragma warning disable CS0693
    public class Graph
    {
        #region Graph.Connection

        /// <summary>
        /// Inner class representing a directed connection between nodes in a graph.
        /// </summary>
        public class Connection
        {
            private Node _from;
            private Node _to;
            private double _weight;

            /// <summary>
            /// Constructs an unweighted connection from a given node to another node in a graph.
            /// </summary>
            /// <param name="from"></param>
            /// <param name="to"></param>
            public Connection(Node from, Node to)
                : this(from, to, 0.0)
            {
                // do nothing
            }

            /// <summary>
            /// Constructs a weighted connection from a given node to another node in a graph.
            /// </summary>
            /// <param name="from"></param>
            /// <param name="to"></param>
            /// <param name="weight"></param>
            public Connection(Node from, Node to, double weight)
            {
                _from = from;
                _to = to;
                _weight = weight;
            }

            /// <summary>
            /// Gets or sets the node from which this connection departs.
            /// </summary>
            public Node From
            {
                get { return _from; }
                set { _from = value; }
            }

            /// <summary>
            /// Gets or sets the node into which this connection arrives.
            /// </summary>
            public Node To
            {
                get { return _to; }
                set { _to = value; }
            }

            /// <summary>
            /// The weight, or cost, of the connection between nodes.
            /// </summary>
            public double Weight
            {
                get { return _weight; }
                set { _weight = value; }
            }

            /// <summary>
            /// True if the weight of this connection is non-zero; false otherwise.
            /// </summary>
            public bool IsWeighted
            {
                get { return _weight != 0.0; }
            }

            /// <summary>
            /// True if the connection is from and to the same node; false otherwise.
            /// </summary>
            public bool IsCircular
            {
                get { return _from == _to; }
            }

            /// <summary>
            /// Gets a string representation of this connection between nodes.
            /// </summary>
            /// <returns></returns>
            public override string ToString()
            {
                return String.Format("From: {0} -> To: {1}", From.ID, To.ID);
            }

            /// <summary>
            /// Tests whether this connection is equal to the given object.
            /// </summary>
            /// <param name="obj"></param>
            /// <returns></returns>
            public override bool Equals(object obj)
            {
                if (!(obj is Connection))
                    return false;
                Connection o = (Connection) obj;
                return (this.From == o.From 
                    && this.To == o.To 
                    && this.Weight == o.Weight);
            }

            /// <summary>
            /// Gets a hash code for this connection.
            /// </summary>
            /// <returns></returns>
            public override int GetHashCode()
            {
                return base.GetHashCode();
            }
        }

        #endregion

        #region Graph.Node

        /// <summary>
        /// Inner class representing a data node in a graph. A data node can optionally 
        /// have an ID and a data object associated with it. It can also have connections
        /// to other nodes in the graph.
        /// </summary>
        public class Node
        {
            private int _id;
            private object _data;
            private List<Connection> _connections;

            #region Constructor

            /// <summary>
            /// Construct a node with the given ID ad null data. A node's ID should
            /// be unique within a data graph.
            /// </summary>
            /// <param name="id"></param>
            public Node(int id)
                : this(id, null)
            {
                // do nothing
            }

            /// <summary>
            /// Construct a node with the given ID and data object. A node's ID
            /// should be unique within a data graph.
            /// </summary>
            /// <param name="id"></param>
            public Node(int id, object data)
            {
                _id = id;
                _data = data;
                _connections = new List<Connection>();
            }

            #endregion

            #region Properties

            /// <summary>
            /// Gets the ID of this node. The ID is assigned when the node is constructed.
            /// A node's ID should be unique within its graph.
            /// </summary>
            public int ID
            {
                get { return _id; }
            }

            /// <summary>
            /// Gets or sets the data object associated with this node.
            /// </summary>
            public object Data
            {
                get { return _data; }
                set { _data = value; }
            }

            /// <summary>
            /// Gets a list of the connections coming into and going out from this node.
            /// Connections should be added and removed with their respective methods,
            /// not by directly modifying this list.
            /// </summary>
            public List<Connection> Connections
            {
                get { return _connections; }
            }

            /// <summary>
            /// Gets the number of connections coming out of or into this node.
            /// </summary>
            /// <returns></returns>
            public int Count
            {
                get { return (_connections != null) ? _connections.Count : 0; }
            }

            #endregion

            #region Testing Connections

            /// <summary>
            /// Tests whether this node has an outgoing connection to the given node.
            /// </summary>
            /// <param name="node">The node to test.</param>
            /// <returns>True if there is a connection from this node into the given node; 
            /// false otherwise.</returns>
            public bool HasConnectionTo(Node node)
            {
                for (int i = 0; i < _connections.Count; i++)
                {
                    if (_connections[i].From == this && _connections[i].To == node)
                        return true;
                }
                return false;
            }

            /// <summary>
            /// Tests whether this node has an outgoing connection to the given node ID.
            /// </summary>
            /// <param name="id">The ID of the node to test.</param>
            /// <returns>True if there is a connection from this node into a node with the
            /// given ID; false otherwise.</returns>
            public bool HasConnectionTo(int id)
            {
                for (int i = 0; i < _connections.Count; i++)
                {
                    if (_connections[i].From == this && _connections[i].To.ID == id)
                        return true;
                }
                return false;
            }

            /// <summary>
            /// Tests whether this node has an incoming connection from the given node.
            /// </summary>
            /// <param name="node">The node to test.</param>
            /// <returns>True if there is a connection from the given node into this node; 
            /// false otherwise.</returns>
            public bool HasConnectionFrom(Node node)
            {
                for (int i = 0; i < _connections.Count; i++)
                {
                    if (_connections[i].From == node && _connections[i].To == this)
                        return true;
                }
                return false;
            }

            /// <summary>
            /// Tests whether this node has an incoming connection from the given node ID.
            /// </summary>
            /// <param name="id">The ID of the node to test.</param>
            /// <returns>True if there is a connection from a node with the given ID into
            /// this node; false otherwise.</returns>
            public bool HasConnectionFrom(int id)
            {
                for (int i = 0; i < _connections.Count; i++)
                {
                    if (_connections[i].From.ID == id && _connections[i].To == this)
                        return true;
                }
                return false;
            }

            /// <summary>
            /// Tests whether this node has either an outgoing connection to or an incoming connection from the given node.
            /// </summary>
            /// <param name="node">The node to test.</param>
            /// <returns>True if there is either a connection from this node into the given node, or from the given node
            /// into this node; false otherwise.</returns>
            public bool HasConnection(Node node)
            {
                return HasConnectionTo(node) || HasConnectionFrom(node);
            }

            /// <summary>
            /// Tests whether this node has either an outgoing connection to or an incoming connection from the given node ID.
            /// </summary>
            /// <param name="id">The ID of the node to test.</param>
            /// <returns>True if there is either a connection from this node into a node with the given ID, or from a node with
            /// the given ID into this node; false otherwise.</returns>
            public bool HasConnection(int id)
            {
                return HasConnectionTo(id) || HasConnectionFrom(id);
            }

            #endregion

            #region Getting Connections

            /// <summary>
            /// Gets the first outgoing connection, if any, from this node to the given node.
            /// </summary>
            /// <param name="node">The node to test.</param>
            /// <returns>The connection to the given node, if one exists.</returns>
            public Connection GetConnectionTo(Node node)
            {
                for (int i = 0; i < _connections.Count; i++)
                {
                    if (_connections[i].From == this && _connections[i].To == node)
                        return _connections[i];
                }
                return null;
            }

            /// <summary>
            /// Gets the first outgoing connection, if any, from this node to a node with the given ID.
            /// </summary>
            /// <param name="id">The ID of the node to test.</param>
            /// <returns>The connection to a node with the given ID, if one exists.</returns>
            public Connection GetConnectionTo(int id)
            {
                for (int i = 0; i < _connections.Count; i++)
                {
                    if (_connections[i].From == this && _connections[i].To.ID == id)
                        return _connections[i];
                }
                return null;
            }

            /// <summary>
            /// Gets the first incoming connection, if any, from the given node into this node.
            /// </summary>
            /// <param name="node">The node to test.</param>
            /// <returns>The connection from the given node, if one exists.</returns>
            public Connection GetConnectionFrom(Node node)
            {
                for (int i = 0; i < _connections.Count; i++)
                {
                    if (_connections[i].From == node && _connections[i].To == this)
                        return _connections[i];
                }
                return null;
            }

            /// <summary>
            /// Gets the first incoming connection, if any, from a node with the given ID into this node.
            /// </summary>
            /// <param name="id">The ID of the node to test.</param>
            /// <returns>The connection from a node with the given ID, if one exists.</returns>
            public Connection GetConnectionFrom(int id)
            {
                for (int i = 0; i < _connections.Count; i++)
                {
                    if (_connections[i].From.ID == id && _connections[i].To == this)
                        return _connections[i];
                }
                return null;
            }

            /// <summary>
            /// Gets all outgoing connections from this node to any nodes.
            /// </summary>
            /// <param name="wantCircular">If true, circular connections will be included.</param>
            /// <returns>A list of all outgoing connections from this node.</returns>
            public List<Connection> GetOutgoingConnections(bool wantCircular)
            {
                List<Connection> cnxns = new List<Connection>(_connections.Count);
                for (int i = 0; i < _connections.Count; i++)
                {
                    if (_connections[i].From == this && (wantCircular || _connections[i].To != this))
                        cnxns.Add(_connections[i]);
                }
                return cnxns;
            }

            /// <summary>
            /// Gets all incoming connections from any nodes into this node.
            /// </summary>
            /// <param name="wantCircular">If true, circular connections will be included.</param>
            /// <returns>A list of all incoming connections to this node.</returns>
            public List<Connection> GetIncomingConnections(bool wantCircular)
            {
                List<Connection> cnxns = new List<Connection>(_connections.Count);
                for (int i = 0; i < _connections.Count; i++)
                {
                    if (_connections[i].To == this && (wantCircular || _connections[i].From != this))
                        cnxns.Add(_connections[i]);
                }
                return cnxns;
            }

            /// <summary>
            /// Gets all connections for this node, whether outgoing (to) or incoming (from).
            /// </summary>
            /// <param name="wantCircular">If true, circular connections will be included.</param>
            /// <returns>A list of all outgoing and incoming connections for this node.</returns>
            public List<Connection> GetAllConnections(bool wantCircular)
            {
                List<Connection> outCnxns = GetOutgoingConnections(wantCircular);
                List<Connection> inCnxns = GetIncomingConnections(wantCircular);
                List<Connection> cnxns = new List<Connection>(outCnxns.Count + inCnxns.Count);
                cnxns.AddRange(outCnxns);
                cnxns.AddRange(inCnxns);
                return cnxns;
            }

            /// <summary>
            /// Gets the connection at the 0-based index given. This indexer is useful for iterating
            /// over all the connections associated with a given node.
            /// </summary>
            /// <param name="index">The 0-based index of the connection to get.</param>
            /// <returns>The connection at the given index.</returns>
            public Connection this[int index]
            {
                get
                {
                    if (0 <= index && index < _connections.Count)
                    {
                        return _connections[index];
                    }
                    else throw new IndexOutOfRangeException("Index out of range in Graph.Node indexer.");
                }
            }

            #endregion

            #region Connecting & Disconnecting

            /// <summary>
            /// Connects from this node to the given node with an unweighted connection.
            /// </summary>
            /// <param name="node">The node to which to connect this node.</param>
            /// <returns>The connection instance formed by this connection.</returns>
            /// <remarks>If the connection already exists, the connection will be updated.</remarks>
            public Connection ConnectTo(Node node)
            {
                return ConnectTo(node, 0.0);
            }

            /// <summary>
            /// Connects from this node to the given node with a weighted connection.
            /// </summary>
            /// <param name="node">The node to which to connect this node.</param>
            /// <param name="weight">The weight to assign this connection.</param>
            /// <returns>The connection instance formed by this connection.</returns>
            /// <remarks>If the connection already exists, the connection will be updated.</remarks>
            public Connection ConnectTo(Node node, double weight)
            {
                Connection cnxn = null;
                if (!this.HasConnectionTo(node))
                {
                    cnxn = new Connection(this, node, weight);
                    _connections.Add(cnxn);
                }
                else // already had the connection, so update it
                {
                    cnxn = this.GetConnectionTo(node);
                    cnxn.From = this;
                    cnxn.To = node;
                    cnxn.Weight = weight;
                }

                // now update the "other end" of the connection
                if (!node.HasConnectionFrom(this))
                {
                    Connection nCnxn = new Connection(this, node, weight);
                    node.Connections.Add(nCnxn);
                }
                else // already had the connection, so update it
                {
                    Connection nCnxn = node.GetConnectionFrom(this);
                    nCnxn.From = this;
                    nCnxn.To = node;
                    nCnxn.Weight = weight;
                }
                return cnxn;
            }

            /// <summary>
            /// Connects from the given node to this node with an unweighted connection.
            /// </summary>
            /// <param name="node">The node from which to connect this node.</param>
            /// <returns>The connection instance formed by this connection.</returns>
            /// <remarks>If the connection already exists, the connection will be updated.</remarks>
            public Connection ConnectFrom(Node node)
            {
                return ConnectFrom(node, 0.0);
            }

            /// <summary>
            /// Connects from the given node to this node with a weighted connection.
            /// </summary>
            /// <param name="node">The node from which to connect this node.</param>
            /// <param name="weight">The weight to assign this connection.</param>
            /// <returns>The connection instance formed by this connection.</returns>
            /// <remarks>If the connection already exists, the connection will be updated.</remarks>
            public Connection ConnectFrom(Node node, double weight)
            {
                Connection cnxn = null;
                if (!this.HasConnectionFrom(node))
                {
                    cnxn = new Connection(node, this, weight);
                    _connections.Add(cnxn);
                }
                else // already had the connection, so update it
                {
                    cnxn = this.GetConnectionFrom(node);
                    cnxn.From = node;
                    cnxn.To = this;
                    cnxn.Weight = weight;
                }

                // now update the other end of the connection
                if (!node.HasConnectionTo(this))
                {
                    Connection nCnxn = new Connection(node, this, weight);
                    node.Connections.Add(nCnxn);
                }
                else // already had the connection, so update it
                {
                    Connection nCnxn = node.GetConnectionTo(this);
                    nCnxn.From = node;
                    nCnxn.To = this;
                    nCnxn.Weight = weight;
                }

                return cnxn;
            }

            /// <summary>
            /// Break the connection from this node to the given node. A unidirectional conection is assumed,
            /// meaning that the connection to break must be from this node into the given node.
            /// </summary>
            /// <param name="node">The node with which to break the connection.</param>
            /// <returns>The number of connections removed: 0 or 1.</returns>
            /// <remarks>Zero severed connections will result in the case that there was no connection
            /// to begin with. One severed connection will result if there was a single directional connection 
            /// from this node into the given node that was severed.</remarks>
            public int Disconnect(Node node)
            {
                return Disconnect(node, false);
            }

            /// <summary>
            /// Break the connection from this node to the given node. A unidirectional or bidirectional
            /// connection can be specified. In the case of the latter, it does not matter whether the
            /// given node is connected 'to' from this node or 'from' into this node.
            /// </summary>
            /// <param name="node">The node with which to break the connection.</param>
            /// <param name="bidirectional">If false, the connection must move from this node to the given
            /// node (outgoing). If true, the connection can also move from the given node into this node
            /// (incoming).</param>
            /// <returns>The number of connections removed: 0, 1, or 2.</returns>
            /// <remarks>Zero severed connections will result in the case that there was no connection
            /// to begin with. One severed connection will result if there was a single connection between 
            /// the nodes that was severed. Two severed connections will result if <i>bidirectional</i> is 
            /// true and this node and the given node each had outgoing connections to each other (and therefore, 
            /// incoming connections from each other).</remarks>
            public int Disconnect(Node node, bool bidirectional)
            {
                // remove outgoing and incoming connections from this node
                int removed = 0;
                for (int i = _connections.Count - 1; i >= 0; i--)
                {
                    if ((_connections[i].From == this && _connections[i].To == node) // outgoing
                        || (bidirectional && _connections[i].From == node && _connections[i].To == this)) // incoming
                    {
                        _connections.RemoveAt(i);
                        removed++;
                    }
                }

                // remove the "other ends" of all connections
                for (int i = node.Connections.Count - 1; i >= 0; i--)
                {
                    if ((node.Connections[i].From == this && node.Connections[i].To == node) // incoming for node
                        || (bidirectional && node.Connections[i].From == node && node.Connections[i].To == this)) // outgoing for node
                    {
                        node.Connections.RemoveAt(i);
                    }
                }
                return removed;
            }

            /// <summary>
            /// Disconnect this node from all other nodes.
            /// </summary>
            /// <returns>The total number of connections removed.</returns>
            public int DisconnectAll()
            {
                int removed = 0;
                for (int i = _connections.Count - 1; i >= 0; i--)
                {
                    if (_connections[i].From == this)
                    {
                        int r = Disconnect(_connections[i].To); // all outgoing connections
                        if (r == 2)
                            i--; // adjust
                        removed += r;
                    }
                }
                for (int i = _connections.Count - 1; i >= 0; i--)
                {
                    if (_connections[i].To == this)
                    {
                        int r = Disconnect(_connections[i].From); // all incoming connections
                        if (r == 2)
                            i--; // adjust
                        removed += r;
                    }
                }
                return removed;
            }

            #endregion

            #region Overrides

            /// <summary>
            /// Gets a string representation of this node.
            /// </summary>
            /// <returns></returns>
            public override string ToString()
            {
                return String.Format("ID: {0}; Data: {1}", this.ID, this.Data?.ToString());
            }

            /// <summary>
            /// Tests whether this node is equal to the given object. The comparison is a deep
            /// comparison, not just whether the references are the same.
            /// </summary>
            /// <param name="obj"></param>
            /// <returns>True if the nodes have the same ID, same data, and all the same connections;
            /// false otherwise.</returns>
            public override bool Equals(object obj)
            {
                if (!(obj is Node))
                    return false;
                Node o = (Node) obj;
                return base.Equals(o);
            }

            /// <summary>
            /// Gets a hash code for this node.
            /// </summary>
            /// <returns></returns>
            public override int GetHashCode()
            {
                return base.GetHashCode();
            }

            #endregion
        }

        #endregion

        #region Graph.NodeAffinityCallback

        /// <summary>
        /// Callback function used in the clustering algorithm for the data graph to test whether 
        /// two nodes have affinity for each other or not. The determination of affinity is left 
        /// to the caller and can be based on edge weights or other properties of the nodes.
        /// </summary>
        /// <param name="cnxn">The connection to test for affinity between two nodes.</param>
        /// <returns>True if affinity exists between the nodes; false otherwise.</returns>
        public delegate bool NodeAffinityCallback(Graph.Connection cnxn);

        #endregion

        #region Graph

        #region Fields

        private List<Node> _nodes;

        #endregion

        #region Constructor 

        /// <summary>
        /// Constructs a new Graph instance.
        /// </summary>
        /// <remarks>Graphs have nodes that contain a numeric ID, a data object, 
        /// and a set of connections to other nodes. </remarks>
        public Graph()
        {
            _nodes = new List<Node>();
        }

        #endregion

        #region Adding & Removing Nodes

        /// <summary>
        /// Gets the number of nodes in this data graph.
        /// </summary>
        /// <returns></returns>
        public int Count
        {
            get { return (_nodes != null) ? _nodes.Count : 0; }
        }

        /// <summary>
        /// Adds a node to this data graph. The node ID must be unique within the
        /// data graph's nodes thus far.
        /// </summary>
        /// <param name="node">The node to add</param>
        /// <returns>True if successful; false otherwise.</returns>
        public bool AddNode(Node node)
        {
            bool found = false;
            for (int i = 0; i < _nodes.Count; i++)
            {
                if (_nodes[i].ID == node.ID)
                {
                    found = true;
                    break;
                }
            }
            if (!found)
                _nodes.Add(node);
            return !found;
        }

        /// <summary>
        /// Adds a node with the given ID to this data graph. The node ID must be 
        /// unique within the data graph's nodes thus far.
        /// </summary>
        /// <param name="id">The unique ID to give this node.</param>
        /// <returns>True if successful; false otherwise.</returns>
        public bool AddNode(int id)
        {
            return AddNode(new Node(id));
        }

        /// <summary>
        /// Adds a node with the given ID and data object to this data graph. The node 
        /// ID must be unique within the data graph's nodes thus far.
        /// </summary>
        /// <param name="id">The unique ID to give this node.</param>
        /// <param name="data">The data object to assign to this node.</param>
        /// <returns>True if successful; false otherwise.</returns>
        public bool AddNode(int id, object data)
        {
            return AddNode(new Node(id, data));
        }

        /// <summary>
        /// Adds a node with a unique ID one greater than the currently-greatest ID in
        /// the graph. No data is associated with the node but the node instance is returned,
        /// allowing the caller to query its ID and associate data with it, if desired.
        /// </summary>
        /// <returns>The newly created node added to the graph.</returns>
        public Node AddNode()
        {
            int max = _nodes.Count > 0 ? int.MinValue : -1;
            for (int i = 0; i < _nodes.Count; i++)
            {
                if (_nodes[i].ID > max)
                    max = _nodes[i].ID;
            }
            Node node = new Node(max + 1);
            bool success = AddNode(node);
            Debug.Assert(success, "Failed to create node with unique ID in Graph.AddNode.");
            return node;
        }

        /// <summary>
        /// Removes the given node from the graph.
        /// </summary>
        /// <param name="node">The node to remove.</param>
        /// <returns>True if successful; false otherwise.</returns>
        public bool RemoveNode(Node node)
        {
            bool found = false;
            for (int i = _nodes.Count - 1; i >= 0; i--)
            {
                if (_nodes[i] == node)
                {
                    _nodes[i].DisconnectAll();
                    _nodes.RemoveAt(i);
                    found = true;
                }
            }
            return found;
        }

        /// <summary>
        /// Removes the node with the given ID from the data graph.
        /// </summary>
        /// <param name="id">The ID of the node to remove.</param>
        /// <remarks>The last node with the given ID to be added to this graph will
        /// be removed.</remarks>
        /// <returns>True if successful; false otherwise.</returns>
        public bool RemoveNode(int id)
        {
            Node node = GetNode(id);
            return RemoveNode(node);
        }

        /// <summary>
        /// Remove all nodes in this data graph. All connections are removed prior to 
        /// removing each node.
        /// </summary>
        public void RemoveAllNodes()
        {
            for (int i = _nodes.Count - 1; i >= 0; i--)
            {
                _nodes[i].DisconnectAll();
                _nodes.RemoveAt(i);
            }
        }

        #endregion

        #region Getting & Finding Nodes

        /// <summary>
        /// Tests whether this data graph has the given node.
        /// </summary>
        /// <param name="node">The node to test.</param>
        /// <returns>True if the node is contained within this graph; false otherwise.</returns>
        /// <remarks>The node does not have to be connected to any others to be considered
        /// a part of this data graph.</remarks>
        public bool HasNode(Node node)
        {
            for (int i = 0; i < _nodes.Count; i++)
            {
                if (_nodes[i] == node)
                    return true;
            }
            return false;
        }

        /// <summary>
        /// Tests whether this data graph has a node with the given ID.
        /// </summary>
        /// <param name="id">The ID of the node to test.</param>
        /// <returns>True if a node with the given ID is contained within this graph;
        /// false otherwise.</returns>
        /// <remarks>The node does not have to be connected to any others to be considered
        /// a part of this data graph.</remarks>
        public bool HasNode(int id)
        {
            for (int i = 0; i < _nodes.Count; i++)
            {
                if (_nodes[i].ID == id)
                    return true;
            }
            return false;
        }

        /// <summary>
        /// Finds the first node with the given data object.
        /// </summary>
        /// <param name="data">The data object to look for.</param>
        /// <returns>The node with the given data object.</returns>
        /// <remarks>This method uses the object.Equals method for comparisons on the data 
        /// object.</remarks>
        public Node FindNode(object data)
        {
            if (_nodes != null && _nodes.Count > 0)
            {
                for (int i = 0; i < _nodes.Count; i++)
                {
                    if (_nodes[i].Data.Equals(data))
                        return _nodes[i];
                }
            }
            return null;
        }

        /// <summary>
        /// Gets the node with the given ID.
        /// </summary>
        /// <param name="id">The node with the given ID to get.</param>
        /// <returns>The node with the given ID.</returns>
        public Node GetNode(int id)
        {
            for (int i = 0; i < _nodes.Count; i++)
            {
                if (_nodes[i].ID == id)
                    return _nodes[i];
            }
            return null;
        }

        /// <summary>
        /// Indexer over the graph's nodes. Note that the index is not the same
        /// as a node's ID, which is better thought of as a unique numeric tag.
        /// </summary>
        /// <param name="index">A 0-based index.</param>
        /// <returns>The node at the given 0-based index.</returns>
        public Node this[int index]
        {
            get
            {
                if (_nodes != null && 0 <= index && index < _nodes.Count)
                {
                    return _nodes[index];
                }
                else throw new IndexOutOfRangeException("Index out of range in NodeGraph indexer.");
            }
        }

        #endregion

        #region Connecting & Disconnecting Individual Nodes

        /// <summary>
        /// Connect the given nodes with an unweighted connection. The connection is unidirectional.
        /// </summary>
        /// <param name="from">The node out from which to form the connection.</param>
        /// <param name="to">The node into which to form the connection.</param>
        /// <returns>True if the connection was a success; false otherwise.</returns>
        public bool Connect(Node from, Node to)
        {
            return Connect(from, to, 0.0);
        }

        /// <summary>
        /// Connect the given nodes with a weighted connection. The connection is unidirectional.
        /// </summary>
        /// <param name="from">The node out from which to form the connection.</param>
        /// <param name="to">The node into which to form the connection.</param>
        /// <param name="weight">The weight of the connection.</param>
        /// <returns>True if the connection was a success; false otherwise.</returns>
        public bool Connect(Node from, Node to, double weight)
        {
            return Connect(from, to, weight, false);
        }

        /// <summary>
        /// Connect the given nodes with a weighted connection and optionally in a bidirectional fashion.
        /// </summary>
        /// <param name="from">The node out from which to form the connection.</param>
        /// <param name="to">The node into which to form the connection.</param>
        /// <param name="weight">The weight of the connection.</param>
        /// <param name="bidirectional">If true, the nodes are connected with two connections, one from each
        /// node to the other.</param>
        /// <returns>True if the connection was a success; false otherwise.</returns>
        public bool Connect(Node from, Node to, double weight, bool bidirectional)
        {
            bool b1 = false, b2 = false;
            if (from != null && to != null)
            {
                Connection cnxn = from.ConnectTo(to, weight); // connect the nodes!
                Debug.Assert(cnxn != null && cnxn.From == from && cnxn.To == to, "Buggy connection in NodeGraph.Connect.");
                b1 = true;

                if (bidirectional)
                {
                    cnxn = from.ConnectFrom(to, weight); // connect the other direction as well
                    Debug.Assert(cnxn != null && cnxn.From == to && cnxn.To == from, "Buggy connection in NodeGraph.Connect.");
                    b2 = true;
                }
            }
            return b1 && (!bidirectional || b2);
        }

        /// <summary>
        /// Connect the nodes with the given IDs with an unweighted connection. The connection is 
        /// unidirectional.
        /// </summary>
        /// <param name="idFrom">The ID of the node out from which to form the connection.</param>
        /// <param name="idTo">The ID of the node into which to form the connection.</param>
        /// <returns>True if the connection was a success; false otherwise.</returns>
        public bool Connect(int idFrom, int idTo)
        {
            return Connect(idFrom, idTo, 0.0);
        }

        /// <summary>
        /// Connect the nodes with the given IDs with a weighted connection. The connection is 
        /// unidirectional.
        /// </summary>
        /// <param name="idFrom">The ID of the node out from which to form the connection.</param>
        /// <param name="idTo">The ID of the node into which to form the connection.</param>
        /// <param name="weight">The weight of the connection.</param>
        /// <returns>True if the connection was a success; false otherwise.</returns>
        public bool Connect(int idFrom, int idTo, double weight)
        {
            return Connect(idFrom, idTo, weight, false);
        }

        /// <summary>
        /// Connect the given nodes with a weighted connection and optionally in a bidirectional fashion.
        /// </summary>
        /// <param name="idFrom">The ID of the node out from which to form the connection.</param>
        /// <param name="idTo">The ID of the node into which to form the connection.</param>
        /// <param name="weight">The weight of the connection.</param>
        /// <param name="bidirectional">If true, the nodes are connected with two connections, one from each
        /// node to the other.</param>
        /// <returns>True if the connection was a success; false otherwise.</returns>
        public bool Connect(int idFrom, int idTo, double weight, bool bidirectional)
        {
            Node from = GetNode(idFrom);
            Node to = GetNode(idTo);
            return Connect(from, to, weight, bidirectional);
        }

        /// <summary>
        /// Disconnect one node connecting to another. The connection is considered unidirectional.
        /// </summary>
        /// <param name="from">The node out from which to break the connection.</param>
        /// <param name="to">The node into which to break the connection.</param>
        /// <returns>True if the nodes are disconnected successfully; false otherwise.</returns>
        /// <remarks>If the nodes were not connected already, false is returned. Also, the connection
        /// must be from the 'from' node to the 'to' node.</remarks>
        public bool Disconnect(Node from, Node to)
        {
            return Disconnect(from, to, false);
        }

        /// <summary>
        /// Disconnect one node connecting to another. The connection may be considered unidirectional
        /// or bidirectional.
        /// </summary>
        /// <param name="from">The node out from which to break the connection.</param>
        /// <param name="to">The node into which to break the connection.</param>
        /// <param name="bidirectional">If true, connections in both directions will be severed. In other
        /// words, it does not matter which node is 'from' and which is 'to'.</param>
        /// <returns>True if the nodes were disconnected successfully; false otherwise.</returns>
        /// <remarks>If the nodes were not connected already, false is returned.</remarks>
        public bool Disconnect(Node from, Node to, bool bidirectional)
        {
            int cnx = 0;
            if (from != null && to != null)
            {
                cnx = from.Disconnect(to);
                if (bidirectional)
                    cnx += to.Disconnect(from);
            }
            return cnx > 0;
        }

        /// <summary>
        /// Disconnect one node connecting to another. The connection is considered unidirectional.
        /// </summary>
        /// <param name="from">The first node ID out from which to break the connection.</param>
        /// <param name="to">The first node ID into which to break the connection.</param>
        /// <returns>True if the nodes are disconnected successfully; false otherwise.</returns>
        /// <remarks>If the nodes were not connected already, false is returned. Also, the connection
        /// must be from the 'from' node to the 'to' node.</remarks>
        public bool Disconnect(int idFrom, int idTo)
        {
            return Disconnect(idFrom, idTo, false);
        }

        /// <summary>
        /// Disconnect one node connecting to another. The connection may be considered unidirectional
        /// or bidirectional.
        /// </summary>
        /// <param name="idFrom">The first node ID out from which to break the connection.</param>
        /// <param name="idTo">The first node ID into which to break the connection.</param>
        /// <param name="bidirectional">If true, connections in both directions will be severed. In other
        /// words, it does not matter which node is 'from' and which is 'to'.</param>
        /// <returns>True if the nodes were disconnected successfully; false otherwise.</returns>
        /// <remarks>If the nodes were not connected already, false is returned.</remarks>
        public bool Disconnect(int idFrom, int idTo, bool bidirectional)
        {
            Node from = GetNode(idFrom);
            Node to = GetNode(idTo);
            return Disconnect(from, to, bidirectional);
        }

        #endregion

        #region Connecting & Disconnecting Groups of Nodes

        /// <summary>
        /// Connects from the given node to all other nodes in the graph with an unweighted
        /// connection. Circular connections are avoided. Unidirectional connections are assumed.
        /// </summary>
        /// <param name="node">The node from which to connect.</param>
        public void ConnectToAll(Node node)
        {
            ConnectToAll(node, 0.0);
        }

        /// <summary>
        /// Connects from the given node to all other nodes in the graph with a 
        /// weighted connection. Circular connections are avoided. Unidirectional 
        /// connections are assumed.
        /// </summary>
        /// <param name="node">The node from which to connect.</param>
        /// <param name="weight">The weight of each connection.</param>
        public void ConnectToAll(Node node, double weight)
        {
            ConnectToAll(node, weight, false);
        }

        /// <summary>
        /// Connects from the given node to all other nodes in the graph with a
        /// weighted connection. The option for circular connections is provided.
        /// Unidirectional connections are assumed.
        /// </summary>
        /// <param name="node">The node from which to connect.</param>
        /// <param name="weight">The weight of each connection.</param>
        /// <param name="wantCircular">If true, a circular connection will be established.</param>
        public void ConnectToAll(Node node, double weight, bool wantCircular)
        {
            ConnectToAll(node, weight, wantCircular, false);
        }

        /// <summary>
        /// Connects from the given node to all other nodes in the graph with a weighted
        /// connection. Options for circular and bidirectional connections are provided.
        /// </summary>
        /// <param name="node">The node from which to connect.</param>
        /// <param name="weight">The weight of each connection.</param>
        /// <param name="wantCircular">If true, a circular connection will be established.</param>
        /// <param name="bidirectional">If true, bidirectional connections will be established.</param>
        public void ConnectToAll(Node node, double weight, bool wantCircular, bool bidirectional)
        {
            if (node != null)
            {
                for (int i = 0; i < _nodes.Count; i++)
                {
                    if (node != _nodes[i] || wantCircular)
                        node.ConnectTo(_nodes[i], weight);
                    if (bidirectional && node != _nodes[i])
                        node.ConnectFrom(_nodes[i], weight);
                }
            }
        }

        /// <summary>
        /// Connects from the node with the given ID to all other nodes in the graph with 
        /// an unweighted connection. Circular connections are avoided. Unidirectional 
        /// connections are assumed.
        /// </summary>
        /// <param name="idFrom">The ID of the node from which to connect.</param>
        public void ConnectToAll(int idFrom)
        {
            ConnectToAll(idFrom, 0.0);
        }

        /// <summary>
        /// Connects from the node with the given ID to all other nodes in the graph with 
        /// a weighted connection. Circular connections are avoided. Unidirectional 
        /// connections are assumed.
        /// </summary>
        /// <param name="idFrom">The ID of the node from which to connect.</param>
        /// <param name="weight">The weight of each connection.</param>
        public void ConnectToAll(int idFrom, double weight)
        {
            ConnectToAll(idFrom, weight, false);
        }

        /// <summary>
        /// Connects from the the node with the given ID to all other nodes in the graph with 
        /// a weighted connection. The option for circular connections is provided. Unidirectional 
        /// connections are assumed.
        /// </summary>
        /// <param name="idFrom">The ID of the node from which to connect.</param>
        /// <param name="weight">The weight of each connection.</param>
        /// <param name="wantCircular">If true, a circular connection will be established.</param>
        public void ConnectToAll(int idFrom, double weight, bool wantCircular)
        {
            ConnectToAll(idFrom, weight, wantCircular, false);
        }

        /// <summary>
        /// Connects from the node with the given ID to all other nodes in the graph with a
        /// weighted connection. Options for circular and bidirectional connections are
        /// provided.
        /// </summary>
        /// <param name="idFrom">The ID of the node from which to connect.</param>
        /// <param name="weight">The weight of each connection.</param>
        /// <param name="wantCircular">If true, a circular connection will be established.</param>
        /// <param name="bidirectional">If true, bidirectional connections will be established.</param>
        public void ConnectToAll(int idFrom, double weight, bool wantCircular, bool bidirectional)
        {
            Node node = GetNode(idFrom);
            ConnectToAll(node, weight, wantCircular, bidirectional);
        }

        /// <summary>
        /// Connects to the given node from all other nodes in the graph with an unweighted
        /// connection. Circular connections are avoided. Unidirectional connections are assumed.
        /// </summary>
        /// <param name="node">The node to which to connect.</param>
        public void ConnectFromAll(Node node)
        {
            ConnectFromAll(node, 0.0);
        }

        /// <summary>
        /// Connects to the given node from all other nodes in the graph with a 
        /// weighted connection. Circular connections are avoided. Unidirectional connections are 
        /// assumed.
        /// </summary>
        /// <param name="node">The node to which to connect.</param>
        /// <param name="weight">The weight of each connection.</param>
        public void ConnectFromAll(Node node, double weight)
        {
            ConnectFromAll(node, weight, false);
        }

        /// <summary>
        /// Connects to the given node from all other nodes in the graph with a weighted 
        /// connection. The option for circular connections is provided. Unidirectional 
        /// connections are assumed.
        /// </summary>
        /// <param name="node">The node to which to connect.</param>
        /// <param name="weight">The weight of each connection.</param>
        /// <param name="wantCircular">If true, a circular connection will be established.</param>
        public void ConnectFromAll(Node node, double weight, bool wantCircular)
        {
            ConnectFromAll(node, weight, wantCircular, false);
        }

        /// <summary>
        /// Connects to the given node from all other nodes in the graph with a weighted 
        /// connection. Options for circular and bidirectional connections are provided.
        /// </summary>
        /// <param name="node">The node to which to connect.</param>
        /// <param name="weight">The weight of each connection.</param>
        /// <param name="wantCircular">If true, a circular connection will be established.</param>
        /// <param name="bidirectional">If true, bidirectional connections will be established.</param>
        public void ConnectFromAll(Node node, double weight, bool wantCircular, bool bidirectional)
        {
            if (node != null)
            {
                for (int i = 0; i < _nodes.Count; i++)
                {
                    if (node != _nodes[i] || wantCircular)
                        node.ConnectFrom(_nodes[i], weight);
                    if (bidirectional && node != _nodes[i])
                        node.ConnectTo(_nodes[i], weight);
                }
            }
        }

        /// <summary>
        /// Connects to the node with the given ID from all other nodes in the graph with 
        /// an unweighted connection. Circular connections are avoided. Unidirectional connections 
        /// are assumed.
        /// </summary>
        /// <param name="idTo">The ID of the node to which to connect.</param>
        public void ConnectFromAll(int idTo)
        {
            ConnectFromAll(idTo, 0.0);
        }

        /// <summary>
        /// Connects to the node with the given ID from all other nodes in the graph with 
        /// a weighted connection. Circular connections are avoided. Unidirectional connections 
        /// are assumed.
        /// </summary>
        /// <param name="idTo">The ID of the node to which to connect.</param>
        /// <param name="weight">The weight of each connection.</param>
        public void ConnectFromAll(int idTo, double weight)
        {
            ConnectFromAll(idTo, weight, false);
        }

        /// <summary>
        /// Connects to the the node with the given ID from all other nodes in the graph with 
        /// a weighted connection. The option for circular connections is provided. Unidirectional 
        /// connections are assumed.
        /// </summary>
        /// <param name="idTo">The ID of the node to which to connect.</param>
        /// <param name="weight">The weight of each connection.</param>
        /// <param name="wantCircular">If true, a circular connection will be established.</param>
        public void ConnectFromAll(int idTo, double weight, bool wantCircular)
        {
            ConnectFromAll(idTo, weight, wantCircular, false);
        }

        /// <summary>
        /// Connects to the node with the given ID from all other nodes in the graph with a weighted 
        /// connection. Options for circular and bidirectional connections are provided.
        /// </summary>
        /// <param name="idTo">The ID of the node to which to connect.</param>
        /// <param name="weight">The weight of each connection.</param>
        /// <param name="wantCircular">If true, a circular connection will be established.</param>
        /// <param name="bidirectional">If true, bidirectional connections will be established.</param>
        public void ConnectFromAll(int idTo, double weight, bool wantCircular, bool bidirectional)
        {
            Node node = GetNode(idTo);
            ConnectFromAll(node, weight, wantCircular, bidirectional);
        }

        /// <summary>
        /// Connects all nodes to all nodes. Since all nodes are connected to all others, bidirectional
        /// connections are inherently established.
        /// </summary>
        /// <param name="wantCircular">If true, a circular connections will be established.</param>
        public void ConnectAll(bool wantCircular)
        {
            for (int i = 0; i < _nodes.Count; i++)
            {
                for (int j = 0; j < _nodes.Count; j++)
                {
                    if (i != j || wantCircular)
                    {
                        _nodes[i].ConnectTo(_nodes[j]);
                    }
                }
            }
        }

        /// <summary>
        /// Disconnects all nodes from all others.
        /// </summary>
        public void DisconnectAll()
        {
            for (int i = 0; i < _nodes.Count; i++)
            {
                _nodes[i].DisconnectAll();
            }
        }

        /// <summary>
        /// Gets a list of all the connections in the data graph.
        /// </summary>
        /// <param name="wantCircular">If true, circular connections will be included.</param>
        /// <returns>A list of all connections.</returns>
        public List<Connection> GetAllConnections(bool wantCircular)
        {
            List<Connection> allCnxns = new List<Connection>();
            for (int i = 0; i < _nodes.Count; i++)
            {
                List<Connection> cnxns = _nodes[i].GetAllConnections(wantCircular);
                if (cnxns != null && cnxns.Count > 0)
                    allCnxns.AddRange(cnxns);
            }
            return allCnxns;
        }

        #endregion

        #region Overrides

        /// <summary>
        /// Gets a string representation of this data graph.
        /// </summary>
        /// <returns>A string representation.</returns>
        public override string ToString()
        {
            return String.Format("Nodes: {0}; Connections: {1}", this.Count, this.GetAllConnections(true).Count);
        }

        /// <summary>
        /// Tests whether this data graph is equal to the given object.
        /// </summary>
        /// <param name="obj">The given object to test.</param>
        /// <returns></returns>
        public override bool Equals(object obj)
        {
            if (!(obj is Graph))
                return false;
            Graph o = (Graph) obj;
            return base.Equals(o);
        }

        /// <summary>
        /// Gets a hash code for this data graph.
        /// </summary>
        /// <returns></returns>
        public override int GetHashCode()
        {
            return base.GetHashCode();
        }

        #endregion

        #region Correlation Clustering

        /// <summary>
        /// Implements the CC-Pivot correlation clustering algorithm of Ailon et al. (2005) to group nodes.
        /// </summary>
        /// <param name="tFn">A callback function used to test whether a node has affinity for the pivot node
        /// from the CC-Pivot function of Alison et al. Affinity between nodes can be as simple as (+,-) signs
        /// on each edge, or more complex like weighted real values. The caller decides whether two nodes
        /// share affinity or not.</param>
        /// <param name="oFn">Optional callback function to indicate progress. If used, it is called to
        /// indicate the number of nodes remaining to be clustered, so progress starts with a count of
        /// all the nodes in the original graph, and descends to zero once all nodes have been clustered.</param>
        /// <returns>A list of node clusters.</returns>
        /// <remarks>A single grouping might or might not be that effective based on the membership test
        /// defined in the caller-supplied callback function. One solution is to simply run the grouping
        /// many, many times and, with some kind of scoring function, take the best grouping that emerges.
        /// Doing so removes the dependency on the random choice made for the pivot node.</remarks>
        public List<List<Node>> Cluster(NodeAffinityCallback tFn, OptimizationCallback oFn = null)
        {
            if (_nodes == null || _nodes.Count == 0)
                return null;

            List<List<Node>> clusters = new List<List<Node>>(_nodes.Count);
            ClusterHelper(clusters, this, tFn, oFn); // continues to update clusters
            return clusters;
        }

        /// <summary>
        /// A helper function for the recursion for implementing correlation clustering.
        /// </summary>
        /// <param name="clusters">The current clusters being built up as a list of node lists.</param>
        /// <param name="graph">The current graph left to cluster. This graph shrinks each recursion, 
        /// making the problem simpler and simpler until all nodes are clustered.</param>
        /// <param name="tFn">See entry for Cluster, above.</param>
        /// <param name="oFn">See entry for Cluster, above.</param>
        private void ClusterHelper(List<List<Node>> clusters, Graph graph, NodeAffinityCallback tFn, OptimizationCallback oFn = null)
        {
            oFn?.Invoke(graph.Count); // pass the number of nodes that have yet to be clustered

            if (graph.Count == 0)
                return; // base case: all nodes clustered!

            int i = RandomEx.Integer(0, graph.Count - 1);
            Node pivot = graph[i]; // uniform random pivot node

            clusters.Add(new List<Node>(graph.Count)); // C in CC-Pivot
            clusters[clusters.Count - 1].Add(pivot); // start a new cluster for this recursion

            Graph subgraph = new Graph(); // V' in CC-Pivot

            for (int j = 0; j < graph.Count; j++) // for all other nodes in the graph
            {
                if (j != i)
                {
                    bool added = false;
                    Node node = graph[j];
                    if (pivot.HasConnectionTo(node))
                    {
                        Connection cnxn = pivot.GetConnectionTo(node);
                        if (tFn.Invoke(cnxn)) // ask the caller for affinity judgment
                        {
                            clusters[clusters.Count - 1].Add(node); // add to C in CC-Pivot
                            added = true;
                        }
                    }
                    if (!added)
                        subgraph.AddNode(node); // add to V' in CC-Pivot
                }
            }
            ClusterHelper(clusters, subgraph, tFn, oFn); // recursive step: subgraph is simpler than graph
        }

        #endregion

        #endregion
    }
}