﻿using System;
using System.IO;
using System.Xml;

namespace WobbrockLib.Data
{
    /// <summary>
    /// Defines an interface for an XML log to be used in a user study. An instance
    /// that is XML loggable can write itself out to XML and create itself by reading XML in.
    /// It can also parse XML and perform analyses, writing the results to a CSV file for 
    /// analysis in programs like Microsoft Excel or the R statistical programming language.
    /// </summary>
    public interface IXmlLoggable
    {
        /// <summary>
        /// Writes part or all of this data object to XML. If this data object owns other
        /// data objects that will also be written, this method may leave some XML elements
        /// open, which will be closed with a later call to <i>WriteXmlFooter</i>.
        /// </summary>
        /// <param name="writer">An open XML writer.</param>
        /// <returns>Returns <b>true</b> if successful; <b>false</b> otherwise.</returns>
        bool WriteXmlHeader(XmlTextWriter writer);

        /// <summary>
        /// Writes any closing XML necessary for this data object. This method can simply
        /// return <b>true</b> if all data was already written in <i>WriteXmlHeader</i>.
        /// </summary>
        /// <param name="writer">An open XML writer.</param>
        /// <returns>Returns <b>true</b> if successful; <b>false</b> otherwise.</returns>
        bool WriteXmlFooter(XmlTextWriter writer);

        /// <summary>
        /// Reads a data object from XML and returns an instance of the object that implements
        /// this interface.
        /// </summary>
        /// <param name="reader">An open XML reader.</param>
        /// <returns>Returns <b>true</b> if successful; <b>false</b> otherwise.</returns>
        /// <remarks>Clients should first create a new instance of the implementer of this interface
        /// using a default constructor, and then call this method to populate the data fields of that 
        /// instance.</remarks>
        bool ReadFromXml(XmlTextReader reader);

        /// <summary>
        /// Performs any analyses on this data object and writes the results to a CSV file for
        /// subsequent analysis in Microsoft Excel or a statistics package like R.
        /// </summary>
        /// <param name="writer">An open stream writer pointed to a CSV file.</param>
		/// <param name="writeColnames">If true, column names will be written; if false, only data
		/// will be written, not the column headings.</param>
		/// <param name="data">Optional additional parameters to be used when writing out results.</param>
        /// <returns>Returns <b>true</b> if writing is successful; <b>false</b> otherwise.</returns>
        bool WriteResultsToCsv(StreamWriter writer, bool writeColnames, params object[] data);
    }
}