﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace WobbrockLib.Data
{
	/// <summary>
	/// Provides a generic first-in, first-out queue that is also indexable.
	/// </summary>
	/// <typeparam name="T"></typeparam>
	public class IndexableQueue<T>
	{
		#region Fields

		private List<T> _queue;

		#endregion

		#region Constructors

		/// <summary>
		/// Default constructor.
		/// </summary>
		public IndexableQueue()
		{
			_queue = new List<T>();
		}

		/// <summary>
		/// Constructor of a queue with a given initial capacity.
		/// </summary>
		/// <param name="capacity">The initial capacity of the queue.</param>
		public IndexableQueue(int capacity)
		{
			_queue = new List<T>(capacity);
		}

		/// <summary>
		/// Constructor of a queue with a given initial collection.
		/// </summary>
		/// <param name="collection"></param>
		public IndexableQueue(IEnumerable<T> collection)
		{
			_queue = new List<T>(collection);
		}

		#endregion

		#region Queuing

		/// <summary>
		/// Adds the given item to the end of the queue.
		/// </summary>
		/// <param name="item">The item to enqueue.</param>
		public void Enqueue(T item)
		{
			_queue.Add(item);
		}

		/// <summary>
		/// Removes and returns the item at the start of the queue.
		/// </summary>
		/// <returns>The item at the start of the queue.</returns>
		public T Dequeue()
		{
			if (_queue.Count > 0)
			{
				T item = _queue[0];
				_queue.RemoveAt(0);
				return item;
			}
			else return default(T);
		}

		#endregion

		#region Indexing

		/// <summary>
		/// Gets or sets the item at the given index.
		/// </summary>
		/// <param name="index">The zero-based index.</param>
		/// <returns>The item at the given index.</returns>
		public T this[int index]
		{
			get { return _queue[index]; }
			set { _queue[index] = value; }
		}

		/// <summary>
		/// Gets the number of items that are currently in this queue.
		/// </summary>
		public int Count
		{
			get { return _queue.Count; }
		}

		#endregion

		#region Peeking

		/// <summary>
		/// Peeks at the item at the given index. It does not remove the item at the index.
		/// </summary>
		/// <param name="index">The zero-based index of the item to peek. Index 0 is the head of
		/// the queue, which was added longest ago. The index of the highest value will be the 
		/// item added most recently.</param>
		/// <returns>The item at the given index.</returns>
		public T PeekAt(int index)
		{
			return _queue[index];
		}

		/// <summary>
		/// Peeks at the item at the start of the queue. This is the item at index 0, which was
		/// added longest ago. This is also the same item that would be returned by a call to 
		/// Dequeue.
		/// </summary>
		/// <returns>The item at the start of the queue, or the item type's default value
		/// if the queue is empty.</returns>
		public T PeekStart()
		{
			return (_queue.Count > 0) ? _queue[0] : default(T);
		}

		/// <summary>
		/// Peeks at the item at the end of the queue. This is the item at the greatest index,
		/// the item added most recently.
		/// </summary>
		/// <returns>The item at the end of the queue, or the item type's default value 
		/// if the queue is empty.</returns>
		public T PeekEnd()
		{
			return (_queue.Count > 0) ? _queue[_queue.Count - 1] : default(T);
		}

		#endregion

		#region Removing

		/// <summary>
		/// Removes the item at the given index.
		/// </summary>
		/// <param name="index">The zero-based index of the item to remove. Index 0 is the head of
		/// the queue, which was added longest ago. The index of the highest value will be the 
		/// item added most recently.</param>
		public void RemoveAt(int index)
		{
			_queue.RemoveAt(index);
		}

		/// <summary>
		/// Removes the item at the start of the queue. This is the item at index 0, which was
		/// added longest ago. This is also the same item that would be returned by a call to 
		/// Dequeue.
		/// </summary>
		public void RemoveStart()
		{
			if (_queue.Count > 0)
			{
				_queue.RemoveAt(0);
			}
		}

		/// <summary>
		/// Removes the item at the end of the queue. This is the item at the greatest index,
		/// the item added most recently.
		/// </summary>
		public void RemoveEnd()
		{
			if (_queue.Count > 0)
			{
				_queue.RemoveAt(_queue.Count - 1);
			}
		}

		/// <summary>
		/// Clears the queue of all its items.
		/// </summary>
		public void Clear()
		{
			_queue.Clear();
		}

		#endregion

		#region Overrides

		/// <summary>
		/// Gets a string representation of this IndexableQueue.
		/// </summary>
		/// <returns></returns>
		public override string ToString()
		{
			return _queue.ToString();
		}

		/// <summary>
		/// Determines whether this IndexableQueue is equal to the one given.
		/// </summary>
		/// <param name="obj">An IndexableQueue instance to compare.</param>
		/// <returns>True if this queue and the one given are equal; false otherwise.</returns>
		public override bool Equals(object obj)
		{
			return _queue.Equals(obj);
		}

		/// <summary>
		/// Gets a hash code for this IndexableQueue instance.
		/// </summary>
		/// <returns>A hash code.</returns>
		public override int GetHashCode()
		{
			return _queue.GetHashCode();
		}

		#endregion

	}
}