﻿using System;
using System.IO;

namespace WobbrockLib.Data
{
	/// <summary>
	/// This class uses a memory mapped file to represent a 1-D array. This can be useful for very large arrays exceeding C#'s 2GB 
	/// object limit.
	/// </summary>
	/// <typeparam name="T">The data type stored in this 1-D array. Note that the value must be a non-nullable struct, which includes 
	/// the atomic C# types like int, char, byte, etc. It cannot be an object reference.</typeparam>
	public class MemoryMappedArray<T> : AbsMemoryMappedArray<T> where T : struct
	{
		// Fields
		private long _length;

		/// <summary>
		/// Constructor for a memory mapped file representing a 1-D array. Before the array can be used, it must be initialized.
		/// </summary>
		public MemoryMappedArray()
		{
			_length = 0L;
		}

		/// <summary>
		/// Initializes the memory mapped 1-D array with the given number of slots.
		/// </summary>
		/// <param name="length">The number of slots to initialize in the 1-D array.</param>
		/// <returns>True if initialized successfully; false otherwise.</returns>
		public new bool Initialize(long length)
		{
			_length = length;
			return base.Initialize(length * _typeSize);
		}

		/// <summary>
		/// Creates a memory mapped 1-D array on disk, rather than from system memory, of the given length and initial value.
		/// </summary>
		/// <param name="file">The file to create on disk.</param>
		/// <param name="length">The length in number of array slots to write.</param>
		/// <param name="useIfExists">If true and a file exists at the given location, it will be used. If false, a new file will
		/// be created.</param>
		/// <returns>True if initialized successfully; false otherwise.</returns>
		public new bool Initialize(string file, long length, bool useIfExists)
		{
			_length = length;
			return base.Initialize(file, length * _typeSize, useIfExists);
		}

		/// <summary>
		/// Gets the length with which this 1-D array was initialized.
		/// </summary>
		public long Length
		{
			get { return _length; }
		}

		/// <summary>
		/// Gets the value at the given index in the 1-D memory mapped array.
		/// </summary>
		/// <param name="index">The index into the array.</param>
		/// <returns>The value at [index] in the 1-D memory mapped array.</returns>
		public new T GetValue(long index)
		{
			return base.GetValue(index * _typeSize);
		}

		/// <summary>
		/// Sets the value at the given index in the 1-D memory mapped array.
		/// </summary>
		/// <param name="index">The index into the array.</param>
		/// <param name="value">The value to set at the given index.</param>
		public new void SetValue(long index, T value)
		{
			base.SetValue(index * _typeSize, value);
		}

		/// <summary>
		/// Indexer for this 1-D array.
		/// </summary>
		/// <param name="index">The index at which to get or set the given array value.</param>
		/// <returns>The array value stored at the given index.</returns>
		public new T this[long index]
		{
			get { return this.GetValue(index); }
			set { this.SetValue(index, value); }
		}

		/// <summary>
		/// Copies a number of items from the source array to the destination array at the given offsets. 
		/// </summary>
		/// <param name="sourceArray">The array from which to copy items.</param>
		/// <param name="sourceIndex">The source index at which to start copying items.</param>
		/// <param name="destinationArray">The array to which to copy items.</param>
		/// <param name="destinationIndex">The destination index to which to start copying items.</param>
		/// <param name="length">The number of items to copy.</param>
		public static void Copy(
			MemoryMappedArray<T> sourceArray, long sourceIndex, 
			MemoryMappedArray<T> destinationArray, long destinationIndex, 
			long length
			)
		{
			AbsMemoryMappedArray<T>.Copy(sourceArray, sourceIndex * _typeSize, destinationArray, destinationIndex * _typeSize, length * _typeSize);
		}
	}
}
