﻿using System;
using System.IO;

namespace WobbrockLib.Data
{
	/// <summary>
	/// This class uses a memory mapped file to represent a 2-D array (i.e., matrix). This can be useful for very large matrices 
	/// exceeding C#'s 2GB object limit.
	/// </summary>
	/// <typeparam name="T">The data type stored in this matrix. Note that the value must be a non-nullable struct, which includes 
	/// the atomic C# types like int, char, byte, etc. It cannot be an object reference.</typeparam>
	public class MemoryMappedMatrix<T> : AbsMemoryMappedArray<T> where T : struct
	{
		// Fields
		private long _cols;
		private long _rows;

		/// <summary>
		/// Constructor for a memory mapped file representing a 2-D array (i.e., matrix). Before the matrix can be used, it must be 
		/// initialized.
		/// </summary>
		public MemoryMappedMatrix()
		{
			_cols = 0L;
			_rows = 0L;
		}

		/// <summary>
		/// Initializes the memory mapped 2-D array (i.e., matrix) with the given columns and rows.
		/// </summary>
		/// <param name="cols">The number of columns in this matrix. This is the first dimension of a conventional 2-D array.</param>
		/// <param name="rows">The number of rows in this matrix. This is the second dimension of a conventional 2-D array.</param>
		/// <returns>True if initialized successfully; false otherwise.</returns>
		/// <remarks>The memory mapped file stores data in a 1-D sequence, like any file does. So we must retain the number of 
		/// columns in this 2-D array so that when indexing into it as [i,j], we can do the math to access the proper file offset.
		/// </remarks>
		public bool Initialize(long cols, long rows)
		{
			_cols = cols;
			_rows = rows;
			return base.Initialize(cols * rows * _typeSize);
		}

		/// <summary>
		/// Creates a memory mapped 2-D array (i.e., matrix) on disk, rather than from system memory, with the given columns and rows.
		/// </summary>
		/// <param name="file">The file to create on disk.</param>
		/// <param name="cols">The number of columns in this matrix. This is the first dimension of a conventional 2-D array.</param>
		/// <param name="rows">The number of rows in this matrix. This is the second dimension of a conventional 2-D array.</param>
		/// <param name="useIfExists">If true and a file exists at the given location, it will be used. If false, a new file will
		/// be created.</param>
		/// <returns>True if initialization succeeds; false otherwise.</returns>
		public bool Initialize(string file, long cols, long rows, bool useIfExists)
		{
			_cols = cols;
			_rows = rows;
			return base.Initialize(file, cols * rows * _typeSize, useIfExists);
		}

		/// <summary>
		/// Gets the number of columns with which this 2-D matrix was initialized. This value corresponds to 'i' in an [i,j] array
		/// indexer.
		/// </summary>
		public long Columns
		{
			get { return _cols; }
		}

		/// <summary>
		/// Gets the number of rows with which this 2-D matrix was initialized. This value corresponds to 'j' in an [i,j] array
		/// indexer.
		/// </summary>
		public long Rows
		{
			get { return _rows; }
		}

		/// <summary>
		/// Gets the value at index [i,j] in the 2-D memory mapped matrix.
		/// </summary>
		/// <param name="i">The column index into the matrix.</param>
		/// <param name="j">The row index into the matrix.</param>
		/// <returns>The value at index [i,j] in the 2-D memory mapped matrix.</returns>
		/// <remarks>
		/// To go from 2-D index [i,j] to indexing into the memory mapped file you use [i + j * nCols].
		/// To go from 1-D index [x] in the memory mapped file to recover the [i,j] indices, you would use [x % nCols, x / nCols].
		/// </remarks>
		public T GetValue(long i, long j)
		{
			long index = i + j * _cols;
			return base.GetValue(index * _typeSize);
		}

		/// <summary>
		/// Sets the value at index [i,j] in the 2-D memory mapped matrix.
		/// </summary>
		/// <param name="i">The column index into the matrix.</param>
		/// <param name="j">The row index into the matrix.</param>
		/// <param name="value">The value to set at index [i,j] in the matrix.</param>
		/// <remarks>
		/// To go from 2-D index [i,j] to indexing into the memory mapped file you use [i + j * nCols].
		/// To go from 1-D index [x] in the memory mapped file to recover the [i,j] indices, you would use [x % nCols, x / nCols].
		/// </remarks>
		public void SetValue(long i, long j, T value)
		{
			long index = i + j * _cols;
			base.SetValue(index * _typeSize, value);
		}

		/// <summary>
		/// Indexer for this 2-D matrix.
		/// </summary>
		/// <param name="i">The column index at which to get or set the given array value.</param>
		/// <param name="j">The row index at which to get or set the given array value.</param>
		/// <returns>The array value stored at the given [i,j] indices.</returns>
		public T this[long i, long j]
		{
			get { return this.GetValue(i, j); }
			set { this.SetValue(i, j, value); }
		}

		/// <summary>
		/// Copies a contiguous number of items from the source matrix to the destination matrix at the given indices. 
		/// </summary>
		/// <param name="sourceMatrix">The array from which to copy items.</param>
		/// <param name="sourceColumnIndex">The source column index from which to start copying items. This is 'i' in the [i,j] indexer.</param>
		/// <param name="sourceRowIndex">The source row index from which to start copying items. This is 'j' in the [i,j] indexer.</param>
		/// <param name="destinationMatrix">The array to which to copy items.</param>
		/// <param name="destinationColumnIndex">The destination column index to which to start copying items. This is 'i' in the [i,j] indexer.</param>
		/// <param name="destinationRowIndex">The destination row index to which to start copying items. This is 'j' in the [i,j] indexer.</param>
		/// <param name="count">The number of items to copy.</param>
		/// <remarks>All 'count' items starting at [sourceColumnIndex, sourceRowIndex] are copied contiguously to 
		/// [destinationColumnIndex, destinationRowIndex].</remarks>
		public static void Copy(
			MemoryMappedMatrix<T> sourceMatrix, long sourceColumnIndex, long sourceRowIndex, 
			MemoryMappedMatrix<T> destinationMatrix, long destinationColumnIndex, long destinationRowIndex, 
			long count
			)
		{
			long sourceIndex = sourceColumnIndex + sourceRowIndex * sourceMatrix._cols;
			long destinationIndex = destinationColumnIndex + destinationRowIndex * destinationMatrix._cols;
			AbsMemoryMappedArray<T>.Copy(sourceMatrix, sourceIndex * _typeSize, destinationMatrix, destinationIndex * _typeSize, count * _typeSize);
		}
	}
}
