﻿using System;
using System.Windows.Forms;
using System.Diagnostics;
using System.Runtime.InteropServices;
using WobbrockLib.Windows;

namespace WobbrockLib.Devices
{
	/// <summary>
	/// 
	/// </summary>
    public class LowLevelKeyboardHook
    {
        public event EventHandler<KeyEventArgs> HookKeyDown;
		public event EventHandler<KeyPressEventArgs> HookKeyPress;
		public event EventHandler<KeyEventArgs> HookKeyUp;
        
        // hook members
        private string _name;
        private IntPtr _hookID;
        private Win32.WindowsHookProc _hookProc;

        /// <summary>
        /// 
        /// </summary>
        /// <param name="name"></param>
        public LowLevelKeyboardHook(string name)
        {
            _name = name;
            _hookID = IntPtr.Zero;
            _hookProc = HookCallback;
            
        }

        /// <summary>
        /// 
        /// </summary>
        ~LowLevelKeyboardHook()
        {
            Uninstall();
        }

        /// <summary>
        /// 
        /// </summary>
        public string Name
        {
            get { return _name; }
        }

        /// <summary>
        /// 
        /// </summary>
        public void Install()
        {
            Uninstall(); // stop the hook if it is already active first

            // then install a new hook
            Process procCurrent = Process.GetCurrentProcess();
            ProcessModule procMod = procCurrent.MainModule;
            _hookID = Win32.SetWindowsHookEx((int) Win32.WH.KEYBOARD_LL, _hookProc, Win32.GetModuleHandle(procMod.ModuleName), 0u);
            procMod.Dispose();
            procCurrent.Dispose();
        }

        /// <summary>
        /// 
        /// </summary>
        public void Uninstall()
        {
            if (_hookID != IntPtr.Zero)
            {
                Win32.UnhookWindowsHookEx(_hookID);
            }
            _hookID = IntPtr.Zero;
        }

        /// <summary>
        /// 
        /// </summary>
        public bool IsActive
        {
            get { return _hookID != IntPtr.Zero; }
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="nCode"></param>
        /// <param name="wParam"></param>
        /// <param name="lParam"></param>
        /// <returns></returns>
        private IntPtr HookCallback(int nCode, int wParam, IntPtr lParam)
        {
            bool handled = false;

            if (nCode >= 0)
            {
                Win32.KBDLLHOOKSTRUCT info = (Win32.KBDLLHOOKSTRUCT) Marshal.PtrToStructure(lParam, typeof(Win32.KBDLLHOOKSTRUCT));
                //
                // KeyDown
                //
                if (HookKeyDown != null && (wParam == (int) Win32.WM.KEYDOWN || wParam == (int) Win32.WM.SYSKEYDOWN))
                {
                    Keys data = (Keys) info.vkCode;
                    Keys mod = GetModifierKey(data); // get if a modifier key was just pressed

                    // test the modifier keys for whether they are down; doesn't apply to them being pressed themselves
                    bool shift = ((Win32.GetKeyState((int) Win32.VK.SHIFT) & 0x80) == 0x80 ? true : false);
                    bool ctrl = ((Win32.GetKeyState((int) Win32.VK.CONTROL) & 0x80) == 0x80 ? true : false);
                    bool alt = ((Win32.GetKeyState((int) Win32.VK.MENU) & 0x80) == 0x80 ? true : false);

                    // if a modifier key is down or was just pressed, add it to the data
                    if (shift || mod == Keys.LShiftKey || mod == Keys.RShiftKey)
                        data |= Keys.Shift;
                    if (ctrl || mod == Keys.LControlKey || mod == Keys.RControlKey)
                        data |= Keys.Control;
                    if (alt || mod == Keys.LMenu || mod == Keys.RMenu)
                        data |= Keys.Alt;

                    KeyEventArgs e = new KeyEventArgs(data);
                    HookKeyDown.Invoke(this, e);
                    handled = e.Handled;
                }
                //
                // KeyPress
                //
                if (HookKeyPress != null && wParam == (int) Win32.WM.KEYDOWN)
                {
                    bool shift = ((Win32.GetKeyState((int) Win32.VK.SHIFT) & 0x80) == 0x80 ? true : false);
                    bool caps = ((Win32.GetKeyState((int) Win32.VK.CAPITAL) & 0x01) == 0x01 ? true : false);
                    byte[] state = new byte[256];
                    Win32.GetKeyboardState(state);
                    short lpChar;
                    if (Win32.ToAscii(info.vkCode, info.scanCode, state, out lpChar, info.flags) == 1)
                    {
                        char key = (char) (lpChar & 0xFF); // take only low-order byte
                        if ((shift ^ caps) && Char.IsLetter(key))
                            key = Char.ToUpper(key);
                        KeyPressEventArgs e = new KeyPressEventArgs(key);
                        HookKeyPress.Invoke(this, e);
                        handled = handled || e.Handled;
                    }
                }
                //
                // KeyUp
                //
                if (HookKeyUp != null && (wParam == (int) Win32.WM.KEYUP || wParam == (int) Win32.WM.SYSKEYUP))
                {
                    Keys data = (Keys) info.vkCode;
                    Keys mod = GetModifierKey(data); // get if a modifier key was just released

                    // test the modifier keys for whether they are down
                    bool shift = ((Win32.GetKeyState((int) Win32.VK.SHIFT) & 0x80) == 0x80 ? true : false);
                    bool ctrl = ((Win32.GetKeyState((int) Win32.VK.CONTROL) & 0x80) == 0x80 ? true : false);
                    bool alt = ((Win32.GetKeyState((int) Win32.VK.MENU) & 0x80) == 0x80 ? true : false);

                    // if a modifier key is down, add it to the data
                    if (shift)
                        data |= Keys.Shift;
                    if (ctrl)
                        data |= Keys.Control;
                    if (alt)
                        data |= Keys.Alt;

                    // if a modifier key was just released, turn off saying that key is down
                    if (mod == Keys.LShiftKey || mod == Keys.RShiftKey)
                        data &= ~Keys.Shift;
                    if (mod == Keys.LControlKey || mod == Keys.RControlKey)
                        data &= ~Keys.Control;
                    if (mod == Keys.LMenu || mod == Keys.RMenu)
                        data &= ~Keys.Alt;

                    KeyEventArgs e = new KeyEventArgs(data);
                    HookKeyUp.Invoke(this, e);
                    handled = handled || e.Handled;
                }
            }

			return handled ? new IntPtr(-1) : Win32.CallNextHookEx(_hookID, nCode, wParam, lParam);
        }

        /// <summary>
        /// The virtual key code we receive in the case of a modifier key event, when cast to a Keys type,
        /// can tell us which modifier key was pressed but it isn't a clean value. For example, when the
        /// left shift button is pressed, Keys.LShiftKey is the value, but when the left control key is
        /// pressed, Keys.LControlKey | Keys.LShiftKey is the value. This method therefore tells us
        /// which control key was pressed.
        /// </summary>
        /// <param name="data">The vkCode field from the KBDLLHOOKSTRUCT cast to a Keys enumeration.</param>
        /// <returns>The modifier key that was pressed, if any, or Keys.None.</returns>
        private Keys GetModifierKey(Keys data)
        {
            if (data == Keys.LShiftKey)
                return Keys.LShiftKey;
            else if (data == (Keys.LShiftKey | Keys.RShiftKey))
                return Keys.RShiftKey;
            else if (data == (Keys.LShiftKey | Keys.LControlKey))
                return Keys.LControlKey;
            else if (data == (Keys.LShiftKey | Keys.RShiftKey | Keys.LControlKey | Keys.RControlKey))
                return Keys.RControlKey;
            else if (data == (Keys.LShiftKey | Keys.LMenu))
                return Keys.LMenu;
            else if (data == (Keys.LShiftKey | Keys.RShiftKey | Keys.LMenu | Keys.RMenu))
                return Keys.RMenu;
            else
                return Keys.None;
        }
    }
}
