﻿using System;
using System.Threading;
using System.Diagnostics;
using System.Drawing;
using System.Windows.Forms;
using System.Runtime.InteropServices;
using WobbrockLib.Windows;
using WobbrockLib.Types;
using WobbrockLib.Extensions;

namespace WobbrockLib.Devices
{
	/// <summary>
	/// Offers functions similar to those in WobbrockLib.Extensions but for the mouse, which
	/// is why it is in WobbrockLib.Devices. These static functions enable the user to synthesize
	/// mouse events programmatically.
	/// </summary>
	public static class MouseEx
	{
		#region Take Action

		/// <summary>
		/// An enumerated type for encoding a button action of the mouse.
		/// </summary>
		public enum ButtonAction : int
		{
			None = 0,
			Down = 1,
			Up = 2,
			Click = 3,
			DoubleClick = 4,
			WheelUp = 5,      // roll away from the user
			WheelDown = 6     // roll back toward the user
		}

		/// <summary>
		/// Takes the given mouse button action at the cursor's current location.
		/// </summary>
		/// <param name="action">The action to take.</param>
		/// <param name="button">The button with which to take the action. In the case of a mouse wheel
		/// action, this value should be a uint for the number of mouse wheel detents scrolled, cast to
		/// a MouseButtons.</param>
		public static void TakeAction(ButtonAction action, MouseButtons button)
		{
			switch (action)
			{
				case ButtonAction.Down:
					Down(button);
					break;
				case ButtonAction.Up:
					Up(button);
					break;
				case ButtonAction.Click:
					Click(button);
					break;
				case ButtonAction.DoubleClick:
					DoubleClick(button);
					break;
				case ButtonAction.WheelUp:
					WheelUp((uint) button);
					break;
				case ButtonAction.WheelDown:
					WheelDown((uint) button);
					break;
				default:
					// do nothing
					break;
			}
		}

		/// <summary>
		/// Takes the given mouse button action at the given screen location.
		/// </summary>
		/// <param name="action">The action to take.</param>
		/// <param name="button">The button with which to take the action.</param>
		/// <param name="x">The x-coordinate of the screen location at which to take the action.</param>
		/// <param name="y">The y-coordinate of the screen location at which to take the action.</param>
		public static void TakeAction(ButtonAction action, MouseButtons button, int x, int y)
		{
			MoveTo(x, y);
			Thread.Sleep(55);
			TakeAction(action, button);
		}

		#endregion

		#region Move

		/// <summary>
		/// Moves the mouse to the given screen coordinate.
		/// </summary>
		/// <param name="x">The x-coordinate screen position.</param>
		/// <param name="y">The y-coordinate screen position.</param>
		public static void MoveTo(int x, int y)
		{
			int bits = Win32.SizeOfIntPtr;
			if (bits == 32) // x86
			{
				Win32.INPUTx86 i = Win32.INPUTx86.Empty;
				i.type = Win32.INPUTF.MOUSE;
				i.mi.dwFlags = Win32.MOUSEEVENTF.MOVE | Win32.MOUSEEVENTF.ABSOLUTE;
				PointR pt = FromScreenToAbsolute(x, y);
				i.mi.dx = (int) pt.X;
				i.mi.dy = (int) pt.Y;
				Win32.SendInput(1, ref i, Marshal.SizeOf(i));
			}
			else if (bits == 64) // x64
			{
				Win32.INPUTx64 i = Win32.INPUTx64.Empty;
				i.type = Win32.INPUTF.MOUSE;
				i.mi.dwFlags = Win32.MOUSEEVENTF.MOVE | Win32.MOUSEEVENTF.ABSOLUTE;
				PointR pt = FromScreenToAbsolute(x, y);
				i.mi.dx = (int) pt.X;
				i.mi.dy = (int) pt.Y;
				Win32.SendInput(1, ref i, Marshal.SizeOf(i));
			}
			else
				Debug.Fail("System is neither x86 nor x64 according to Win32.SizeOfIntPtr.");
		}

		/// <summary>
		/// Helper function to convert from screen coordinates to absolute values used by
		/// SendInput for MOUSEEVENTF_MOVE | MOUSEEVENTF_ABSOLUTE. The screen coordinates
		/// must be happed to values between 0 and UInt16.MaxValue (65,535).
		/// </summary>
		/// <param name="x">The x-coordinate of the screen location to map into an absolute value.</param>
		/// <param name="y">The x-coordinate of the screen location to map into an absolute value.</param>
		/// <returns>A point containing the mapped screen coordinate.</returns>
		private static PointR FromScreenToAbsolute(int x, int y)
		{
			Screen screen = Screen.FromPoint(new Point(x, y));
			double newX = GeotrigEx.MapValue(x, 0.0, screen.Bounds.Width, 0.0, UInt16.MaxValue);
			double newY = GeotrigEx.MapValue(y, 0.0, screen.Bounds.Height, 0.0, UInt16.MaxValue);
			return new PointR(
				Math.Round(newX),
				Math.Round(newY)
				);
		}

		/// <summary>
		/// Moves the mouse the given number of pixels from its current position.
		/// </summary>
		/// <param name="dx">The number of pixels to move left, if negative, or right, if positive.</param>
		/// <param name="dy">The number of pixels to move up, if negative, or down, if positive.</param>
		public static void MoveBy(int dx, int dy)
		{
			int bits = Win32.SizeOfIntPtr;
			if (bits == 32) // x86
			{
				Win32.INPUTx86 i = Win32.INPUTx86.Empty;
				i.type = Win32.INPUTF.MOUSE;
				i.mi.dwFlags = Win32.MOUSEEVENTF.MOVE;
				i.mi.dx = dx;
				i.mi.dy = dy;
				Win32.SendInput(1, ref i, Marshal.SizeOf(i));
			}
			else if (bits == 64) // x64
			{
				Win32.INPUTx64 i = Win32.INPUTx64.Empty;
				i.type = Win32.INPUTF.MOUSE;
				i.mi.dwFlags = Win32.MOUSEEVENTF.MOVE;
				i.mi.dx = dx;
				i.mi.dy = dy;
				Win32.SendInput(1, ref i, Marshal.SizeOf(i));
			}
			else Debug.Fail("System is neither x86 nor x64 according to Win32.SizeOfIntPtr.");
		}

		#endregion

		#region Down

		/// <summary>
		/// Performs a mouse down wherever the mouse currently is with the given button.
		/// </summary>
		/// <param name="button">The button that has gone down.</param>
		public static void Down(MouseButtons button)
		{
			Win32.MOUSEEVENTF down;
			uint data = 0u;

			switch (button)
			{
				case MouseButtons.Left:
					down = Win32.MOUSEEVENTF.LEFTDOWN;
					break;
				case MouseButtons.Right:
					down = Win32.MOUSEEVENTF.RIGHTDOWN;
					break;
				case MouseButtons.Middle:
					down = Win32.MOUSEEVENTF.MIDDLEDOWN;
					break;
				case MouseButtons.XButton1:
					down = Win32.MOUSEEVENTF.XDOWN;
					data = (uint) Win32.MOUSEEVENTF.XBUTTON1;
					break;
				case MouseButtons.XButton2:
					down = Win32.MOUSEEVENTF.XDOWN;
					data = (uint) Win32.MOUSEEVENTF.XBUTTON2;
					break;
				default:
					down = Win32.MOUSEEVENTF.LEFTDOWN;
					break;
			}
			Send(down, data);
		}

		#endregion

		#region Up

		/// <summary>
		/// Performs a mouse up wherever the mouse currently is with the given button.
		/// </summary>
		/// <param name="button">The button that has gone up.</param>
		public static void Up(MouseButtons button)
		{
			Win32.MOUSEEVENTF up;
			uint data = 0u;

			switch (button)
			{
				case MouseButtons.Left:
					up = Win32.MOUSEEVENTF.LEFTUP;
					break;
				case MouseButtons.Right:
					up = Win32.MOUSEEVENTF.RIGHTUP;
					break;
				case MouseButtons.Middle:
					up = Win32.MOUSEEVENTF.MIDDLEUP;
					break;
				case MouseButtons.XButton1:
					up = Win32.MOUSEEVENTF.XUP;
					data = (uint) Win32.MOUSEEVENTF.XBUTTON1;
					break;
				case MouseButtons.XButton2:
					up = Win32.MOUSEEVENTF.XUP;
					data = (uint) Win32.MOUSEEVENTF.XBUTTON2;
					break;
				default:
					up = Win32.MOUSEEVENTF.LEFTUP;
					break;
			}
			Send(up, data);
		}

		#endregion

		#region Click

		/// <summary>
		/// Performs a mouse click wherever the mouse currently is with the given button.
		/// </summary>
		/// <param name="button">The button that has clicked.</param>
		public static void Click(MouseButtons button)
		{
			Win32.MOUSEEVENTF down;
			Win32.MOUSEEVENTF up;
			uint data = 0u;

			switch (button)
			{
				case MouseButtons.Left:
					down = Win32.MOUSEEVENTF.LEFTDOWN;
					up = Win32.MOUSEEVENTF.LEFTUP;
					break;
				case MouseButtons.Right:
					down = Win32.MOUSEEVENTF.RIGHTDOWN;
					up = Win32.MOUSEEVENTF.RIGHTUP;
					break;
				case MouseButtons.Middle:
					down = Win32.MOUSEEVENTF.MIDDLEDOWN;
					up = Win32.MOUSEEVENTF.MIDDLEUP;
					break;
				case MouseButtons.XButton1:
					down = Win32.MOUSEEVENTF.XDOWN;
					up = Win32.MOUSEEVENTF.XUP;
					data = (uint) Win32.MOUSEEVENTF.XBUTTON1;
					break;
				case MouseButtons.XButton2:
					down = Win32.MOUSEEVENTF.XDOWN;
					up = Win32.MOUSEEVENTF.XUP;
					data = (uint) Win32.MOUSEEVENTF.XBUTTON2;
					break;
				default:
					down = Win32.MOUSEEVENTF.LEFTDOWN;
					up = Win32.MOUSEEVENTF.LEFTUP;
					break;
			}

			Send(down, data);
			Send(up, data);
		}

		#endregion

		#region Double Click

		/// <summary>
		/// Performs a mouse double-click wherever the mouse currently is with the given button.
		/// </summary>
		/// <param name="button">The button that has double-clicked.</param>
		public static void DoubleClick(MouseButtons button)
		{
			Win32.MOUSEEVENTF down;
			Win32.MOUSEEVENTF up;
			uint data = 0u;

			switch (button)
			{
				case MouseButtons.Left:
					down = Win32.MOUSEEVENTF.LEFTDOWN;
					up = Win32.MOUSEEVENTF.LEFTUP;
					break;
				case MouseButtons.Right:
					down = Win32.MOUSEEVENTF.RIGHTDOWN;
					up = Win32.MOUSEEVENTF.RIGHTUP;
					break;
				case MouseButtons.Middle:
					down = Win32.MOUSEEVENTF.MIDDLEDOWN;
					up = Win32.MOUSEEVENTF.MIDDLEUP;
					break;
				case MouseButtons.XButton1:
					down = Win32.MOUSEEVENTF.XDOWN;
					up = Win32.MOUSEEVENTF.XUP;
					data = (uint) Win32.MOUSEEVENTF.XBUTTON1;
					break;
				case MouseButtons.XButton2:
					down = Win32.MOUSEEVENTF.XDOWN;
					up = Win32.MOUSEEVENTF.XUP;
					data = (uint) Win32.MOUSEEVENTF.XBUTTON2;
					break;
				default:
					down = Win32.MOUSEEVENTF.LEFTDOWN;
					up = Win32.MOUSEEVENTF.LEFTUP;
					break;
			}

			Send(down, data);
			Send(up, data);
			
			Send(down, data);
			Send(up, data);
		}

		#endregion

		#region Wheel

		/// <summary>
		/// Perform a mouse wheel operation that rolls the mouse wheel up, or away from the user. This
		/// can also be thought of as the action necessary to move a scroll bar up on a vertical scrolling
		/// window.
		/// </summary>
		/// <param name="detents">The number of mouse wheel detents, or clicks, that determine the scroll
		/// amount. Most often, this value should be 1u.</param>
		public static void WheelUp(uint detents)
		{
			uint delta = detents * Win32.MOUSEINPUT.WHEEL_DELTA;
			Send(Win32.MOUSEEVENTF.WHEEL, delta);
		}

		/// <summary>
		/// Perform a mouse wheel operation that rolls the mouse wheel down, or back towards the user. This
		/// can also be thought of as the action necessary to move a scroll bar down on a vertical scrolling
		/// window.
		/// </summary>
		/// <param name="detents">The number of mouse wheel detents, or clicks, that determine the scroll
		/// amount. Most often, this value should be 1u.</param>
		/// <remarks>The Windows API wants the mouseData value, a DWORD, to be negative when scrolling down. 
		/// But a DWORD (uint) is unsigned. So we first compute the delta, turn it negative, and then cast
		/// it back to an unsigned integer.</remarks>
		public static void WheelDown(uint detents)
		{
			uint delta = (uint) -(detents * Win32.MOUSEINPUT.WHEEL_DELTA);
			Send(Win32.MOUSEEVENTF.WHEEL, delta);
		}

		#endregion

		#region Send Function

		/// <summary>
		/// Helper function to send the given mouse flags with the given mouse data.
		/// </summary>
		/// <param name="dwFlags">The dwFlags parameter for SendInput.</param>
		/// <param name="mouseData">The mouseData parameter for SendInput.</param>
		private static void Send(Win32.MOUSEEVENTF dwFlags, uint mouseData)
		{
			int bits = Win32.SizeOfIntPtr;
			if (bits == 32) // x86
			{
				Win32.INPUTx86 i = Win32.INPUTx86.Empty;
				i.type = Win32.INPUTF.MOUSE;
				i.mi.dwFlags = dwFlags;
				i.mi.mouseData = mouseData;
				Win32.SendInput(1, ref i, Marshal.SizeOf(i));
			}
			else if (bits == 64) // x64
			{
				Win32.INPUTx64 i = Win32.INPUTx64.Empty;
				i.type = Win32.INPUTF.MOUSE;
				i.mi.dwFlags = dwFlags;
				i.mi.mouseData = mouseData;
				Win32.SendInput(1, ref i, Marshal.SizeOf(i));
			}
			else Debug.Fail("System is neither x86 nor x64 according to Win32.SizeOfIntPtr.");
		}

		#endregion
	}
}
