﻿using System;
using System.Drawing;
using System.Drawing.Imaging;
using System.Runtime.InteropServices;
using System.Windows.Forms;
using WobbrockLib.Types;
using WobbrockLib.Windows;

namespace WobbrockLib.Extensions
{
    public static class DrawingEx
    {
        #region Screen Scraping

        /// <summary>
        /// Takes a screen shot of the screen at the given screen coordinate.
        /// </summary>
        /// <param name="screenPt">A screen coordinate.</param>
        /// <returns>A bitmap of a screen corresponding to the screen at the given point.</returns>
        public static Bitmap Screenshot(Point screenPt)
		{
            Screen screen = Screen.FromPoint(screenPt);
            return Screenshot(screen);
		}

        /// <summary>
        /// Takes a screen shot of the given screen.
        /// </summary>
        /// <param name="screen">A reference to a screen.</param>
        /// <returns>A bitmap of a screen for which to take a screen shot.</returns>
        public static Bitmap Screenshot(Screen screen)
		{
            if (screen == null)
                return null;

            // we need the current display scaling since it may not be 100%
            // and that will affect our bitmap that we capture
            double scale = ControlPanelEx.GetDisplayScaling(screen);

            // create a bitmap capable of holding the scaled screen shot
            Bitmap bmp = new Bitmap(
                (int) (screen.Bounds.Width * scale),
                (int) (screen.Bounds.Height * scale)
                );

            // get a graphics context for drawing to this bitmap
            Graphics g = Graphics.FromImage(bmp);

            // copy the screen pixels to the bitmap
            g.CopyFromScreen(
                screen.Bounds.X, screen.Bounds.Y,  // src
                0, 0,                              // dst
                bmp.Size,                          // size
                CopyPixelOperation.SourceCopy
                );

            // we're done with the graphics context
            g.Dispose();

            return bmp;
        }

        /// <summary>
        /// Scrape pixels directly from the screen at the given rectangular bounds. The screen at
        /// the top-left corner of the rectangle will be the one scraped.
        /// </summary>
        /// <param name="bounds">The rectangle from which to scrape pixels from the screen.</param>
        /// <param name="screenAt">A reference point from the given bounds that determines which screen
        /// is scraped.</param>
        /// <returns>A bitmap of the scraped pixels.</returns>
        public static Bitmap ScrapePixelsFromScreen(Rectangle bounds, ContentAlignment screenAt)
		{
            return ScrapePixelsFromScreen(bounds.Location, bounds.Size, screenAt);
		}

        /// <summary>
        /// Scrape pixels directly from the given screen at the given rectangular bounds.
        /// </summary>
        /// <param name="bounds">The rectangle from which to scrape pixels from the given screen.</param>
        /// <param name="screen">The screen whose pixels to scrape.</param>
        /// <returns>A bitmap of the scraped pixels.</returns>
        public static Bitmap ScrapePixelsFromScreen(Rectangle bounds, Screen screen)
        {
            return ScrapePixelsFromScreen(bounds.Location, bounds.Size, screen);
        }

        /// <summary>
        /// Scape pixels directly from the screen at the given location and size. The screen at the
        /// location coordinate will be the one scraped.
        /// </summary>
        /// <param name="location">The top-left location from which to scrape pixels.</param>
        /// <param name="size">The size extending rightward and downward from the top-left location from 
        /// which to scrape pixels from the screen.</param>
        /// <param name="screenAt">A reference point from the given dimensions that determines which screen
        /// is scraped.</param>
        /// <returns>A bitmap of the scraped pixels.</returns>
        public static Bitmap ScrapePixelsFromScreen(Point location, Size size, ContentAlignment screenAt)
		{
            Screen screen;
            switch (screenAt)
            {
                case ContentAlignment.TopLeft:
                    screen = Screen.FromPoint(location);
                    break;
                case ContentAlignment.TopCenter:
                    screen = Screen.FromPoint(new Point(location.X + size.Width / 2, location.Y));
                    break;
                case ContentAlignment.TopRight:
                    screen = Screen.FromPoint(new Point(location.X + size.Width, location.Y));
                    break;
                case ContentAlignment.MiddleLeft:
                    screen = Screen.FromPoint(new Point(location.X, location.Y + size.Height / 2));
                    break;
                case ContentAlignment.MiddleCenter:
                    screen = Screen.FromPoint(new Point(location.X + size.Width / 2, location.Y + size.Height / 2));
                    break;
                case ContentAlignment.MiddleRight:
                    screen = Screen.FromPoint(new Point(location.X + size.Width, location.Y + size.Height / 2));
                    break;
                case ContentAlignment.BottomLeft:
                    screen = Screen.FromPoint(new Point(location.X, location.Y + size.Height));
                    break;
                case ContentAlignment.BottomCenter:
                    screen = Screen.FromPoint(new Point(location.X + size.Width / 2, location.Y + size.Height));
                    break;
                case ContentAlignment.BottomRight:
                    screen = Screen.FromPoint(new Point(location.X + size.Width, location.Y + size.Height));
                    break;
                default:
                    screen = null;
                    break;
            }
            return ScrapePixelsFromScreen(location, size, screen);
		}

        /// <summary>
        /// Scape pixels directly from the given screen at the given location and size.
        /// </summary>
        /// <param name="location">The top-left location from which to scape pixels from the given screen, 
        /// in screen coordinates.</param>
        /// <param name="size">The size extending rightward and downward from the top-left location from 
        /// which to scrape pixels from the given screen.</param>
        /// <param name="screen">The screen whose pixels to scrape.</param>
        /// <returns>A bitmap of the scraped pixels.</returns>
        /// <remarks>
        /// In a multimon setup, the display scaling may differ on different monitors. This function is robust 
        /// to differences in display scaling.
        /// </remarks>
        public static Bitmap ScrapePixelsFromScreen(Point location, Size size, Screen screen)
		{
            if (screen == null)
                return null;

            // get the display scaling factor, e.g., 1.00 = 100%, 1.50 = 150%
            double scale = ControlPanelEx.GetDisplayScaling(screen);

            // translate the location by the top-left corner of the screen bounds
            // so that the origin of the screen is effectively (0,0) for this 
            // translated location. only matters in a multimon setup.
            location.X -= screen.Bounds.X;
            location.Y -= screen.Bounds.Y;

            // now scale the location on its (0,0)-origin screen by the display scaling factor
            double scaledX = location.X * scale;
            double scaledY = location.Y * scale;

            // now map that scaled point from its (0,0)-origin screen back to its actual screen
            PointR newPt = GeotrigEx.MapPoint(
                new PointR(scaledX, scaledY),
                new RectangleR(0, 0, screen.Bounds.Width, screen.Bounds.Height),
                screen.Bounds
                );

            // next create a bitmap of the appropriate scaled size to hole the scraped pixels
            Bitmap bmp = new Bitmap((int) (size.Width * scale),  (int) (size.Height * scale));
            
            // finally scrape the pixels from the screen!
            Graphics g = Graphics.FromImage(bmp);
			g.CopyFromScreen(
				(Point) newPt,  // src
				Point.Empty,    // dst
				bmp.Size,       // size
				CopyPixelOperation.SourceCopy
				);
            g.Dispose();
            
            return bmp;
		}

        /// <summary>
        /// Scrapes pixels from the given bitmap at the given bounds and returns them as a bitmap of
        /// their own.
        /// </summary>
        /// <param name="src">The source bitmap to scrape.</param>
        /// <param name="bounds">The rectangular bounds from which to scrape the pixels of the given
        /// bitmap.</param>
        /// <returns>A rectangular bitmap of pixels from the given source bitmap.</returns>
        public static Bitmap ScrapePixelsFromBitmap(Bitmap src, Rectangle bounds)
        {
            return ScrapePixelsFromBitmap(src, bounds.Location, bounds.Size);
        }

        /// <summary>
        /// Scrapes pixels from the given bitmap at the given location and size and returns them as
        /// a bitmap of their own.
        /// </summary>
        /// <param name="src">The source bitmap to scrape.</param>
        /// <param name="location">The (x,y) location within the source bitmap defining the top-left
        /// corner of the pixels to scrape.</param>
        /// <param name="size">The width and height within the source bitmap defining the size of
        /// the area to scrape.</param>
        /// <returns>A rectangular bitmap of pixels from the given source bitmap.</returns>
        public static Bitmap ScrapePixelsFromBitmap(Bitmap src, Point location, Size size)
		{
            if (src == null)
                return null;

            // create a bitmap to hold the scraped pixels
            Bitmap dst = new Bitmap(size.Width, size.Height);

            // get a graphics context for drawing to this bitmap
            Graphics g = Graphics.FromImage(dst);

            // draw pixels from the source bitmap to the destination bitmap
            g.DrawImage(
                src,
                new Rectangle(0, 0, dst.Width, dst.Height), // dst
                new Rectangle(location.X, location.Y, size.Width, size.Height), // src
                GraphicsUnit.Pixel
                );

            // we're done with the graphics context
            g.Dispose();

            return dst;
		}

        #endregion

        #region Blurring Techniques

        /// <summary>
        /// Blurs a given bitmap using the Box filter blurring technique. The larger the dimension
        /// given, the more the blurring. Runs much faster than the Gaussian technique.
        /// </summary>
        /// <param name="bmp">The bitmap to blur.</param>
        /// <param name="dimension">The dimension of the box filter, with minimum of 3 for a
        /// 3 x 3 filter matrix.</param>
        /// <returns>A blurred image.</returns>
        /// <see cref="https://softwarebydefault.com/2013/06/09/image-blur-filters/"/>
        public static Bitmap BoxBlur(Bitmap bmp, int dimension)
		{
            if (dimension < 3)
                dimension = 3;
            if ((dimension % 2) == 0)
                dimension++;

            double[,] filter = new double[dimension, dimension];

            double cell = 1.0 / (dimension * dimension);

            for (int i = 0; i < dimension; i++) // down the rows
			{
                for (int j = 0; j < dimension; j++) // across the cols
				{
                    filter[i, j] = cell;
				}
			}

            return Convolve(bmp, filter);
		}

		/// <summary>
		/// Blurs a given bitmap using a Gaussian blurring filter by the given amount. The larger the
		/// standard deviation, the more blur applied.
		/// </summary>
		/// <param name="bmp">The bitmap to blur.</param>
		/// <param name="stdev">An amount to blur defined as an odd integer 3 or greater. If the value
		/// is less than 3, it will be upgraded to 3. If the value is even, it will be increased by 1.</param>
		/// <returns>A bitmap blurred by a given amount.</returns>
		public static Bitmap GaussianBlur(Bitmap bmp, int stdev)
        {
            // get the 2-D Gaussian kernel filter
            double[,] filter = GaussianFilter(stdev);

            // use the filter to convolve the image pixels
            return Convolve(bmp, filter);
        }

        /// <summary>
        /// Creates a 2-D Gaussian filter to be used as a convolution matrix for blurring images. 
        /// </summary>
        /// <param name="stdev">The standard deviation for parameterizing the 2-D Gaussian equation.
        /// Greater standard deviations result in larger filter matrices and more blurring.</param>
        /// <returns>A 2-D array representing a matrix populated with discretized Gaussian values.</returns>
        /// <see cref="https://homepages.inf.ed.ac.uk/rbf/HIPR2/gsmooth.htm"/>
        public static double[,] GaussianFilter(int stdev)
        {
            if (stdev < 1)
                stdev = 1;

            int size = 3 * stdev * 2 + 1;
            double[,] kernel2D = new double[size, size];

            for (int i = 0, y = -size / 2; i < size; i++, y++) // down the rows
            {
                for (int j = 0, x = -size / 2; j < size; j++, x++) // across the cols
                {
                    kernel2D[i, j] = (1.0 / (2.0 * Math.PI * stdev * stdev)) * Math.Pow(Math.E, -((x * x + y * y) / (2.0 * stdev * stdev)));
                }
            }

            return kernel2D;
        }

        /// <summary>
        /// Applies a 2-D convolution matrix filter to a bitmap, returning the resulting bitmap.
        /// </summary>
        /// <param name="bmp">The bitmap whose pixels to convolve with the given filter.</param>
        /// <param name="filter">The filter matrix to use to convolve the given bitmap with.</param>
        /// <returns>A bitmap affected by the given convolution filter matrix.</returns>
        /// <see cref="https://softwarebydefault.com/2013/06/09/image-blur-filters/"/>
        public static Bitmap Convolve(Bitmap bmp, double[,] filter)
        {
            BitmapData bmpData = bmp.LockBits(new Rectangle(0, 0, bmp.Width, bmp.Height), ImageLockMode.ReadOnly, PixelFormat.Format32bppArgb);

            byte[] pixelBuffer = new byte[bmpData.Stride * bmpData.Height];
            byte[] resultBuffer = new byte[bmpData.Stride * bmpData.Height];

            Marshal.Copy(bmpData.Scan0, pixelBuffer, 0, pixelBuffer.Length);
            bmp.UnlockBits(bmpData);

            double blue = 0.0;
            double green = 0.0;
            double red = 0.0;

            int filterWidth = filter.GetLength(1);
            int filterHeight = filter.GetLength(0);

            int filterOffset = (filterWidth - 1) / 2;
            int calcOffset = 0;
            int byteOffset = 0;

            for (int offsetY = filterOffset; offsetY < bmp.Height - filterOffset; offsetY++)
            {
                for (int offsetX = filterOffset; offsetX < bmp.Width - filterOffset; offsetX++)
                {
                    blue = 0.0;
                    green = 0.0;
                    red = 0.0;

                    byteOffset = offsetY * bmpData.Stride + offsetX * 4;

                    for (int filterY = -filterOffset; filterY <= filterOffset; filterY++)
                    {
                        for (int filterX = -filterOffset; filterX <= filterOffset; filterX++)
                        {
                            calcOffset = byteOffset + (filterX * 4) + (filterY * bmpData.Stride);

                            blue += (pixelBuffer[calcOffset + 0]) * filter[filterY + filterOffset, filterX + filterOffset];
                            green += (pixelBuffer[calcOffset + 1]) * filter[filterY + filterOffset, filterX + filterOffset];
                            red += (pixelBuffer[calcOffset + 2]) * filter[filterY + filterOffset, filterX + filterOffset];
                        }
                    }

                    blue = (blue > 255 ? 255 : (blue < 0 ? 0 : blue));
                    green = (green > 255 ? 255 : (green < 0 ? 0 : green));
                    red = (red > 255 ? 255 : (red < 0 ? 0 : red));

                    resultBuffer[byteOffset + 0] = (byte) blue;
                    resultBuffer[byteOffset + 1] = (byte) green;
                    resultBuffer[byteOffset + 2] = (byte) red;
                    resultBuffer[byteOffset + 3] = (byte) 0xFF;
                }
            }

            // make the result bitmap
            Bitmap bmpResult = new Bitmap(bmp.Width, bmp.Height);
            BitmapData bmpResultData = bmpResult.LockBits(new Rectangle(0, 0, bmpResult.Width, bmpResult.Height), ImageLockMode.WriteOnly, PixelFormat.Format32bppArgb);

            Marshal.Copy(resultBuffer, 0, bmpResultData.Scan0, resultBuffer.Length);
            bmpResult.UnlockBits(bmpResultData);

            return bmpResult;
        }

        #endregion

        #region "Fast" Blurring Techniques

        /// <summary>
        /// Fast version of the Box blurring function.
        /// </summary>
        /// <param name="bmp">Bitmap to blur.</param>
        /// <param name="radius">The size of the box. Greater size results in more blurring.</param>
        /// <returns>A blurred bitmap using the box blur technqiue.</returns>
        /// <remarks>Also a helper function for the FastGaussianBlur.</remarks>
        /// <see cref="https://medium.com/@RoardiLeone/fast-image-blurring-algorithm-photoshop-level-w-c-code-87516d5cee87"/>
        public static Bitmap FastBoxBlur(Bitmap bmp, int radius)
        {
            int kSize = radius;
            if (kSize % 2 == 0)
                kSize++;

            Bitmap Hblur = (Bitmap) bmp.Clone();

            float Avg = 1f / kSize;

            for (int j = 0; j < bmp.Height; j++)
            {

                float[] hSum = new float[] { 0f, 0f, 0f, 0f };
                float[] iAvg = new float[] { 0f, 0f, 0f, 0f };

                for (int x = 0; x < kSize; x++)
                {
                    Color tmpColor = bmp.GetPixel(x, j);
                    hSum[0] += tmpColor.A;
                    hSum[1] += tmpColor.R;
                    hSum[2] += tmpColor.G;
                    hSum[3] += tmpColor.B;
                }

                iAvg[0] = hSum[0] * Avg;
                iAvg[1] = hSum[1] * Avg;
                iAvg[2] = hSum[2] * Avg;
                iAvg[3] = hSum[3] * Avg;

                for (int i = 0; i < bmp.Width; i++)
                {
                    if (i - kSize / 2 >= 0 && i + 1 + kSize / 2 < bmp.Width)
                    {
                        Color tmp_pColor = bmp.GetPixel(i - kSize / 2, j);
                        hSum[0] -= tmp_pColor.A;
                        hSum[1] -= tmp_pColor.R;
                        hSum[2] -= tmp_pColor.G;
                        hSum[3] -= tmp_pColor.B;

                        Color tmp_nColor = bmp.GetPixel(i + 1 + kSize / 2, j);
                        hSum[0] += tmp_nColor.A;
                        hSum[1] += tmp_nColor.R;
                        hSum[2] += tmp_nColor.G;
                        hSum[3] += tmp_nColor.B;

                        iAvg[0] = hSum[0] * Avg;
                        iAvg[1] = hSum[1] * Avg;
                        iAvg[2] = hSum[2] * Avg;
                        iAvg[3] = hSum[3] * Avg;
                    }
                    Hblur.SetPixel(i, j, Color.FromArgb((int) iAvg[0], (int) iAvg[1], (int) iAvg[2], (int) iAvg[3]));
                }
            }

            Bitmap total = (Bitmap) Hblur.Clone();
            for (int i = 0; i < Hblur.Width; i++)
            {
                float[] tSum = new float[] { 0f, 0f, 0f, 0f };
                float[] iAvg = new float[] { 0f, 0f, 0f, 0f };
                for (int y = 0; y < kSize; y++)
                {
                    Color tmpColor = Hblur.GetPixel(i, y);
                    tSum[0] += tmpColor.A;
                    tSum[1] += tmpColor.R;
                    tSum[2] += tmpColor.G;
                    tSum[3] += tmpColor.B;
                }

                iAvg[0] = tSum[0] * Avg;
                iAvg[1] = tSum[1] * Avg;
                iAvg[2] = tSum[2] * Avg;
                iAvg[3] = tSum[3] * Avg;

                for (int j = 0; j < Hblur.Height; j++)
                {
                    if (j - kSize / 2 >= 0 && j + 1 + kSize / 2 < Hblur.Height)
                    {
                        Color tmp_pColor = Hblur.GetPixel(i, j - kSize / 2);
                        tSum[0] -= tmp_pColor.A;
                        tSum[1] -= tmp_pColor.R;
                        tSum[2] -= tmp_pColor.G;
                        tSum[3] -= tmp_pColor.B;

                        Color tmp_nColor = Hblur.GetPixel(i, j + 1 + kSize / 2);
                        tSum[0] += tmp_nColor.A;
                        tSum[1] += tmp_nColor.R;
                        tSum[2] += tmp_nColor.G;
                        tSum[3] += tmp_nColor.B;

                        iAvg[0] = tSum[0] * Avg;
                        iAvg[1] = tSum[1] * Avg;
                        iAvg[2] = tSum[2] * Avg;
                        iAvg[3] = tSum[3] * Avg;
                    }
                    total.SetPixel(i, j, Color.FromArgb((int) iAvg[0], (int) iAvg[1], (int) iAvg[2], (int) iAvg[3]));
                }
            }
            return total;
        }

        /// <summary>
        /// A fast version of the Gaussian blur procedure that relies upon successive Box blurs.
        /// </summary>
        /// <param name="bmp">The source bitmap to blur.</param>
        /// <param name="stdev">The standard deviation of the Gaussian. Higher values result in more blurring.</param>
        /// <returns>A blurred bitmap.</returns>
        /// <see cref="https://medium.com/@RoardiLeone/fast-image-blurring-algorithm-photoshop-level-w-c-code-87516d5cee87"/>
        public static Bitmap FastGaussianBlur(Bitmap bmp, int stdev)
        {
            int[] boxes = BoxesForGaussian(stdev, 3);
            Bitmap img_1 = FastBoxBlur(bmp, boxes[0]);
            Bitmap img_2 = FastBoxBlur(img_1, boxes[1]);
            Bitmap img_3 = FastBoxBlur(img_2, boxes[2]);
            return img_3;
        }

        /// <summary>
        /// Helper function for the FastGaussianBlur.
        /// </summary>
        /// <param name="stdev">The standard deviation of the Gaussian function.</param>
        /// <param name="n">Should be 3. See the link below.</param>
        /// <returns>The three boxes used for the fast Gaussian blur.</returns>
        /// <see cref="https://medium.com/@RoardiLeone/fast-image-blurring-algorithm-photoshop-level-w-c-code-87516d5cee87"/>
        private static int[] BoxesForGaussian(double stdev, int n)
        {

            double wIdeal = Math.Sqrt((12 * stdev * stdev / n) + 1);
            double wl = Math.Floor(wIdeal);

            if (wl % 2 == 0)
                wl--;
            double wu = wl + 2;

            double mIdeal = (12 * stdev * stdev - n * wl * wl - 4 * n * wl - 3 * n) / (-4 * wl - 4);
            double m = Math.Round(mIdeal);

            int[] sizes = new int[n];
            for (int i = 0; i < n; i++)
            {
                if (i < m)
                {
                    sizes[i] = (int) wl;
                }
                else
                {
                    sizes[i] = (int) wu;
                }
            }
            return sizes;
        }

        #endregion
    }
}
