﻿using System;
using System.Collections.Generic;
using WobbrockLib.Types;

namespace WobbrockLib.Extensions
{
    /// <summary>
    /// Statistical utility functions.
    /// </summary>
    public static class StatsEx
    {
        /// <summary>
        /// Computes the minimum of a series of numbers.
        /// </summary>
        /// <param name="d">An array of values.</param>
        /// <returns>The minimum value of the array.</returns>
        public static double Min(double[] d)
        {
            double min = double.NaN;
            for (int i = 0; i < d.Length; i++)
            {
                if (d[i] < min || Double.IsNaN(min))
                    min = d[i];
            }
            return min;
        }

        /// <summary>
        /// Computes the minimum of three values.
        /// </summary>
        /// <param name="d1">The first value.</param>
        /// <param name="d2">The second value.</param>
        /// <param name="d3">The third value.</param>
        /// <returns>The minimum of the three values.</returns>
        public static double Min3(double d1, double d2, double d3)
        {
            double m = Math.Min(d1, d2);
            return (m < d3) ? m : d3;
        }

        /// <summary>
        /// Computes the maximum of a series of numbers.
        /// </summary>
        /// <param name="d">An array of values.</param>
        /// <returns>The maximum value of the array.</returns>
        public static double Max(double[] d)
        {
            double max = double.NaN;
            for (int i = 0; i < d.Length; i++)
            {
                if (d[i] > max || Double.IsNaN(max))
                    max = d[i];
            }
            return max;
        }

        /// <summary>
        /// Computes the maximum of three values.
        /// </summary>
        /// <param name="d1">The first value.</param>
        /// <param name="d2">The second value.</param>
        /// <param name="d3">The third value.</param>
        /// <returns>The maximum of the three values.</returns>
        public static double Max3(double d1, double d2, double d3)
        {
            double m = Math.Max(d1, d2);
            return (m > d3) ? m : d3;
        }

        /// <summary>
        /// Computes the sum of the values in the given array.
        /// </summary>
        /// <param name="d">An array of double-precision floating point values.</param>
        /// <returns>The arithmetic sum.</returns>
        public static double Sum(double[] d)
        {
            double x = 0.0;
            for (int i = 0; i < d.Length; i++)
            {
                x += d[i];
            }
            return x;
        }

        /// <summary>
        /// Computes the mean of the values in the given array.
        /// </summary>
        /// <param name="d">An array of double-precision floating point values.</param>
        /// <returns>The arithmetic mean, or double.NaN if d is of length zero.</returns>
        public static double Mean(double[] d)
        {
            double x = 0.0;
            for (int i = 0; i < d.Length; i++)
            {
                x += d[i];
            }
            return x / d.Length;
        }

        /// <summary>
        /// Compute the weighted mean of an array of values <i>d</i> using a corresponding array
        /// of weights <i>w</i>, where value <i>d[i]</i> is weighted by weight <i>w[i]</i>.
        /// </summary>
        /// <param name="d">The array of values.</param>
        /// <param name="w">The array of weights.</param>
        /// <returns>The weighted mean of the array of values, or double.NaN if d is of length zero.</returns>
        /// <url>http://pygsl.sourceforge.net/reference/pygsl/node36.html</url>
        public static double Mean(double[] d, double[] w)
        {
            double wsum = 0.0; // weighted sum of values 
            double sumw = 0.0; // sum of used weights
            for (int i = 0; i < d.Length; i++)
            {
                wsum += d[i] * w[i]; // weight each value d[i] by w[i]
                sumw += w[i]; // keep sum of used weights
            }
            return wsum / sumw;
        }

        /// <summary>
        /// Computes the standard deviation of the sample values in the given array.
        /// </summary>
        /// <param name="d">An array of double-precision floating point values.</param>
        /// <returns>The sample standard deviation, or double.NaN if d is of length zero.</returns>
        /// <remarks>The population standard deviation is similar but does not contain the "- 1"
        /// in the denominator of the final quotient.</remarks>
        public static double StdDev(double[] d)
        {
            if (d.Length == 0)
                return double.NaN;

            double ssd = 0.0;
            double x_bar = Mean(d);

            for (int i = 0; i < d.Length; i++)
            {
                double dx = d[i] - x_bar;
                ssd += dx * dx; // square and add
            }

            return Math.Sqrt(ssd / (d.Length - 1));
        }

        /// <summary>
        /// Compute the weighted standard deviation of an array of values <i>d</i> using a corresponding 
        /// array of weights <i>w</i>, where value <i>d[i]</i> is weighted by weight <i>w[i]</i>.
        /// </summary>
        /// <param name="d">The array of values.</param>
        /// <param name="w">The array of weights.</param>
        /// <returns>The weighted sample standard deviation of the array of values, or double.NaN if d is of length zero.</returns>
        /// <url>http://pygsl.sourceforge.net/reference/pygsl/node36.html</url>
        public static double StdDev(double[] d, double[] w)
        {
            double wmean = Mean(d, w);
            return StdDev(d, w, wmean);
        }

        /// <summary>
        /// Compute the weighted standard deviation of an array of values <i>d</i> using a corresponding 
        /// array of weights <i>w</i>, where value <i>d[i]</i> is weighted by weight <i>w[i]</i>. In this
        /// overload, the weighted mean is not computed but given directly.
        /// </summary>
        /// <param name="d">The array of values.</param>
        /// <param name="w">The array of weights.</param>
        /// <param name="wmean">The weighted mean to use.</param>
        /// <returns>The weighted sample standard deviation of the array of values, or double.NaN if d is of length zero.</returns>
        /// <url>http://pygsl.sourceforge.net/reference/pygsl/node36.html</url>
        public static double StdDev(double[] d, double[] w, double wmean)
        {
            double sumw = 0.0; // sum of weights
            double sumsqw = 0.0; // sum of squared weights
            double sumwsqdev = 0.0; // sum of weighted squared deviations

            for (int i = 0; i < d.Length; i++)
            {
                double dx = d[i] - wmean; // deviation
                sumwsqdev += w[i] * (dx * dx); // sum of weighted squared deviations
                sumsqw += w[i] * w[i]; // keep sum of squared weights
                sumw += w[i]; // keep sum of weights
            }

            double variance = (sumw / (sumw * sumw - sumsqw)) * sumwsqdev;
            return Math.Sqrt(variance);
        }

        /// <summary>
        /// Computes the standard deviation of the given points about their centroid. This is
        /// analogous to a bivariate (2D) standard deviation function. Intuitively, it conveys 
        /// how spread out are points about their centroid.
        /// </summary>
        /// <param name="points">A list of points.</param>
        /// <returns>The standard deviation of points about their centroid, or double.NaN if d is of length zero.</returns>
        public static double StdDev2D(List<PointR> points)
        {
            if (points.Count == 0)
                return double.NaN;

            double ssd = 0.0;
            PointR c = GeotrigEx.Centroid(points); // (x_bar, y_bar)

            for (int i = 0; i < points.Count; i++)
            {
                double dx = GeotrigEx.Distance(c, points[i]);
                ssd += dx * dx; // square and add
            }

            return Math.Sqrt(ssd / (points.Count - 1));
        }

        /// <summary>
        /// Computes the weighted standard deviation of the given points about their centroid. This is
        /// analogous to a bivariate (2D) weighted standard deviation function. Intuitively, it conveys 
        /// how spread out are points about their centroid, weighted by the given weights.
        /// </summary>
        /// <param name="points">A list of points.</param>
        /// <param name="w">The array of weights.</param>
        /// <returns>The weighted standard deviation of points about their centroid, or double.NaN if points is of length zero.</returns>
        public static double StdDev2D(List<PointR> points, double[] w)
        {
            PointR wcentroid = GeotrigEx.Centroid(points, w);
            return StdDev2D(points, w, wcentroid);
        }

        /// <summary>
        /// Computes the weighted standard deviation of the given points about their centroid. This is
        /// analogous to a bivariate (2D) weighted standard deviation function. Intuitively, it conveys 
        /// how spread out are points about their centroid, weighted by the given weights. In this version
        /// of the function, the weighted centroid is provided by the caller.
        /// </summary>
        /// <param name="points">A list of points.</param>
        /// <param name="w">The array of weights.</param>
        /// <param name="wcentroid">The weighted centroid provided.</param>
        /// <returns>The weighted standard deviation of points about their centroid, or double.NaN if points is of length zero.</returns>
        public static double StdDev2D(List<PointR> points, double[] w, PointR wcentroid)
        {
            double sumw = 0.0; // sum of weights
            double sumsqw = 0.0; // sum of squared weights
            double sumwsqdev = 0.0; // sum of weighted squared deviations

            for (int i = 0; i < points.Count; i++)
            {
                double dx = GeotrigEx.Distance(points[i], wcentroid); // deviation
                sumwsqdev += w[i] * (dx * dx); // sum of weighted squared deviations
                sumsqw += w[i] * w[i]; // keep sum of squared weights
                sumw += w[i]; // keep sum of weights
            }

            double variance = (sumw / (sumw * sumw - sumsqw)) * sumwsqdev;
            return Math.Sqrt(variance);
        }

        /// <summary>
        /// Returns the maximum standard deviation for <i>n</i> values and an expected mean.
        /// </summary>
        /// <param name="n">The number of values.</param>
        /// <param name="mean">The expected mean value.</param>
        /// <returns>The maximum possible standard deviation. If n &lt; 2, <b>double.PositiveInfinity</b> is returned.</returns>
        /// <seealso cref="http://www.blackwell-synergy.com/doi/abs/10.1111/j.1467-9639.2005.00206.x">http://www.blackwell-synergy.com/doi/abs/10.1111/j.1467-9639.2005.00206.x</seealso>
        public static double MaxStdDev(int n, double mean)
        {
            if (n >= 2)
            {
                if ((n % 2) == 0) // even n
                {
                    return mean * Math.Sqrt(
                        ((double) n) / (n - 1)
                        );
                }
                else // odd n
                {
                    return mean * Math.Sqrt(
                        (n + 1) / ((double) n)
                        );
                }
            }
            else return double.PositiveInfinity;
        }

        /// <summary>
        /// Normalizes a vector of weights such that the contents of the vector sum to 1.0.
        /// </summary>
        /// <param name="w">The given weights to normalize.</param>
        public static void Normalize(double[] w)
        {
            double sumw = 0.0;
            for (int i = 0; i < w.Length; i++)
            {
                sumw += w[i];
            }
            for (int i = 0; i < w.Length; i++)
            {
                w[i] /= sumw;
            }
        }

        /// <summary>
        /// Compute the slope of a regression line for a set of (x,y) values.
        /// </summary>
        /// <param name="x">The x-values used to predict y-value outcomes.</param>
        /// <param name="y">The y-value outcomes. It is assumed to have as many values as the x-value array.</param>
        /// <returns>The slope of a regression line fitted to these values.</returns>
        /// <remarks>
        /// <para>
        /// Regression line coefficients are computed using the least squares method.  Typically, 
        /// the x-values are controlled (independent) and the y-values are measured (dependent).  
        /// In computing the regression line, the y-values are "regressed on" the x-values.  
        /// The regression line equation predicts y as a linear function of x.
        /// </para>
        /// <para>Shamelessly taken from I. Scott MacKenzie's MyUtil.java class, with permission.</para>
        /// </remarks>
        public static double Slope(double[] x, double[] y)
        {
            double sx2 = 0.0;
            double sx = 0.0;
            double sxy = 0.0;
            double sy = 0.0;

            for (int i = 0; i < x.Length; i++)
            {
                sx2 += x[i] * x[i];
                sx += x[i];
                sxy += x[i] * y[i];
                sy += y[i];
            }

            double sxx = x.Length * sx2 - sx * sx;
            sxy = x.Length * sxy - sx * sy;

            return sxy / sxx;
        }

        /// <summary>
        /// Compute the y-axis intercept of a regression line for a set of (x,y) values.
        /// </summary>
        /// <param name="x">The x-values used to predict y-value outcomes.</param>
        /// <param name="y">The y-value outcomes. It is assumed to have as many values as the x-value array.</param>
        /// <returns>The y-axis intercept of a regression line fitted to these values.</returns>
        /// <remarks>
        /// <para>
        /// Regression line coefficients are computed using the least squares method.  Typically, 
        /// the x-values are controlled (independent) and the y-values are measured (dependent).  
        /// In computing the regression line, the y-values are "regressed on" the x-values.  
        /// The regression line equation predicts y as a linear function of x.
        /// </para>
        /// <para>Shamelessly taken from I. Scott MacKenzie's MyUtil.java class, with permission.</para>
        /// </remarks>
        public static double Intercept(double[] x, double[] y)
        {
            return Mean(y) - Slope(x, y) * Mean(x); // == y - mx
        }

        /// <summary>
        /// Calculate the covariance between two sets of values.
        /// </summary>
        /// <param name="d1">A set of values.</param>
        /// <param name="d2">A set of values. Must be of the same length as d1.</param>
        /// <returns>The covariance between two sets of values.</returns>
        /// <remarks>
        /// <para>
        /// The covariance is a measure of association in many problems in physical sciences and 
        /// engineering. It is an adequate measure as long as the scales (means and variances) of the
        /// variables are not arbitrary. But, most variables in social and behavioral sciences are 
        /// measured on an arbitrary scale; hence, correlation coefficients are generally preferred 
        /// to covariance as measures of relationship.
        /// </para>
        /// <para>Shamelessly taken from I. Scott MacKenzie's MyUtil.java class, with permission.</para>
        /// </remarks>
        public static double Covariance(double[] d1, double[] d2)
        {
            double cov = 0.0;
            double m1 = Mean(d1);
            double m2 = Mean(d2);
            for (int i = 0; i < d1.Length; ++i)
            {
                cov += (d1[i] - m1) * (d2[i] - m2);
            }
            return cov / (d1.Length - 1);
        }

        /// <summary>
        /// Compute the Pearson coefficient of correlation, usually termed 'r'. This value indicates
        /// the degree and direction of the association of two variables.
        /// </summary>
        /// <param name="d1">A set of values.</param>
        /// <param name="d2">A set of values. Must be of the same length as d1.</param>
        /// <returns>The Pearson correlation coefficient 'r' between the two sets of values.</returns>
        /// <para>Shamelessly taken from I. Scott MacKenzie's MyUtil.java class, with permission.</para>
        public static double Pearson(double[] d1, double[] d2)
        {
            double cov = Covariance(d1, d2);
            double sd1 = StdDev(d1);
            double sd2 = StdDev(d2);
            return cov / (sd1 * sd2);
        }

        /// <summary>
        /// Calculates the y-value corresponding to the given x-value of a normal distribution with the given
        /// mean and standard deviation.
        /// </summary>
        /// <param name="x">The x-value for which to calculate the y-value.</param>
        /// <param name="mean">The mean of the normal distribution.</param>
        /// <param name="stdev">The standard deviation of the normal distribution.</param>
        /// <returns>The y-value for the given x-value of a normal distribution.</returns>
        /// <seealso cref="https://en.wikipedia.org/wiki/Normal_distribution"/>
        /// <see cref="https://www.desmos.com/calculator"/>
        public static double NormalDistribution(double x, double mean, double stdev)
        {
            double v = stdev * stdev; // variance
            double b = Math.Sqrt(2.0 * Math.PI * v);
            double m = (x - mean) * (x - mean); // (x - mu)^2
            double p = (m / (2.0 * v));
            return (1.0 / b) * Math.Exp(-p);
        }

        /// <summary>
        /// Calculates the y-value corresponding to the given x-value of an exponential distribution with
        /// the given lambda parameter.
        /// </summary>
        /// <param name="x">A nonnegative x-value for which to calculate the y-value. Negative x-values
        /// result in zero y-values.</param>
        /// <param name="lambda">The rate parameter defining the exponential distribution.</param>
        /// <returns>The y-value for the given x-value of an exponential distribution.</returns>
        /// <seealso cref="https://en.wikipedia.org/wiki/Exponential_distribution"/>
        /// <see cref="https://www.desmos.com/calculator"/>
        public static double ExponentialDistribution(double x, double lambda)
        {
            if (x < 0.0)
                return 0.0;
            else
                return lambda * Math.Exp(-lambda * x);
        }

        /// <summary>
        /// Calculates the area under he standard normal distribution using the approximation from Shah (1985).
        /// This is used in calculating probabilites based on z-scores.
        /// </summary>
        /// <param name="z">The x-axis value used to calculate the area from (0, <i>z</i>).</param>
        /// <returns>The area under the standard normal curve from (0, <i>z</i>). If <i>z</i> is negative,
        /// then -1.0 is returned as an error value.</returns>
        public static double ShahNormal(double z)
        {
            double ans = -1.0;
            if (0.0 <= z && z <= 2.2)
            {
                ans = z * (4.4 - z) / 10.0;
            }
            else if (2.2 < z && z < 2.6)
            {
                ans = 0.49;
            }
            else if (2.6 <= z)
            {
                ans = 0.50;
            }
            return ans;
        }

        /// <summary>
        /// Returns the cumulative distribution function (CDF) of the standard normal distribution 
        /// evaluated at a given value.
        /// </summary>
        /// <param name="z">A position on the x-axis.</param>
        /// <returns>The standard normal cumulative distribution function evaluated at <c>z</c>.</returns>
        /// <remarks>
        /// <para>The value of the cumulative distribution function at a point <c>z</c> is
        /// the probability that the value of a random variable having the standard normal density is
        /// less than or equal to <c>z</c>.</para>
        /// <para>This algorithm is ported from dcdflib: Cody, W.D. (1993). "ALGORITHM 715: 
        /// SPECFUN - A Portable FORTRAN Package of Special Function Routines and Test Drivers."
        /// ACM Transactions on Mathematical Software 19 (1), pp. 22-32. (March 1993 issue.)</para>
        /// </remarks>
        /// <see cref="http://en.wikipedia.org/wiki/Cumulative_distribution_function"/>
        public static double CDF(double mean, double stdev, double z)
        {
            double[] a = { 2.2352520354606839287e00, 1.6102823106855587881e02, 1.0676894854603709582e03, 1.8154981253343561249e04, 6.5682337918207449113e-2 };
            double[] b = { 4.7202581904688241870e01, 9.7609855173777669322e02, 1.0260932208618978205e04, 4.5507789335026729956e04 };
            double[] c = { 3.9894151208813466764e-1, 8.8831497943883759412e00, 9.3506656132177855979e01, 5.9727027639480026226e02, 2.4945375852903726711e03, 6.8481904505362823326e03, 1.1602651437647350124e04, 9.8427148383839780218e03, 1.0765576773720192317e-8 };
            double[] d = { 2.2266688044328115691e01, 2.3538790178262499861e02, 1.5193775994075548050e03, 6.4855582982667607550e03, 1.8615571640885098091e04, 3.4900952721145977266e04, 3.8912003286093271411e04, 1.9685429676859990727e04 };
            double[] p = { 2.1589853405795699e-1, 1.274011611602473639e-1, 2.2235277870649807e-2, 1.421619193227893466e-3, 2.9112874951168792e-5, 2.307344176494017303e-2 };
            double[] q = { 1.28426009614491121e00, 4.68238212480865118e-1, 6.59881378689285515e-2, 3.78239633202758244e-3, 7.29751555083966205e-5 };

            int i;
            double del;
            double tmp;
            double xden;
            double xnum;
            double xsq;
            double result;
            double ccum;

            double arg = (z - mean) / stdev;
            double min = Double.Epsilon;
            double x = arg;
            double y = Math.Abs(x);

            if (y <= 0.66291e0) // Evaluate  anorm  for  |X| <= 0.66291
            {
                xsq = 0.0e0;
                if (y > 1.0e-12)
                {
                    xsq = x * x;
                }
                xnum = a[4] * xsq;
                xden = xsq;
                for (i = 0; i < 3; i++)
                {
                    xnum = (xnum + a[i]) * xsq;
                    xden = (xden + b[i]) * xsq;
                }
                result = x * (xnum + a[3]) / (xden + b[3]);
                tmp = result;
                result = 0.5e0 + tmp;
                ccum = 0.5e0 - tmp;
            }
            else if (y <= 5.656854248e0) // Evaluate  anorm  for 0.66291 < |X| <= sqrt(32)
            {
                xnum = c[8] * y;
                xden = y;
                for (i = 0; i < 7; i++)
                {
                    xnum = (xnum + c[i]) * y;
                    xden = (xden + d[i]) * y;
                }
                result = (xnum + c[7]) / (xden + d[7]);
                xsq = Math.Floor(y * 1.60e0) / 1.60e0;
                del = (y - xsq) * (y + xsq);
                result = Math.Exp(-(xsq * xsq * 0.5e0)) * Math.Exp(-(del * 0.5e0)) * result;
                ccum = 1.0e0 - result;
                if (x > 0.0e0)
                {
                    tmp = result;
                    result = ccum;
                    ccum = tmp;
                }
            }
            else // Evaluate  anorm  for |X| > sqrt(32)
            {
                result = 0.0e0;
                xsq = 1.0e0 / (x * x);
                xnum = p[5] * xsq;
                xden = xsq;
                for (i = 0; i < 4; i++)
                {
                    xnum = (xnum + p[i]) * xsq;
                    xden = (xden + q[i]) * xsq;
                }
                result = xsq * (xnum + p[4]) / (xden + q[4]);
                result = (3.9894228040143267794e-1 - result) / y;
                xsq = Math.Floor(x * 1.60e0) / 1.60e0;
                del = (x - xsq) * (x + xsq);
                result = Math.Exp(-(xsq * xsq * 0.5e0)) * Math.Exp(-(del * 0.5e0)) * result;
                ccum = 1.0e0 - result;
                if (x > 0.0e0)
                {
                    tmp = result;
                    result = ccum;
                    ccum = tmp;
                }
            }
            if (result < min)
            {
                result = 0.0e0;
            }
            if (ccum < min) // Fix up for negative argument, erf, etc.
            {
                ccum = 0.0e0;
            }
            return result;
        }

        /// <summary>
        /// Returns an approximation of the error function of z. Note that Erf(z/sqrt(2)) is
        /// equal to 2 * CDF(0,1,z) - 1, which is a better approximation. Also
        /// note that 1 - Erf(z/sqrt(2)) equals 2 * (1 - CDF(0,1,z)). So only use
        /// this approximation if you are not dividing z by sqrt(2).
        /// </summary>
        /// <param name="z">The value for which the error function is computed.</param>
        /// <returns>The error function of z.</returns>
        /// <see cref="http://en.wikipedia.org/wiki/Error_function" />
        public static double Erf(double z)
        {
            double a = (8.0 * (Math.PI - 3.0)) / (3.0 * Math.PI * (4.0 - Math.PI));
            return Math.Sqrt(
                1.0 - Math.Exp(
                        -(z * z) * (4.0 / Math.PI + a * z * z) / (1.0 + a * z * z)
                    )
                );
        }

        /// <summary>
        /// Perform the Anderson-Darling (1952) test for normality of a sample taken from an unknown distribution 
        /// of unknown mean and variance. These values will be estimated from the sample itself.
        /// </summary>
        /// <param name="x">The values on which to perform the Anderson-Darling test for normality.</param>
        /// <param name="alpha">The significance level required for the test. Use only values of .10, .05, .025, .01, .005.</param>
        /// <param name="A2">Returned A^2 statistic, i.e., the Anderson-Darling statistic.</param>
        /// <param name="CV">Returned critical value for this level of alpha. If A2 > CV, the test returns true.</param>
        /// <returns>True if the test is significant; false otherwise. This means that if true is returned,
        /// the data is deemed non-normal and the null hypothesis is rejected.</returns>
        /// <remarks>The Anderson-Darling test was published and refined in the following articles on which this function draws:
        /// Anderson, T. W. and Darling, D. A. (1952). Asymptotic theory of certain "goodness of fit" criteria based on stochastic 
        ///     processes. Annals of Mathematical Statistics 23 (2), pp. 193–212.
        /// Anderson, T.W. and Darling, D.A. (1954). A test of goodness of fit. Journal of the American Statistical Association 
        ///     49 (268), pp. 765–769.
        /// Stephens, M. A. (1974). EDF Statistics for Goodness of Fit and Some Comparisons. Journal of the American Statistical 
        ///     Association 69 (347), pp. 730–737.
        /// D'Agostino, R. B. (1986). Tests for the Normal Distribution. In D'Agostino, R. B. and Stephens, M.A. (eds.) Goodness-of-Fit 
        ///     Techniques. New York: Marcel Dekker, pp. 367-420.
        /// </remarks>
        public static bool AndersonDarling(double[] x, double alpha, out double A2, out double CV)
        {
            // Find the appropriate critical value for the given alpha value.
            //                  10%,   5%,    2.5%,  1%,    0.5%,
            double[] alphas = { 0.100, 0.050, 0.025, 0.010, 0.005 };
            double[] critix = { 0.631, 0.752, 0.873, 1.035, 1.159 }; // from D'Agostino (1986); slightly lower ("easier") than case 4 at http://en.wikipedia.org/wiki/Anderson%E2%80%93Darling_test

            CV = 0.0;
            for (int i = 0; i < alphas.Length; i++)
            {
                if (alpha == alphas[i])
                {
                    CV = critix[i];
                    break;
                }
            }
            if (CV == 0.0)
                throw new ArgumentException("Undefined alpha value for GMeans. Use .10, .05, .025, .01 or .005.");

            // Sort ascending and then transform x so it has mean 0.0 and variance 1.0.
            Array.Sort<double>(x);
            double mean = Mean(x);
            double stdev = StdDev(x);
            for (int i = 0; i < x.Length; i++)
            {
                x[i] = (x[i] - mean) / stdev; // translate and stretch or squish
            }

            // Now compute the Anderson-Darling statistic over the values x.
            int n = x.Length;
            double S = 0.0;
            for (int i = 0; i < n; i++)
            {
                double zi = CDF(0.0, 1.0, x[i]);
                double zni = CDF(0.0, 1.0, x[n - 1 - i]);
                S += (2.0 * (i + 1) - 1.0) * (Math.Log(zi) + Math.Log(1.0 - zni));
            }
            A2 = -n - (S / n); // raw A^2 statistic

            // Correction because we are using means and stdev estimated from the data. See M. A. Stephens. (September 1974). EDF statistics 
            // for goodness of fit and some comparisons. Journal of the American Statistical Association 69 (347), pp. 730–737. See Table 1.3.
            A2 *= (1.0 + (4.0 / n) - (25.0 / (n * n))); // A^2* statistic (http://en.wikipedia.org/wiki/Anderson%E2%80%93Darling_test)

            // Correction for small sample sizes. Ralph B. D'Agostino. (1986). "Tests for the Normal Distribution." In D'Agostino, R.B. and 
            // Stephens, M.A.. Goodness-of-Fit Techniques. New York: Marcel Dekker. See especially Table 4.7, p. 123 and pp. 372-373.
            if (n <= 25)
            {
                A2 *= (1.0 + (0.75 / n) + (2.25 / (n * n))); // A^2m statistic (http://www.statisticalengineering.com/goodness.htm)
            }
            return (A2 > CV);
        }

        /// <summary>
        /// Computes the exact binomial test for a number of successes out of a total number of events with an
        /// expected probability. The test can be a one-sided test or a two-sided test. The p-value of the test
        /// is returned.
        /// </summary>
        /// <param name="successes">The number of successes. A success can be interpreted as an event of interest.</param>
        /// <param name="total">The total number of events. Must be greater than or equal to the number of successes.</param>
        /// <param name="prob">The expected probability of the successes in question.</param>
        /// <param name="direction">The direction of the test. If less than zero (e.g., -1), then it indicates a one-sided
        /// "test for being significantly less than the expected probability." If greater than zero (e.g., +1), then it 
        /// indicates a one-sided "test for being significantly greater than the expected probability." And if zero, it indicates
        /// a "test for being significantly different from the expected probability."</param>
        /// <exapmle>
        /// A six-sided die is rolled 100 times, and "6" comes up 25 times. The question is whether this is greater
        /// than the expected probability for a fair die, which should be (1/6)*100 = 16.6666... times out of 100, not 25. 
        /// Calling <c>BinomialTest(25, 100, (1.0 / 6.0), +1)</c> asks this question. The return p-value of .021703 is 
        /// less than the traditional alpha = .05 threshold for statistical significance, so we conclude that the die is
        /// not fair and 25 sixes are statistically significantly greater than chance.
        /// </exapmle>
        /// <returns>The p-value of the statistical test.</returns>
        public static double BinomialTest(int successes, int total, double prob = 0.50, int direction = 0)
        {
            if (successes < 0 || successes > total)
                throw new ArgumentOutOfRangeException("The argument 'successes' is out of range in StatsEx.BinomialTest.");
            if (total <= 0)
                throw new ArgumentOutOfRangeException("The argument 'total' is out of range in StatsEx.BinomialTest.");
            if (prob < 0.0 || prob > 1.0)
                throw new ArgumentOutOfRangeException("The argument 'prob' is out of range in StatsEx.BinomialTest.");

            double pvalue;
            if (direction == 0)
            {
                pvalue = TwoSidedBinomialTest(successes, total, prob);
            }
            else
            {
                if (direction < 0)
                    pvalue = OneSidedBinomialTest(successes, total, prob);
                else // direction > 0
                    pvalue = OneSidedBinomialTest(total - successes, total, 1.0 - prob);
            }
            return pvalue;
        }

        /// <summary>
        /// Constant values used in the binomial test calculations.
        /// </summary>
        private const double EXACT_TEST_BIAS = 0.00000000000000000000000010339757656912845935892608650874535669572651386260986328125;
        private const double ONE_PLUS_EPSILON = 1.00000000000001 * EXACT_TEST_BIAS;
        private const double ONE_MINUS_EPSILON = .99999999999999 * EXACT_TEST_BIAS;

        /// <summary>
        /// Performs a two-sided binomial test, testing the proportion of successes out of a total number
        /// of observations against a hypothesized or expected probability.
        /// </summary>
        /// <param name="successes">The total number of success observations.</param>
        /// <param name="total">The total number of observations.</param>
        /// <param name="prob">The hypothesized or expected probabilty to test against.</param>
        /// <returns>The p-value for the two-sided binomial test.</returns>
        /// <remarks>
        /// This code was adapted from https://www.cog-genomics.org/static/js/software_stats.js,
        /// which is loaded to run this page: https://www.cog-genomics.org/software/stats.
        /// </remarks>
        private static double TwoSidedBinomialTest(int successes, int total, double prob)
        {
            double cur_succ_t2 = successes;
            double cur_fail_t2 = total - successes;
            double tailp = EXACT_TEST_BIAS;
            double centerp = 0.0;
            double lastp2 = tailp;
            double lastp1 = tailp;
            double prob_mult_incr = prob / (1 - prob);
            double prob_mult_decr = (1 - prob) / prob;
            double cur_succ_t1;
            double cur_fail_t1;
            double preaddp;

            if (total * prob < successes)
            {
                while (cur_succ_t2 > 0.5)
                {
                    cur_fail_t2++;
                    lastp2 *= prob_mult_decr * cur_succ_t2 / cur_fail_t2;
                    cur_succ_t2--;
                    if (lastp2 < ONE_PLUS_EPSILON)
                    {
                        tailp += lastp2;
                        break;
                    }
                    centerp += lastp2;
                    if (Double.IsPositiveInfinity(centerp))
                        return 0.0;
                }
                if (centerp == 0.0)
                    return 1.0;
                while (cur_succ_t2 > 0.5)
                {
                    cur_fail_t2++;
                    lastp2 *= prob_mult_decr * cur_succ_t2 / cur_fail_t2;
                    cur_succ_t2--;
                    preaddp = tailp;
                    tailp += lastp2;
                    if (tailp <= preaddp)
                        break;
                }
                cur_succ_t1 = successes + 1.0;
                cur_fail_t1 = total - successes;
                while (cur_fail_t1 > 0.5)
                {
                    lastp1 *= prob_mult_incr * cur_fail_t1 / cur_succ_t1;
                    preaddp = tailp;
                    tailp += lastp1;
                    if (tailp <= preaddp)
                        break;
                    cur_succ_t1++;
                    cur_fail_t1--;
                }
            }
            else
            {
                while (cur_fail_t2 > 0.5)
                {
                    cur_succ_t2++;
                    lastp2 *= prob_mult_incr * cur_fail_t2 / cur_succ_t2;
                    cur_fail_t2--;
                    if (lastp2 < ONE_PLUS_EPSILON)
                    {
                        tailp += lastp2;
                        break;
                    }
                    centerp += lastp2;
                    if (Double.IsPositiveInfinity(centerp))
                        return 0.0;
                }
                if (centerp == 0.0)
                    return 1.0;
                while (cur_fail_t2 > 0.5)
                {
                    cur_succ_t2++;
                    lastp2 *= prob_mult_incr * cur_fail_t2 / cur_succ_t2;
                    cur_fail_t2--;
                    preaddp = tailp;
                    tailp += lastp2;
                    if (tailp <= preaddp)
                        break;
                }
                cur_succ_t1 = successes;
                cur_fail_t1 = total - successes;
                while (cur_succ_t1 > 0.5)
                {
                    cur_fail_t1++;
                    lastp1 *= prob_mult_decr * cur_succ_t1 / cur_fail_t1;
                    preaddp = tailp;
                    tailp += lastp1;
                    if (tailp <= preaddp)
                        break;
                    cur_succ_t1--;
                }
            }
            return tailp / (tailp + centerp);
        }

        /// <summary>
        /// Performs a one-sided binomial test, testing the proportion of successes out of a total number
        /// of observations against a hypothesized or expected probability.
        /// </summary>
        /// <param name="successes">The total number of success observations.</param>
        /// <param name="total">The total number of observations.</param>
        /// <param name="prob">The hypothesized or expected probabilty to test against.</param>
        /// <returns>The p-value for the one-sided binomial test.</returns>
        /// <remarks>
        /// This code was adapted from https://www.cog-genomics.org/static/js/software_stats.js, 
        /// which is loaded to run this page: https://www.cog-genomics.org/software/stats. 
        /// </remarks>
        private static double OneSidedBinomialTest(int successes, int total, double prob)
        {
            double cur_prob = EXACT_TEST_BIAS;
            double left_prob = cur_prob;
            double right_prob = 0.0;
            double prob_mult_incr = prob / (1.0 - prob);
            double prob_mult_decr = (1.0 - prob) / prob;
            double cur_succ = successes;
            double cur_fail = total - successes;
            double preaddp;

            if (total * prob < successes)
            {
                while (cur_succ > 0.5)
                {
                    cur_fail++;
                    cur_prob *= prob_mult_decr * cur_succ / cur_fail;
                    cur_succ--;
                    preaddp = left_prob;
                    left_prob += cur_prob;
                    if (left_prob <= preaddp)
                        break;
                    if (left_prob >= 1.0)
                        return 1.0;
                }
                cur_succ = successes;
                cur_fail = total - successes;
                cur_prob = EXACT_TEST_BIAS;
                right_prob = left_prob;
                while (cur_fail > 0.5)
                {
                    cur_succ++;
                    cur_prob *= prob_mult_incr * cur_fail / cur_succ;
                    cur_fail--;
                    preaddp = right_prob;
                    right_prob += cur_prob;
                    if (right_prob <= preaddp)
                        break;
                }
                return left_prob / right_prob;
            }
            else
            {
                while (cur_fail > 0.5)
                {
                    cur_succ++;
                    cur_prob *= prob_mult_incr * cur_fail / cur_succ;
                    cur_fail--;
                    preaddp = right_prob;
                    right_prob += cur_prob;
                    if (Double.IsPositiveInfinity(right_prob))
                        return 0.0;
                    if (right_prob <= preaddp)
                        break;
                }
                cur_succ = successes;
                cur_fail = total - successes;
                cur_prob = EXACT_TEST_BIAS;
                while (cur_succ > 0.5)
                {
                    cur_fail++;
                    cur_prob *= prob_mult_decr * cur_succ / cur_fail;
                    cur_succ--;
                    preaddp = left_prob;
                    left_prob += cur_prob;
                    if (left_prob <= preaddp)
                        break;
                }
                return left_prob / (left_prob + right_prob);
            }
        }

    }
}
