﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.IO;

namespace WobbrockLib.Extensions
{
    /// <summary>
    /// Utility functions for working with strings.
    /// </summary>
    public static class StringEx
    {
        /// <summary>
        /// Returns a string representation of an array in human-readable format. The type
        /// of the array can be anything, since its elements must simply implement the
        /// <b>ToString()</b> method, which all C# objects must and do.
        /// </summary>
        /// <param name="array"></param>
        /// <returns>A string in the form of <i>[a,b,c,d,e]</i> representing the elements in an array.</returns>
        public static string Array2String(Array array)
        {
            string s = "[";
            for (int i = 0; array != null && i < array.Length; i++)
            {
                s += array.GetValue(i).ToString() + ",";
            }
            s = s.TrimEnd(new char[1] { ',' });
            s += "]";
            return s;
        }

        /// <summary>
        /// This function implements the reverse of <b>StringEx.Array2String</b> for an integer
        /// array. The string format is assumed to be <i>[a,b,c,d,e]</i>.
        /// </summary>
        /// <param name="s">The string representation of an array.</param>
        /// <returns>The integer array, or <b>null</b> if the format is incorrect or the array is empty.</returns>
        public static int[] String2IntArray(string s)
        {
            if (s == "[]")
                return null; // no elements

            int commas = 0;
            for (int i = 0; i < s.Length; i++)
            {
                if (s[i] == ',')
                    commas++;
            }
            int start = 1;
            int end = 0; // comma position
            int[] array = new int[commas + 1];
            for (int i = 0; i < array.Length; i++)
            {
                end = s.IndexOf(i == array.Length - 1 ? ']' : ',', start);
                try { array[i] = Int32.Parse(s.Substring(start, end - start)); }
                catch { return null; }
                start = end + 1;
            }
            return array;
        }

        /// <summary>
        /// This function implements the reverse of <b>StringEx.Array2String</b> for an array
        /// of double-precision floating-point numbers. The string format is assumed to 
        /// be <i>[a,b,c,d,e]</i>.
        /// </summary>
        /// <param name="s">The string representation of an array.</param>
        /// <returns>The double array, or <b>null</b> if the format is incorrect.</returns>
        public static double[] String2DoubleArray(string s)
        {
            if (s == "[]")
                return null; // no elements
            
            int commas = 0;
            for (int i = 0; i < s.Length; i++)
            {
                if (s[i] == ',')
                    commas++;
            }
            int start = 1;
            int end = 0; // comma position
            double[] array = new double[commas + 1];
            for (int i = 0; i < array.Length; i++)
            {
                end = s.IndexOf(i == array.Length - 1 ? ']' : ',', start);
                try { array[i] = Double.Parse(s.Substring(start, end - start)); }
                catch { return null; }
                start = end + 1;
            }
            return array;
        }

        /// <summary>
        /// This function implements the reverse of <b>StringEx.Array2String</b> for an array
        /// of decimal numbers. The string format is assumed to be <i>[a,b,c,d,e]</i>.
        /// </summary>
        /// <param name="s">The string representation of an array.</param>
        /// <returns>The decimal array, or <b>null</b> if the format is incorrect.</returns>
        public static decimal[] String2DecimalArray(string s)
        {
            if (s == "[]")
                return null; // no elements

            int commas = 0;
            for (int i = 0; i < s.Length; i++)
            {
                if (s[i] == ',')
                    commas++;
            }
            int start = 1;
            int end = 0; // comma position
            decimal[] array = new decimal[commas + 1];
            for (int i = 0; i < array.Length; i++)
            {
                end = s.IndexOf(i == array.Length - 1 ? ']' : ',', start);
                try { array[i] = Decimal.Parse(s.Substring(start, end - start)); }
                catch { return null; }
                start = end + 1;
            }
            return array;
        }

        /// <summary>
        /// This function implements the reverse of <b>Rectangle.ToString()</b>, creating a <b>Rectangle</b> 
        /// structure from its string representation. All members are assumed to be integer types.
        /// </summary>
        /// <param name="s">The string representation of a <b>Rectangle</b>. Example: {X=20,Y=20,Width=100,Height=50}.</param>
        /// <returns>The <b>Rectangle</b> structure representing string <i>s</i>.</returns>
        /// <remarks>For converting a string to a double-precision floating-point RectangleR, use RectangleR.FromString.</remarks>
        public static Rectangle String2Rectangle(string s)
        {
            Rectangle r = Rectangle.Empty;
            try
            {
                int xpos = s.IndexOf('X'); // "X"
                int comma = s.IndexOf(',', xpos + 2);
                int x = int.Parse(s.Substring(xpos + 2, comma - (xpos + 2)));

                int ypos = s.IndexOf('Y'); // "Y"
                comma = s.IndexOf(',', ypos + 2);
                int y = int.Parse(s.Substring(ypos + 2, comma - (ypos + 2)));

                int wpos = s.IndexOf('W'); // "Width"
                comma = s.IndexOf(',', wpos + 6);
                int w = int.Parse(s.Substring(wpos + 6, comma - (wpos + 6)));

                int hpos = s.IndexOf('H'); // "Height"
                comma = s.IndexOf('}', hpos + 7);
                int h = int.Parse(s.Substring(hpos + 7, comma - (hpos + 7)));

                r = new Rectangle(x, y, w, h);
            }
            catch { }
            return r;
        }

        /// <summary>
        /// This function implements the reverse of <b>Point.ToString()</b>, creating a <b>Point</b> 
        /// structure from its string representation. The (x,y) coordinates are assumed to be integer
        /// types.
        /// </summary>
        /// <param name="s">The string representation of a <b>Point</b>. Example: {X=20,Y=30}.</param>
        /// <returns>The <b>Point</b> structure representing string <i>s</i>.</returns>
        /// <remarks>For converting a string to a double-precision floating-point PointR, use PointR.FromString.</remarks>
        public static Point String2Point(string s)
        {
            Point pt = Point.Empty;
            try
            {
                int xpos = s.IndexOf('X'); // "X"
                int comma = s.IndexOf(',', xpos + 2);
                int x = int.Parse(s.Substring(xpos + 2, comma - (xpos + 2)));

                int ypos = s.IndexOf('Y'); // "Y"
                comma = s.IndexOf('}', ypos + 2);
                int y = int.Parse(s.Substring(ypos + 2, comma - (ypos + 2)));

                pt = new Point(x, y);
            }
            catch { }
            return pt;
        }

        
        /// <summary>
        /// This function implements the reverse of <b>Size.ToString()</b>, creating a <b>Size</b> 
        /// structure from its string representation. The width and height are assumed to be integer 
        /// types.
        /// </summary>
        /// <param name="s">The string representation of a <b>Size</b>. Example: {Width=20,Height=30}.</param>
        /// <returns>The <b>Size</b> structure representing string <i>s</i>.</returns>
        /// <remarks>For converting a string to a double-precision floating-point SizeR, use SizeR.FromString.</remarks>
        public static Size String2Size(string s)
        {
            Size sz = Size.Empty;
            try
            {
                int wpos = s.IndexOf('W'); // "W"idth
                int comma = s.IndexOf(',', wpos + 6);
                int w = int.Parse(s.Substring(wpos + 6, comma - (wpos + 6)));

                int hpos = s.IndexOf('H'); // "H"eight
                comma = s.IndexOf('}', hpos + 7);
                int h = int.Parse(s.Substring(hpos + 7, comma - (hpos + 7)));

                sz = new Size(w, h);
            }
            catch { }
            return sz;
        }

        /// <summary>
        /// Makes the given name a valid filename, stripping it of any invalid characters.
        /// </summary>
        /// <param name="name">The filename to strip of any invalid characters.</param>
        /// <returns>A valid filename.</returns>
        public static string MakeValidFilename(string name)
        {
            char[] invalids = Path.GetInvalidFileNameChars();
            int idx = name.IndexOfAny(invalids);
            while (idx != -1)
            {
                name = name.Remove(idx, 1);
                idx = name.IndexOfAny(invalids);
            }
            return name;
        }

        /// <summary>
        /// Gets a random string of alphanumeric characters at a random length between
        /// a minimum and maximum prescribed length, inclusive.
        /// </summary>
        /// <param name="minlen">The minimum length of the desired string, inclusive.</param>
        /// <param name="maxlen">The maximum length of the desired string, inclusive.</param>
        /// <returns>A random alphanumeric string.</returns>
        public static string Random(int minlen, int maxlen)
        {
            return RandomEx.String(minlen, maxlen); // just patch through to RandomEx
        }
    }
}
