using System;
using System.Net;
using System.Net.Sockets;
using System.Text;

namespace WobbrockLib.Net
{
    /// <summary>
    /// Implements a TCP client that sends messages to a listening TCP server.
    /// </summary>
    public class NetSender
    {
        protected IPAddress     _ip;
        protected int           _port;
        protected TcpClient     _client;

        /// <summary>
        /// Constructs a client capable of connecting to a server and sending data over TCP.
        /// </summary>
        public NetSender()
        {
            _ip = IPAddress.Parse("127.0.0.1");
            _port = 9079;
            _client = null;
        }

        /// <summary>
        /// The IP address over which this client will send information to a listening server.
        /// </summary>
        public virtual string IP
        {
            get
            {
                return _ip.ToString();
            }
            set
            {
                try
                {
                    _ip = IPAddress.Parse(value);
                }
                catch (Exception ex)
                {
                    Console.WriteLine(ex);
                }
            }
        }

        /// <summary>
        /// The port over which this client will send information to a listening server.
        /// </summary>
        public virtual int Port
        {
            get
            {
                return _port;
            }
            set
            {
                if (value > 0)
                {
                    _port = value;
                }
            }
        }

        /// <summary>
        /// Gets whether or not this client is connected to a listening server.
        /// </summary>
        public virtual bool IsConnected
        {
            get
            {
                return (_client != null);
            }
        }

        /// <summary>
        /// Connects or reconnects this client to a listening server. If the server is not already there 
        /// listening, this method will fail.
        /// </summary>
        /// <returns>True if a connection is made successfully; false otherwise.</returns>
        public virtual bool Connect()
        {
            if (IsConnected)
            {
                Disconnect();
            }

            bool success = true;
            try
            {
                _client = new TcpClient();
                _client.NoDelay = true;
                _client.Connect(_ip, _port);
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex);
                _client = null;
                success = false;
            }
            return success;
        }

        /// <summary>
        /// Disconnect this client from a listening server.
        /// </summary>
        public virtual void Disconnect()
        {
            if (_client != null)
            {
                try
                {
                    NetworkStream stream = _client.GetStream();

                    // have to close the stream itself so the listener ceases to block on Read.
                    // the docs say closing the client closes the stream, but that isn't sufficient.
                    stream.Close();
                }
                catch (Exception ex)
                {
                    Console.WriteLine(ex);
                }
                _client.Close();
                _client = null;
            }
        }

        /// <summary>
        /// Send a message to a listening server over TCP.
        /// </summary>
        /// <param name="msg"></param>
        /// <returns></returns>
        public virtual bool Send(string msg)
        {
            bool success = true;
            try
            {
                // encode our message for sending
                byte[] buffer = Encoding.UTF8.GetBytes(msg);

                // send the message over the stream to the server
                NetworkStream stream = _client.GetStream();
                stream.Write(buffer, 0, buffer.Length);
            }
            catch (SocketException sex)
            {
                Console.WriteLine(sex);
                success = false;
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex);
                success = false;
            }
            finally
            {
                if (!success)
                {
                    _client.Close();
                    _client = null;
                }
            }
            return success;
        }

    }
}