﻿using System;
using System.IO;
using System.Text;
using System.Net.Sockets;
using System.Threading;
using System.Windows.Forms;
using System.Text.RegularExpressions;
using System.Security.Cryptography;

namespace WobbrockLib.Net
{
    /// <summary>
    /// Implements a WebSockets server that listens for client messages.
    /// </summary>
    public class WebSocketListener : NetListener
    {
        public override event EventHandler NetEvent;
        public override event NetEventHandler NetDataEvent;

        /// <summary>
        /// Constructs a WebSocketListener, a server capable of receiving client connections over
        /// WebSockets, e.g., connections from web pages.
        /// </summary>
        /// <param name="owner"></param>
        public WebSocketListener(Control owner) : base(owner)
        {
            // do nothing
        }

        /// <summary>
        /// Sends the given message to a connected WebSockets client. Does not handle partial messages or
        /// message continuation. Assumes text is being sent, not binary messages.
        /// </summary>
        /// <param name="msg">The text message to send to the client.</param>
        /// <returns>True if the message is sent successfully; false otherwise.</returns>
        /// <see cref="https://developer.mozilla.org/en-US/docs/Web/API/WebSockets_API/Writing_WebSocket_servers"/> 
        public override bool Send(string msg)
        {
            if (_clientStream == null)
                return false;

            bool success = true;
            try
            {
                // encode our message for sending
                byte[] buffer = Encoding.UTF8.GetBytes(msg);

                // MSB sets the "fin" bit (whole message is being sent).
                // the LSB sets opcode of 1 for sending UTF-8 text.
                // if we were sending binary, it'd be 2.
                MemoryStream m = new MemoryStream();
                byte byte0 = (0x01 << 7) | 0x01; // 1000 0001 == 129
                m.WriteByte(byte0);

                // encode the length field
                if (buffer.Length <= 125) // if msg length is <= 125 bytes
                {
                    byte byte1 = (byte) buffer.Length;
                    m.WriteByte(byte1);
                }
                else if (buffer.Length <= 0xFFFF) // if msg length is (125 < len <= 65,535) [1111 1111 1111 1111] (i.e., 16 'on' bits)
                {
                    byte byte1 = 126;
                    m.WriteByte(byte1);
                    ushort len = (ushort) buffer.Length;
                    for (int i = sizeof(UInt16) - sizeof(Byte); i >= 0; i -= sizeof(Byte))
                    {
                        byte b = (byte) ((len >> i) & 0xFF); // 1111 1111
                        m.WriteByte(b);
                    }
                }
                else // the length of the message is greater than ushort.MaxValue (65,535)
                {
                    byte byte1 = 127;
                    m.WriteByte(byte1);
                    ulong len = (ulong) buffer.Length;
                    for (int i = sizeof(UInt64) - sizeof(Byte); i >= 0; i -= sizeof(Byte))
                    {
                        byte b = (byte) ((len >> i) & 0xFF); // 1111 1111
                        m.WriteByte(b);
                    }
                }
                m.Write(buffer, 0, buffer.Length); // finally write the message payload itself
                byte[] bytes = m.ToArray(); // get an array of bytes for the message
                _clientStream.Write(bytes, 0, bytes.Length); // send the message over the stream to the client!
            }
            catch (SocketException sex)
            {
                Console.WriteLine(sex);
                success = false;
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex);
                success = false;
            }
            return success;
        }

        /// <summary>
        /// The method run by the background listener thread that receives data from a WebSockets client.
        /// </summary>
        protected override void ThreadListen()
        {
            try
            {
                _server = new TcpListener(_ip, _port);
                _server.Start(); // start listening

                while (true) // no break from this loop -- CLR will close background thread
                {
                    TcpClient client = _server.AcceptTcpClient(); // blocking until a connection
                    _clientStream = client.GetStream(); // connection!

                    while (!_clientStream.DataAvailable)
                        Thread.Sleep(0); // yield
                    byte[] buffer = new byte[client.Available];
                    string data = String.Empty;

                    // loop to receive all the data sent by the client
                    int i = 0;
                    while ((i = _clientStream.Read(buffer, 0, buffer.Length)) != 0) // blocks until it can read
                    {
                        // translate the raw bytes into an UTF-8 string
                        data = Encoding.UTF8.GetString(buffer, 0, i);

                        if (new Regex("^GET").IsMatch(data)) // WebSockets handshake
                        {
                            byte[] response = Encoding.UTF8.GetBytes("HTTP/1.1 101 Switching Protocols" + Environment.NewLine
                                + "Connection: Upgrade" + Environment.NewLine
                                + "Upgrade: websocket" + Environment.NewLine
                                + "Sec-WebSocket-Accept: " + Convert.ToBase64String(
                                    SHA1.Create().ComputeHash(
                                        Encoding.UTF8.GetBytes(
                                            new Regex("Sec-WebSocket-Key: (.*)").Match(data).Groups[1].Value.Trim() + "258EAFA5-E914-47DA-95CA-C5AB0DC85B11"
                                            )
                                        )
                                    ) + Environment.NewLine
                                    + Environment.NewLine);
                            _clientStream.Write(response, 0, response.Length);
                        }
                        else // not the WebSockets handshake
                        {
                            // decode the received bytes according to the WebSockets spec
                            data = DecodeMessage(buffer);
                            
                            // fire the data received events to the client control
                            if (NetEvent != null)
                                _owner.BeginInvoke(NetEvent); // fire event
                            if (NetDataEvent != null)
                                _owner.BeginInvoke(NetDataEvent, new object[2] { this, new NetEventArgs(data) }); // fire event
                        }

                        // yield the processor until we get new data
                        while (!_clientStream.DataAvailable)
                            Thread.Sleep(0); // yield
                        buffer = new byte[client.Available];
                        data = String.Empty;
                    }

                    client.Close();
                    _clientStream.Close();
                    _clientStream = null;
                }
            }
            catch (ThreadAbortException taex)
            {
                Console.WriteLine("ThreadAbortException: {0}", taex);
            }
            catch (SocketException sex)
            {
                Console.WriteLine("SocketException: {0}", sex);
            }
            catch (Exception ex)
            {
                Console.WriteLine("Exception: {0}", ex);
            }
            finally
            {
                if (_server != null)
                {
                    _server.Stop();
                    _server = null;
                }
            }
        }

        /// <summary>
        /// Decode the bytes in the given buffer according to the WebSockets spec.
        /// </summary>
        /// <param name="buffer">Bytes received from a client over a WebSockets connection.</param>
        /// <returns>The string representation of the recieved bytes.</returns>
        /// <see cref="https://developer.mozilla.org/en-US/docs/Web/API/WebSockets_API/Writing_WebSocket_servers"/> 
        private string DecodeMessage(byte[] buffer)
        {
            byte[] data;
            byte[] masks = new byte[4];
            uint len = buffer[1] & 127u; // length encoded at second byte except MSB: 127 == 0111 1111.
                                         // if 126, then length is next 16 bits. if 127, then length
                                         // is next 64 bits.

            int idx = 2; // index of first mask (default: 2 for len <= 125)
            if (len == 126u)
                idx = 4;
            else if (len == 127u)
                idx = 10;
            
            // set up the four mask bytes
            Array.Copy(buffer, idx, masks, 0, 4);

            // set up the data bytes
            data = new byte[buffer.Length - idx - 4];
            Array.Copy(buffer, idx + 4, data, 0, data.Length);

            // decode the data bytes using the masks
            byte[] decoded = new byte[data.Length];
            for (int i = 0; i < data.Length; i++)
            {
                decoded[i] = (byte) (data[i] ^ masks[i % 4]);
            }

            return Encoding.UTF8.GetString(decoded);
        }

    }
}
