﻿using System;
using WobbrockLib.Optimization;

namespace WobbrockLib.Optimization.Genetic
{
    /// <summary>
    /// This abstract class defines an organism for a genetic algorithm, a computational 
    /// creature that encodes a possible solution to a problem and who can be 
    /// judged as being fit or unfit based on how close that solution is to 
    /// one satisfying a problem definition.
    /// </summary>
    public abstract class Organism
    {
        /// <summary>
        /// This member codifies the generation to which this organism belongs. By definition,
        /// the 'Eve' organism used to produce the initial random population is generation 0.
        /// Her random offspring are the first full generation, generation 1. Their offspring
        /// are generation 2, and so on.
        /// </summary>
        protected int _generation;

        /// <summary>
        /// This member is optionally set upon the birth of this organism if the world state is
        /// supplied. The assumption in such cases is that the world state is unchanging, and
        /// therefore the fitness needs to only be computed once, dramatically speeding up the
        /// sorting process in during evolution.
        /// </summary>
        protected double _fitness;

        /// <summary>
        /// Default constructor for this class.
        /// </summary>
        public Organism()
        {
            _generation = 0;
            _fitness = double.NaN;
        }

        /// <summary>
        /// Gets or sets the generation number of this organism. 'Eve' is considered generation 
        /// 0, and produces generation 1 randomly. The mutated spawn from generation 1 is
        /// generation 2, and so on.
        /// </summary>
        /// <remarks>The setter is provided so that parents who survive into subsequent
        /// generations can have their own generation number updated. Parents survive
        /// this way when none of their offspring are more fit than they are. If any
        /// offspring are more fit, then the parent does not survive. For updating
        /// offsprings' generation numbers from their parents, this should happpen
        /// automatically in BirthMutated.</remarks>
        public virtual int Generation
        {
            get { return _generation; }
            set { _generation = value; }
        }

        /// <summary>
        /// Gets the solution encoded in this organism. Metaphorically, this is known as 
        /// the organism's "DNA."
        /// </summary>
        public abstract object DNA { get; }

        /// <summary>
        /// Gets the printable string representation of this organism's DNA; that is,
        /// the solution currently encoded by this organism. Many subclasses will want
        /// to craft custom strings to represent their organism's solution.
        /// </summary>
        /// <returns>A string representation of the solution encoded in this organism's DNA.</returns>
        public override string ToString()
        {
            return DNA.ToString();
        }

        /// <summary>
        /// Makes a deep copy of this organism.
        /// </summary>
        /// <returns>A deep copy of this organism.</returns>
        public abstract Organism Copy();

        /// <summary>
        /// Produces a random organism of the same kind as this one. This method is 
        /// used in establishing initial populations from an 'Eve' organism. 
        /// Generation-wise, 'Eve' is considered 0 and the first random generation 
        /// created from her is considered 1. Thus, implementations of this method 
        /// should produce organisms with generation values of 1.
        /// </summary>
        /// <param name="state">If the fitness can be calculated and stored as a static
        /// value thereafter, the world state can be passed in to do so, saving future
        /// computations of fitness and speeding up evolution considerably.</param>
        /// <returns>A new randomly generated organism with a generation of 1.</returns>
        public abstract Organism BirthRandom(object state = null);

        /// <summary>
        /// Produces a new organism that represents a mutation of this organism.
        /// The number of random mutations to perform is provided. Generation-wise,
        /// the mutated offspring of the parent will be +1 generation greater than
        /// its parent (i.e., this organism).
        /// </summary>
        /// <param name="nMutations">The number of random mutations to perform.</param>
        /// <param name="state">If the fitness can be calculated and stored as a static
        /// value thereafter, the world state can be passed in to do so, saving future
        /// computations of fitness and speeding up evolution considerably.</param>
        /// <returns>A new organism derived from this organism with a generation 
        /// +1 greater than this organism.</returns>
        public abstract Organism BirthMutated(int nMutations, object state = null);

        /// <summary>
        /// Gets the size of the organism defined as the number of data elements the organism 
        /// encodes in its DNA.
        /// </summary>
        /// <remarks>This property is used to adjust the number of mutations per generation so that 
        /// as generations progress, fewer mutations occur per organism in producing offspring.</remarks>
        /// <example>If the organism encodes a possible solution as a string of integers, its Size would 
        /// be the number of integers. If an organism encodes a character string, its Size would be its 
        /// string length. If it encodes a 2-D bitmap, its Size would be its 2x2 pixel area.</example>
        public abstract int Size { get; }

        /// <summary>
        /// Indexer that gets and sets individual elements encoded within this organism. Getting is 
        /// provided to inspect an organism's elements; setting is provided only for advanced 
        /// applications (e.g., manual overriding), as mutations should be the way an organism's elements 
        /// are set. This indexer shold be rarely, if ever, used.
        /// </summary>
        /// <param name="index">The index of the organism's element to get or set.</param>
        /// <returns>An object representing the organism's element at that index.</returns>
        public abstract object this[int index] { get; set; }

        /// <summary>
        /// If a static fitness score has been assigned, then the fitness value will not be NaN.
        /// </summary>
        public bool FitnessAssignedAtBirth
        {
            get { return !Double.IsNaN(_fitness); }
        }

        /// <summary>
        /// Gets a fitness score for this organism. The range of the score can be whatever the implementer 
        /// wishes. By necessity for sorting, it should be that a higher score is "more fit."
        /// </summary>
        /// <param name="state">Any world state that is necessary to evaluate the fitness of this organism.
        /// May be null if no extra state is necessary to judge the fitness of this organism. If it is
        /// null, then an attempt should be made to return a pre-calculated fitness score assigned at
        /// birth.</param>
        /// <returns>A fitness score.</returns>
        public double GetFitness(object state = null)
        {
            return (state == null) ? _fitness : ComputeFitness(state);
        }

        /// <summary>
        /// Computes the fitness score afresh for this organism. This is often an involved computation
        /// and ideally is only carried out once at organism birth. But for world states that are dynamic,
        /// fitness might not be assignable at birth.
        /// </summary>
        /// <param name="state">
        /// The world state that is necessary to evaluate the fitness of this organism.
        /// </param>
        /// <returns>A freshly computed fitness score.</returns>
        public abstract double ComputeFitness(object state);

        /// <summary>
        /// Tests whether or not this organism encodes an adequate solution to the problem
        /// for which it was designed. If an organism is deemed "fit enough," there is no sense
        /// in further evolving an ecosystem to improve upon the solution, and the evolution
        /// will short-circuit. Implementors may just return 'false' if the problem is not one that
        /// has a definably optimal solution, i.e., if it can always be improved with successive
        /// generations. (An example of this would be an organism seeking out ever more digits of pi.)
        /// </summary>
        /// <param name="state">Any world state that is necessary to evaluate the fitness of this organism.
        /// May be null if no extra state is necessary to judge the fitness of this organism.</param>
        /// <returns>True if this organism cannot be improved upon; false otherwise.</returns>
        public abstract bool IsFitEnough(object state = null);

        /// <summary>
        /// Tests whether or not this organism is fitter than a given organism. Higher values are more fit.
        /// </summary>
        /// <param name="o">The given organism to compare this organism to.</param>
        /// <param name="state">Any world state that is necessary to evaluate the fitness of this organism.
        /// May be null if no extra state is necessary to judge the fitness of this organism.</param>
        /// <returns>
        /// True if this organism is fitter than the given one; false otherwise.
        /// </returns>
        public virtual bool IsFitterThan(Organism o, object state = null)
        {
            double thisFitness = this.GetFitness(state);
            double oFitness = o.GetFitness(state);
            return thisFitness > oFitness;
        }

        /// <summary>
        /// Tests whether or not this organism encodes a solution that has "converged with"
        /// the given organism. To "converge" is organism-specific, but means that the two
        /// organisms encode the same solution or a solution that is substantially similar. In
        /// the evolution process, if convergence persists a given number of times in a row, the 
        /// evolution process can be short-circuited and the solution returned prior to evolving 
        /// all generations, the idea being that no better solution is likely to be found.
        /// </summary>
        /// <param name="o">The given organism to test for convergence with this organism.</param>
        /// <param name="state">Any world state that is necessary to evaluate the fitness of this organism.
        /// May be null if no extra state is necessary to judge the fitness of this organism.</param>
        /// <returns>True if convergence has been determined; false otherwise. Organisms wishing to
        /// never allow for convergence because a solution can perpetually be improved upon can simply
        /// return false. (An example of this would be an organism seeking out ever more digits of pi.)
        /// </returns>
        public abstract bool HasConvergedWith(Organism o, object state = null);
    }
}
