﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace WobbrockLib.Optimization
{
    /// <summary>
    /// Defines a class capable of executing the Steepest Ascent Hill Climbing iterative improvement algorithm
    /// over a solution space. See https://en.wikipedia.org/wiki/Hill_climbing for more.
    /// </summary>
    public class HillClimber : Optimizer
    {
        /// <summary>
        /// Constrcuts a climber.
        /// </summary>
        public HillClimber()
        {
            // do nothing
        }

        /// <summary>
        /// The main hill-climbing loop that searches for a solution in the solution space. A solution is deemed
        /// to be found when a location within the space is:
        ///     (a) perfect (has zero cost), 
        ///     (b) satisfices (has epsilon cost), or 
        ///     (c) has no adjacent locations with a better solution.
        /// </summary>
        /// <param name="start">The starting solution from which to climb.</param>
        /// <param name="state">Optional world state to inform the calculation of cost during the optimization
        /// process.</param>
        /// <param name="fn">An optional callback function that reports the moves made thus far by the
        /// hill climber. Note that the callback is made synchronously on the calling thread, so if the 
        /// subscriber intends to carry out extensive operations or update UI controls as a result, a
        /// new background thread should be spawned to handle that.</param>
        /// <returns>The final solution found at a local, and hopefully global, optimum.</returns>
        public override Optimization Optimize(Optimization start, object state = null, OptimizationCallback fn = null)
        {
            _nMoves = 0uL; // reset

            // see if the starting position happens to satisfice
            if (IsSolution(start, state))
                return start;

            // start the iterative hill climbing
            while (true)
            {
                double bestDeltaCost = 0.0; // best decrease in cost
                int bestIdx = -1; // index of best move
                double thisCost = start.GetCost(state); // cost of current solution

                List<Optimization> moves = start.GetMoves(); // get all possible next moves ("neighbors")
                for (int i = 0; i < moves.Count; i++) // loop over all possible moves from this state
                {
                    double moveCost = moves[i].GetCost(state);
                    if (IsSolution(moveCost)) // (a) perfect, or (b) good enough?
                    {
                        _nMoves++;
                        fn?.Invoke(_nMoves);
                        return moves[i];
                    }
                    else // make next best move
                    {
                        double deltaCost = thisCost - moveCost; // is there a reduction in cost?
                        if (deltaCost > bestDeltaCost)
                        {
                            bestDeltaCost = deltaCost;
                            bestIdx = i;
                        }
                    }
                }
                if (bestIdx != -1) // take the best move
                {
                    start = moves[bestIdx]; // best move becomes next round's start position
                    _nMoves++;
                    fn?.Invoke(_nMoves);
                }
                else break; // (c) no moves in any direction reduce cost, so we're at an optimum (local or global)
            }
            return start;
        }
    }
}
