﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace WobbrockLib.Optimization
{
    /// <summary>
    /// Defines a single solution, i.e., location, within a searchable solution space.
    /// </summary>
    public abstract class Optimization
    {
        /// <summary>
        /// Default constructor leaves an unassigned.
        /// </summary>
        public Optimization()
        {
            // do nothing
        }

        /// <summary>
        /// Gets the state encapsulated by this object.
        /// </summary>
        public abstract object State { get; }

        /// <summary>
        /// Gets a string representation of this solution. As the part of a solution that we 
        /// care about is its value, or data, the string representation of a solution is the
        /// string representation of its encapsulated value.
        /// </summary>
        /// <returns>The string representation for this AbsSolution.</returns>
        public override string ToString()
        {
            return State.ToString();
        }

        /// <summary>
        /// Makes and returns a deep copy of this solution.
        /// </summary>
        /// <returns></returns>
        public abstract Optimization Copy();

        /// <summary>
        /// Gets a random instance of this solution type. This can be thought of as jumping to a random
        /// state within the solution space, as opposed to moving to a neighboring solution state.
        /// </summary>
        public abstract Optimization GetRandom();

        /// <summary>
        /// Gets the neighboring states that constitute the possible moves from this state to the next state.
        /// </summary>
        /// <returns>A list of one or more neighbors of this solution.</returns>
        public abstract List<Optimization> GetMoves();

        /// <summary>
        /// Computes the cost of this solution in the solution space. A cost of 0.00 represents a "perfect 
        /// solution," and usually does not exist for most problems.
        /// </summary>
        /// <param name="state">Optional world state to help compute the cost.</param>
        /// <returns>The cost of this solution within the solution space.</returns>
        /// <remarks>Computing the cost is problem-specific, and easily can be the most computationally expensive part 
        /// of an optimization search process, so care should be taken to make computing the cost of any given solution 
        /// as efficient as possible.
        /// </remarks>
        public abstract double GetCost(object state = null);
    }
}
