﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace WobbrockLib.Optimization
{
    /// <summary>
    /// Abstract base class for optimization algorithms that search over a state
    /// space of possible solutions. Examples of such optimization algorithms 
    /// are hill climbing, shotgun hill climbing, or simulated annealing.
    /// </summary>
    public abstract class Optimizer
    {
        /// <summary>
        /// Fields.
        /// </summary>
        protected ulong _nMoves; // number of moves to a neighbor we have made when searching
        protected double _satisficingCost; // high-enough goodness score to short-circuit the optimizer

        /// <summary>
        /// Constructor for the abstract base class for iterative improvement optimization 
        /// algorithms, such as hill climbing, shotgun hill climbing, or simulated annealing.
        /// </summary>
        public Optimizer()
        {
            _nMoves = 0uL; // increments when each move is taken by the optimizer
            _satisficingCost = double.NaN; // initially unset and optional
        }

        /// <summary>
        /// Gets the number of moves taken by this optimizer during its current search process.
        /// The protected member variable should be reset to zero at the start of a new search
        /// process.
        /// </summary>
        public ulong NumMoves
        {
            get { return _nMoves; }
        }

        /// <summary>
        /// Optionally, defines the threshold that, if the cost function produces a result that is 
        /// this low or lower, the optimizer short-circuits the search process and stops. This number 
        /// is double.NaN by default and unused, unless explicitly set.
        /// </summary>
        /// <example>If an optimizer was searching over real-valued numbers in search of Math.PI,
        /// and 3.14 was "close enough," then this "cost" would be Math.Abs(Math.PI - 3.14).</example>
        public double SatisficingCost
        {
            get { return _satisficingCost; }
            set
            {
                if (value > 0.00)
                    _satisficingCost = value;
                else
                    throw new ArgumentOutOfRangeException("Argument out of range in Optimizer.SatisficingCost.");
            }
        }

        /// <summary>
        /// Gets whether this climber can be satisficed, i.e., can reach a point where the solution's
        /// distance or cost metric is within epsilon of the ideal so as to short-circuit the climbing. 
        /// </summary>
        public bool IsSatisficeable
        {
            get { return !Double.IsNaN(_satisficingCost); }
        }

        /// <summary>
        /// Checks to see whether the given solution is perfect (has zero cost) or has a low enough cost 
        /// that it can satisfice, assuming a satisficing cost has been set for this optimizer.
        /// </summary>
        /// <param name="soln">A potential solution to test.</param>
        /// <param name="state">Optional world state to inform the calculation of cost during the optimization
        /// process.</param>
        /// <remarks>If the cost has already been calculated, callers should use the other version
        /// of IsSolution to avoid re-computing the cost.</remarks>
        /// <returns>True if the given solution is indeed a perfect or satisficing solution to the optimization
        /// problem; false otherwise.</returns>
        protected bool IsSolution(Optimization soln, object state = null)
        {
            double cost = soln.GetCost(state);
            return IsSolution(cost);
        }

        /// <summary>
        /// Checks to see whether the given cost represents a solution by being perfect (having zero cost)
        /// or has low enough cost that it can satisfice, assuming a satisficing cost has been set for 
        /// this optimizer.
        /// </summary>
        /// <param name="cost">The cost of a solution.</param>
        /// <returns>True if the given cost is indeed a perfect or satisficing solution to the optimization
        /// problem; false otherwise.</returns>
        protected bool IsSolution(double cost)
        {
            return (cost == 0.0 || (this.IsSatisficeable && cost <= this.SatisficingCost));
        }

        /// <summary>
        /// The main optimization search procedure for the iterative improvement algorithm. Implementers
        /// can add additional Optimize overloads with additional other parameters.
        /// </summary>
        /// <param name="start">A starting solution from which to begin to search the solution space. Usually
        /// callers will want to start with a random starting solution, which can be obtained using 
        /// AbsOptimizationSolution.Random.</param>
        /// <param name="state">Optional world state to inform the calculation of cost during the optimization
        /// process.</param>
        /// <param name="fn">An optional callback function that reports on the progress made thus far by
        /// the optimization search procedure. Note that the callback is made synchronously on the Optimize 
        /// thread, so if the subscriber intends to carry out expensive operations or update UI controls as 
        /// a result (e.g., a progress bar), a new background thread should be spawned to handle that.</param>
        /// <returns>A solution representing, ideally, a global optimum within the given solution space defined
        /// by the starting solution and the recursive moves which it can produce.</returns>
        /// <remarks>The value reported by the callback function is specific to the type of optimization 
        /// algorithm being used. For a hill climbing algorithm, it might be the number of upward moves made. 
        /// For a simulated annealing algorithm, it might be the decreasing temperature. For a genetic 
        /// algorithm, it might be the number of generations that have transpired.</remarks>
        public abstract Optimization Optimize(Optimization start, object state = null, OptimizationCallback fn = null);
    }
}
