﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.Threading;
using WobbrockLib;
using WobbrockLib.Types;
using WobbrockLib.Extensions;
using WobbrockLib.Windows;

namespace Test
{
	/// <summary>
	/// This testing form shows two different ways to animate a blue circle across the form. In
	/// Stepwise Animation, a for-loop is used to animated based on a loop index and a total number
	/// of desired steps; the more steps, the smoother the animation. This animation is drawn in 
	/// a separate thread so as not to lock up the UI. However, there is no control over the total
	/// duration of the animation with this approach, only how smooth it is. In Timer-Driven Animation,
	/// a timer thread is used to periodically fire off timer events at a desired interval, and to do
	/// so over a total duration. Thus, total animation time is controllable, and smoothness comes
	/// from having a lower timer interval, i.e., higher periodicity.
	/// </summary>
	/// <remarks>This form exercises the AnimationEx extension and its pacer functions.</remarks>
	public partial class AnimationForm : Form
	{
		#region Fields

		private const float Radius = 25.0f;
		private PointF _center;
		private System.Threading.Thread _thread; // animation thread for stepwise animation (for-loop driven)
		private System.Threading.Timer _timer; // animation timer for timer-driven animation
		private long _startTime;
		private int _period;
		private int _duration;

		#endregion

		#region Constructor

		/// <summary>
		/// Default constructor.
		/// </summary>
		public AnimationForm()
		{
			InitializeComponent();
		}

		#endregion

		#region Form Events

		/// <summary>
		/// Set the initial position of the blue circle in the Load event handler.
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void AnimationForm_Load(object sender, EventArgs e)
		{
			_center = new PointF(Radius, this.ClientRectangle.Height / 2.0f - Radius);
		}

		/// <summary>
		/// Catch a close command as soon as we can so we can stop any timers or threads.
		/// </summary>
		/// <param name="m">The Windows message to process.</param>
		protected override void WndProc(ref Message m)
		{
			if ((m.Msg == (int) Win32.WM.SYSCOMMAND) && (m.WParam == (IntPtr) Win32.SC.CLOSE))
			{
				StopTimer();
				StopThread();
			}
			base.WndProc(ref m);
		}

		/// <summary>
		/// Stop the animation timer from firing. Used in Timer-Driven Animation.
		/// </summary>
		private void StopTimer()
		{
			if (_timer != null)
			{
				_timer.Change(Timeout.Infinite, Timeout.Infinite);
				_timer = null;
			}
		}

		/// <summary>
		/// Stop the animation thread from running. Used in Stepwise Animation.
		/// </summary>
		private void StopThread()
		{
			if (_thread != null && _thread.IsAlive)
			{
				try { _thread.Abort(); }
				catch (ThreadAbortException) { _thread = null; }
			}
		}

		/// <summary>
		/// Paint the blue circle based on its center location.
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void AnimationForm_Paint(object sender, PaintEventArgs e)
		{
			e.Graphics.FillEllipse(Brushes.DeepSkyBlue, _center.X - Radius, _center.Y - Radius, Radius * 2.0f, Radius * 2.0f);
			e.Graphics.DrawEllipse(Pens.DodgerBlue, _center.X - Radius, _center.Y - Radius, Radius * 2.0f, Radius * 2.0f);
		}

		#endregion

		#region Stepwise Animation

		private void cmdUniformSteps_Click(object sender, EventArgs e)
		{
			StopThread();
			_thread = new Thread(new ThreadStart(OnAnimateUniformSteps));
			_thread.Start();
		}

		private void cmdSlowInSteps_Click(object sender, EventArgs e)
		{
			StopThread();
			_thread = new Thread(new ThreadStart(OnAnimateSlowInSteps));
			_thread.Start();
		}

		private void cmdSlowOutSteps_Click(object sender, EventArgs e)
		{
			StopThread();
			_thread = new Thread(new ThreadStart(OnAnimateSlowOutSteps));
			_thread.Start();
		}

		private void cmdSlowInOutSteps_Click(object sender, EventArgs e)
		{
			StopThread();
			_thread = new Thread(new ThreadStart(OnAnimateSlowInOutSteps));
			_thread.Start();
		}

		private void OnAnimateUniformSteps()
		{
			_center = new PointF(Radius, this.ClientRectangle.Height / 2.0f - Radius);
			this.Invoke(new MethodInvoker(this.Refresh)); // must invoke on UI thread

			PointF endpoint = new PointF(this.ClientRectangle.Width - Radius, this.ClientRectangle.Height / 2.0f - Radius);
			float distance = (endpoint.X - _center.X);
			int steps = (int) numSteps.Value;

			for (int i = 0; i < steps; i++)
			{
				RectangleR rcOld = new RectangleR(_center.X - Radius, _center.Y - Radius, Radius * 2.0 + 1.0, Radius * 2.0 + 1.0);
				this.Invalidate((Rectangle) rcOld);

				float pct = (float) AnimationEx.Uniform(i, steps);
				_center.X = Radius + (pct * distance);

				RectangleR rcNew = new RectangleR(_center.X - Radius, _center.Y - Radius, Radius * 2.5, Radius * 2.0 + 1.0);
				this.Invalidate((Rectangle) rcNew);
				this.Invoke(new MethodInvoker(this.Update)); // must invoke on UI thread
			}
		}

		private void OnAnimateSlowInSteps()
		{
			_center = new PointF(Radius, this.ClientRectangle.Height / 2.0f - Radius);
			this.Invoke(new MethodInvoker(this.Refresh)); // must invoke on UI thread

			PointF endpoint = new PointF(this.ClientRectangle.Width - Radius, this.ClientRectangle.Height / 2.0f - Radius);
			float distance = (endpoint.X - _center.X);
			int steps = (int) numSteps.Value;

			for (int i = 0; i < steps; i++)
			{
				RectangleR rcOld = new RectangleR(_center.X - Radius, _center.Y - Radius, Radius * 2.0 + 1.0, Radius * 2.0 + 1.0);
				this.Invalidate((Rectangle) rcOld);

				float pct = (float) AnimationEx.SlowIn(i, steps);
				_center.X = Radius + (pct * distance);

				RectangleR rcNew = new RectangleR(_center.X - Radius, _center.Y - Radius, Radius * 2.5, Radius * 2.0 + 1.0);
				this.Invalidate((Rectangle) rcNew);
				this.Invoke(new MethodInvoker(this.Update)); // must invoke on UI thread
			}
		}

		private void OnAnimateSlowOutSteps()
		{
			_center = new PointF(Radius, this.ClientRectangle.Height / 2.0f - Radius);
			this.Invoke(new MethodInvoker(this.Refresh)); // must invoke on UI thread

			PointF endpoint = new PointF(this.ClientRectangle.Width - Radius, this.ClientRectangle.Height / 2.0f - Radius);
			float distance = (endpoint.X - _center.X);
			int steps = (int) numSteps.Value;

			for (int i = 0; i < steps; i++)
			{
				RectangleR rcOld = new RectangleR(_center.X - Radius, _center.Y - Radius, Radius * 2.0 + 1.0, Radius * 2.0 + 1.0);
				this.Invalidate((Rectangle) rcOld);

				float pct = (float) AnimationEx.SlowOut(i, steps);
				_center.X = Radius + (pct * distance);

				RectangleR rcNew = new RectangleR(_center.X - Radius, _center.Y - Radius, Radius * 2.5, Radius * 2.0 + 1.0);
				this.Invalidate((Rectangle) rcNew);
				this.Invoke(new MethodInvoker(this.Update)); // must invoke on UI thread
			}
		}

		private void OnAnimateSlowInOutSteps()
		{
			_center = new PointF(Radius, this.ClientRectangle.Height / 2.0f - Radius);
			this.Invoke(new MethodInvoker(this.Refresh)); // must invoke on UI thread

			PointF endpoint = new PointF(this.ClientRectangle.Width - Radius, this.ClientRectangle.Height / 2.0f - Radius);
			float distance = (endpoint.X - _center.X);
			int steps = (int) numSteps.Value;

			for (int i = 0; i < steps; i++)
			{
				RectangleR rcOld = new RectangleR(_center.X - Radius, _center.Y - Radius, Radius * 2.0 + 1.0, Radius * 2.0 + 1.0);
				this.Invalidate((Rectangle) rcOld);

				float pct = (float) AnimationEx.SlowInOut(i, steps);
				_center.X = Radius + (pct * distance);

				RectangleR rcNew = new RectangleR(_center.X - Radius, _center.Y - Radius, Radius * 2.5, Radius * 2.0 + 1.0);
				this.Invalidate((Rectangle) rcNew);
				this.Invoke(new MethodInvoker(this.Update)); // must invoke on UI thread
			}
		}

		#endregion

		#region Timer-Driven Animation

		private void numPeriod_ValueChanged(object sender, EventArgs e)
		{
			if (numPeriod.Value > numDuration.Value)
				numPeriod.Value = numDuration.Value;
			lblHertz.Text = String.Format("= ~{0:g3} cycles / second (Hz)", 1000m / numPeriod.Value);
		}

		private void numDuration_ValueChanged(object sender, EventArgs e)
		{
			if (numDuration.Value < numPeriod.Value)
				numDuration.Value = numPeriod.Value;
			lblSeconds.Text = String.Format("= ~{0:g3} total seconds", numDuration.Value / 1000m);
		}

		private void cmdUniformTimer_Click(object sender, EventArgs e)
		{
			StopTimer();

			_center = new PointF(Radius, this.ClientRectangle.Height / 2.0f - Radius);
			this.Refresh();

			_period = (int) numPeriod.Value;
			_duration = (int) numDuration.Value;
			_startTime = TimeEx.NowMs;
			_timer = new System.Threading.Timer(new TimerCallback(OnAnimateUniformTimer), this, _period, Timeout.Infinite);
		}

		private void cmdSlowInTimer_Click(object sender, EventArgs e)
		{
			StopTimer();

			_center = new PointF(Radius, this.ClientRectangle.Height / 2.0f - Radius);
			this.Refresh();

			_period = (int) numPeriod.Value;
			_duration = (int) numDuration.Value;
			_startTime = TimeEx.NowMs;
			_timer = new System.Threading.Timer(new TimerCallback(OnAnimateSlowInTimer), this, _period, Timeout.Infinite);
		}

		private void cmdSlowOutTimer_Click(object sender, EventArgs e)
		{
			StopTimer();

			_center = new PointF(Radius, this.ClientRectangle.Height / 2.0f - Radius);
			this.Refresh();

			_period = (int) numPeriod.Value;
			_duration = (int) numDuration.Value;
			_startTime = TimeEx.NowMs;
			_timer = new System.Threading.Timer(new TimerCallback(OnAnimateSlowOutTimer), this, _period, Timeout.Infinite);
		}

		private void cmdSlowInOutTimer_Click(object sender, EventArgs e)
		{
			StopTimer();

			_center = new PointF(Radius, this.ClientRectangle.Height / 2.0f - Radius);
			this.Refresh();

			_period = (int) numPeriod.Value;
			_duration = (int) numDuration.Value;
			_startTime = TimeEx.NowMs;
			_timer = new System.Threading.Timer(new TimerCallback(OnAnimateSlowInOutTimer), this, _period, Timeout.Infinite);
		}

		private void OnAnimateUniformTimer(object state)
		{
			_timer.Change(Timeout.Infinite, Timeout.Infinite); // stop
			long now = TimeEx.NowMs;

			RectangleR rcOld = new RectangleR(_center.X - Radius, _center.Y - Radius, Radius * 2.0 + 1.0, Radius * 2.0 + 1.0);
			this.Invalidate((Rectangle) rcOld);

			float pct = (float) AnimationEx.Uniform(_startTime, now, _duration);

			PointF startpoint = new PointF(Radius, this.ClientRectangle.Height / 2.0f - Radius);
			PointF endpoint = new PointF(this.ClientRectangle.Width - Radius, this.ClientRectangle.Height / 2.0f - Radius);
			float distance = (endpoint.X - startpoint.X);

			_center.X = Radius + (pct * distance);

			RectangleR rcNew = new RectangleR(_center.X - Radius, _center.Y - Radius, Radius * 2.5, Radius * 2.0 + 1.0);
			this.Invalidate((Rectangle) rcNew);
			try { this.Invoke(new MethodInvoker(this.Update)); } // must invoke on UI thread
			catch (ObjectDisposedException) { }

			if (pct < 1.00 && _timer != null)
				_timer.Change(_period, Timeout.Infinite); // restart
		}

		private void OnAnimateSlowInTimer(object state)
		{
			_timer.Change(Timeout.Infinite, Timeout.Infinite); // stop
			long now = TimeEx.NowMs;

			RectangleR rcOld = new RectangleR(_center.X - Radius, _center.Y - Radius, Radius * 2.0 + 1.0, Radius * 2.0 + 1.0);
			this.Invalidate((Rectangle) rcOld);

			float pct = (float) AnimationEx.SlowIn(_startTime, now, _duration);

			PointF startpoint = new PointF(Radius, this.ClientRectangle.Height / 2.0f - Radius);
			PointF endpoint = new PointF(this.ClientRectangle.Width - Radius, this.ClientRectangle.Height / 2.0f - Radius);
			float distance = (endpoint.X - startpoint.X);

			_center.X = Radius + (pct * distance);

			RectangleR rcNew = new RectangleR(_center.X - Radius, _center.Y - Radius, Radius * 2.5, Radius * 2.0 + 1.0);
			this.Invalidate((Rectangle) rcNew);
			try { this.Invoke(new MethodInvoker(this.Update)); } // must invoke on UI thread
			catch (ObjectDisposedException) { }

			if (pct < 1.00 && _timer != null)
				_timer.Change(_period, Timeout.Infinite); // restart
		}

		private void OnAnimateSlowOutTimer(object state)
		{
			_timer.Change(Timeout.Infinite, Timeout.Infinite); // stop
			long now = TimeEx.NowMs;

			RectangleR rcOld = new RectangleR(_center.X - Radius, _center.Y - Radius, Radius * 2.0 + 1.0, Radius * 2.0 + 1.0);
			this.Invalidate((Rectangle) rcOld);

			float pct = (float) AnimationEx.SlowOut(_startTime, now, _duration);

			PointF startpoint = new PointF(Radius, this.ClientRectangle.Height / 2.0f - Radius);
			PointF endpoint = new PointF(this.ClientRectangle.Width - Radius, this.ClientRectangle.Height / 2.0f - Radius);
			float distance = (endpoint.X - startpoint.X);

			_center.X = Radius + (pct * distance);

			RectangleR rcNew = new RectangleR(_center.X - Radius, _center.Y - Radius, Radius * 2.5, Radius * 2.0 + 1.0);
			this.Invalidate((Rectangle) rcNew);
			try { this.Invoke(new MethodInvoker(this.Update)); } // must invoke on UI thread
			catch (ObjectDisposedException) { }

			if (pct < 1.00 && _timer != null)
				_timer.Change(_period, Timeout.Infinite); // restart
		}

		private void OnAnimateSlowInOutTimer(object state)
		{
			_timer.Change(Timeout.Infinite, Timeout.Infinite); // stop
			long now = TimeEx.NowMs;

			RectangleR rcOld = new RectangleR(_center.X - Radius, _center.Y - Radius, Radius * 2.0 + 1.0, Radius * 2.0 + 1.0);
			this.Invalidate((Rectangle) rcOld);

			float pct = (float) AnimationEx.SlowInOut(_startTime, now, _duration);

			PointF startpoint = new PointF(Radius, this.ClientRectangle.Height / 2.0f - Radius);
			PointF endpoint = new PointF(this.ClientRectangle.Width - Radius, this.ClientRectangle.Height / 2.0f - Radius);
			float distance = (endpoint.X - startpoint.X);

			_center.X = Radius + (pct * distance);

			RectangleR rcNew = new RectangleR(_center.X - Radius, _center.Y - Radius, Radius * 2.5, Radius * 2.0 + 1.0);
			this.Invalidate((Rectangle) rcNew);
			try { this.Invoke(new MethodInvoker(this.Update)); } // must invoke on UI thread
			catch (ObjectDisposedException) { }

			if (pct < 1.00 && _timer != null)
				_timer.Change(_period, Timeout.Infinite); // restart
		}

		#endregion

		
	}
}
