﻿using System;
using System.Collections.Generic;
using System.Text;
using WobbrockLib.Extensions;
using WobbrockLib.Optimization.Genetic;

namespace Test
{
    /// <summary>
    /// Implements an organism that encodes a string for the purposes of matching
    /// it with a target string the ecosystem is seeking. The closer the encoded
    /// string to the target string, the more fit this organism is. Note that
    /// a mutuble character array is used in lieu of an immutable C# string to make
    /// mutation of individual letters easier.
    /// </summary>
    public class StringSeekingOrganism : Organism
    {
        private char[] _s; // mutable string
        private int _minLen;
        private int _maxLen;

        /// <summary>
        /// Constructs a new organism with the given generation number and
        /// minimum and maximum possible string lengths. The string created
        /// is random.
        /// </summary>
        /// <param name="generation">The generation number of this organism.</param>
        /// <param name="minLen">The minimum allowed length of the represented string.</param>
        /// <param name="maxLen">The maximum allowed length of the represented string.</param>
        public StringSeekingOrganism(int generation, int minLen, int maxLen)
        {
            _minLen = minLen;
            _maxLen = maxLen;
            int length = RandomEx.Integer(minLen, maxLen);
            _s = new char[length];
            for (int i = 0; i < length; i++)
                _s[i] = (char) RandomEx.Integer((int) 'a', (int) 'z');
            _generation = generation;
        }

        /// <summary>
        /// Constructs a new organism with the given generation number and string.
        /// This constructor can be used as a copy constructor.
        /// </summary>
        /// <param name="generation">The generation number of this organism.</param>
        /// <param name="s">A mutable character-array string to represent with this organism.</param>
        public StringSeekingOrganism(int generation, char[] s)
        {
            _s = new char[s.Length];
            Array.Copy(s, _s, s.Length);
            _generation = generation;
        }

        /// <summary>
        /// Copies this organism.
        /// </summary>
        /// <returns>A deep copy of this organism.</returns>
        public override Organism Copy()
        {
            return new StringSeekingOrganism(_generation, _s);
        }

        /// <summary>
        /// Gets the value of this string as currently encoded by a character array.
        /// </summary>
        public override object DNA
        {
            get { return _s; }
        }

        /// <summary>
        /// Provides a string representation of this organism, which is the
        /// string this organism encodes.
        /// </summary>
        /// <returns>The string represented by this organism.</returns>
        public override string ToString()
        {
            return new string(_s);
        }

        /// <summary>
        /// Produces a random first-generation string between the minimum and maximum
        /// lengths provided to this organism's contructor.
        /// </summary>
        /// <returns>A new first-generation organism randomly generated containing a string
        /// between the minimum and maximum length's provided to this organism's constructor.</returns>
        public override Organism BirthRandom(object state = null)
        {
            StringSeekingOrganism o = new StringSeekingOrganism(1, _minLen, _maxLen); // first generation is random

            // compute and set fitness now
            if (state != null)
                o._fitness = o.ComputeFitness(state);

            return o;
        }

        /// <summary>
        /// Produces a new Nth-generation string based on this organism. This parent string
        /// will be mutated 'n' times to produce the offspring. This organism, the parent,
        /// remains unchanged.
        /// </summary>
        /// <param name="nMutations">The number of times this string is mutated to produce an offspring.</param>
        /// <param name="state">The world state used to compute the fitness of this organism. If null, the
        /// fitness is not set at birth, and so calls to GetFitness will need to pass the world state to
        /// compute it.</param>
        /// <returns>A new organism derived from the current one.</returns>
        public override Organism BirthMutated(int nMutations, object state = null)
        {
            StringSeekingOrganism o = (StringSeekingOrganism) this.Copy();
            o._generation = _generation + 1; // next generation
            for (int i = 0; i < nMutations; i++)
            {
                // random chance of being 1 char less or 1 char more
                int lenAdjust = 0;
                if (RandomEx.Integer(1, 100) <= 10) // chance of a possible length adjustment
                {
                    lenAdjust = RandomEx.Integer(-1, +1);
                }
                if (lenAdjust == -1 && o._s.Length > 1)
                {
                    char[] tmp = new char[o._s.Length - 1];
                    Array.Copy(o._s, tmp, o._s.Length - 1);
                    o._s = tmp; // update reference
                }
                else if (lenAdjust == +1)
                {
                    char[] tmp = new char[o._s.Length + 1];
                    Array.Copy(o._s, tmp, o._s.Length);
                    tmp[o._s.Length] = (char) RandomEx.Integer((int) 'a', (int) 'z'); // set random letter
                    o._s = tmp; // update reference
                }
                // random mutation equals flip a random letter randomly
                int pos = RandomEx.Integer(0, o._s.Length - 1); // random index
                o._s[pos] = (char) RandomEx.Integer((int) 'a', (int) 'z'); // change to a random letter
            }

            // compute and set fitness now, if the world state was provided
            if (state != null)
                o._fitness = o.ComputeFitness(state);

            return o;
        }

        /// <summary>
        /// Gets the lenth of the string represented by this organism.
        /// </summary>
        public override int Size
        {
            get { return _s.Length; }
        }

        /// <summary>
        /// Indexer to get or set a specific character in this organism's string.
        /// </summary>
        /// <param name="index">A 0-based index at which to get or set a character.</param>
        /// <returns>An object that can be cast to a char type.</returns>
        public override object this[int index]
        {
            get { return _s[index]; }
            set { _s[index] = (char) value; }
        }

        /// <summary>
        /// Computes the fitness of this organism defined as the complement of the uncorrected text 
        /// entry error rate from Soukoreff and MacKenzie (CHI 2003). The measure uses the Levenshtein
        /// minimum string distance (MSD) metric.
        /// </summary>
        /// <param name="state">The objective goal state, namely, the string we're seeking to converge 
        /// to.</param>
        /// <returns>The fitness as measured by the complement of the uncorrected error rate 
        /// (Soukoreff and MacKenzie 2003).</returns>
        public override double ComputeFitness(object state)
        {
            string target = (string) state;
            string str = this.ToString();
            double errors = (double) MSD(str, target) / Math.Max(str.Length, target.Length);
            return (1.0 - errors);
        }

        /// <summary>
        /// Gets whether or not this organism is fit enough to warrant terminating the evolution process
        /// and no longer seek a more perfect organism. We define fit enough here to be perfect, that is,
        /// the target string has been found.
        /// </summary>
        /// <param name="state">The target string that must be found.</param>
        /// <returns>True if this organism encodes the target string; false otherwise.</returns>
        public override bool IsFitEnough(object state = null)
        {
            return this.ToString() == (string) state;
        }

        /// <summary>
        /// Gets whether or not this organism has converged with a given organism.
        /// </summary>
        /// <param name="o">The given organism.</param>
        /// <param name="state">The target string that must be found.</param>
        /// <returns>True if convergence has occurred; false otherwise.</returns>
        public override bool HasConvergedWith(Organism o, object state = null)
        {
            StringSeekingOrganism sso = (StringSeekingOrganism) o;
            if (_s.Length == sso._s.Length)
            {
                for (int i = 0; i < _s.Length; i++)
                {
                    if (_s[i] != sso._s[i])
                        return false;
                }
                return true;
            }
            return false;
        }

        /// <summary>
        /// The Levenshtein minimum string distance (MSD) metric computed as
        /// the number of insertions, omissions, or substitutions describing
        /// the difference between string a and b.
        /// </summary>
        /// <param name="a">The first string.</param>
        /// <param name="b">The second string.</param>
        /// <returns>The Levensthein minimum string distance between the strings.</returns>
        private int MSD(string a, string b)
        {
            int[,] D = new int[a.Length + 1, b.Length + 1];
            for (int i = 0; i <= a.Length; i++)
                D[i, 0] = i;
            for (int j = 0; j <= b.Length; j++)
                D[0, j] = j;
            for (int i = 1; i <= a.Length; i++)
                for (int j = 1; j <= b.Length; j++)
                    D[i, j] = Math.Min(
                                    Math.Min(D[i - 1, j] + 1, D[i, j - 1] + 1),
                                D[i - 1, j - 1] + (a[i - 1] == b[j - 1] ? 0 : 1)
                                );
            return D[a.Length, b.Length];
        }
    }
}
