﻿using System;
using System.Text;
using System.Collections.Generic;
using System.Threading;

namespace WobbrockLib.Windows
{
	/// <summary>
	/// Abstract base class for UI Automation menu event listeners. Clients that wish to use UI Automation may 
	/// need to have an app.manifest file that has the uiAccess property set to "true". This, in turn, requires
	/// digitally signing the application and running it from a trusted directory like C:\Windows\System32 or
	/// C:\Program Files (x86)\...
	/// </summary>
	/// <see cref="https://docs.microsoft.com/en-us/windows/win32/winauto/uiauto-securityoverview"/>
	/// <see cref="https://docs.microsoft.com/en-us/visualstudio/ide/how-to-sign-application-and-deployment-manifests?view=vs-2017"/>
	/// <see cref="https://stackoverflow.com/questions/8671579/what-is-the-difference-between-asinvoker-and-highestavailable-execution-leve"/>
	/// <see cref="https://stackoverflow.com/questions/84847/how-do-i-create-a-self-signed-certificate-for-code-signing-on-windows"/>
	/// <see cref="https://blog.frogslayer.com/windows-store-applications-and-automated-testing-part-three/"/>
	/// <see cref="https://social.msdn.microsoft.com/Forums/windowsdesktop/en-US/7a42efab-5ce8-456f-8a58-dfedbc2cefcb/debugging-with-uiaccesstrue?forum=windowsaccessibilityandautomation"/>
	public abstract class UIAMenus
	{
		#region Types

		/// <summary>
		/// Indicates a menu action. Implemented as a class so MenuEventEx can inhert.
		/// </summary>
		public class MenuEvent
		{
			public const uint None = 0x00;
			public const uint Open = 0x01;
			public const uint Close = 0x02;
			public static string ToString(uint id)
			{
				switch (id)
				{
					case None:
						return "None";
					case Open:
						return "Open";
					case Close:
						return "Close";
					default:
						return String.Empty;
				}
			}
		}

		/// <summary>
		/// Type indicating the user interface tree that will be registered.
		/// </summary>
		public enum UITree
		{
			None = 0,       // TreeScope.None
			Self = 1,       // TreeScope.Element => self only
			Kids = 2,       // TreeScope.Children => immediate children
			SelfKids = 3,   // TreeScope.Element | TreeScope.Children
			Subtree = 7     // TreeScope.Subtree => self plus all descendants
		}

		/// <summary>
		/// Struct for holding information for registering a menu listener.
		/// </summary>
		protected struct RegState
		{
			public static readonly RegState Empty;
			public IntPtr Hwnd;		// register to listen on this window
			public uint Events;		// register for these menu events
			public UITree Tree;		// register for this UI tree scope
		}

		#endregion

		#region Events

		/// <summary>
		/// Event for when a menu is opened on any registered window.
		/// </summary>
		public event EventHandler<IntPtr> MenuOpened;

		/// <summary>
		/// Event for when a menu is closed on any registered window.
		/// </summary>
		public event EventHandler<IntPtr> MenuClosed;

		#endregion

		#region Fields

		/// <summary>
		/// Keep track of registered window handles for menu events and prevent the same
		/// window from being registered more than once with the same window handle.
		/// </summary>
		protected readonly List<KeyValuePair<IntPtr, uint>> _listeners;

		/// <summary>
		/// Object used with the "lock" keyword, which provides synrchonization via the
		/// underlying Monitor class to the UI Automation registering and unregistering
		/// methods.
		/// </summary>
		protected readonly Object _lockObj;

		#endregion

		#region Constructor

		/// <summary>
		/// Default constructor.
		/// </summary>
		public UIAMenus()
		{
			_listeners = new List<KeyValuePair<IntPtr, uint>>();
			_lockObj = new Object();
		}

		/// <summary>
		/// Destructor.
		/// </summary>
		/// <remarks>Destructors are run on the system's finalizer thread, not the client's main UI thread, 
		/// so we can directly call our unregistration method, rather than spawning a new thread to do the
		/// work.</remarks>
		/// <see cref="https://stackoverflow.com/questions/1259628/when-is-my-destructor-called-in-this-circumstance-c"/>
		~UIAMenus()
		{
			if (_listeners?.Count > 0)
			{
				UIAUnregister(); // direct call to subclass impl.
			}
		}

		#endregion

		#region Registering/Unregistering

		/// <summary>
		/// Register for MenuOpened and MenuClosed events through UI Automation.
		/// </summary>
		/// <param name="hWnd">A handle to the window or control on which to listen for menu events.</param>
		/// <param name="menuEvents">Bitwise values indicating which menu event(s) to listen for.</param>
		/// <param name="tree">Where in the UI tree to listen for the menu events relative to hWnd.</param>
		/// <param name="async">If true, this method returns immediately after spawning a thread to carry
		/// out the work. If false, the calling thread waits for this method to completely finish.</param>
		/// <remarks>This method spawns a new background thread to register the new menu listener via 
		/// UI Automation. Also, it is important that it uses an MTA apartment state.</remarks>
		/// <see cref="https://docs.microsoft.com/en-us/dotnet/framework/ui-automation/ui-automation-threading-issues"/>
		/// <see cref="https://stackoverflow.com/questions/4154429/apartmentstate-for-dummies"/>
		public virtual void RegisterListener(IntPtr hWnd, uint menuEvents, UITree tree, bool async)
		{
			Thread thread = new Thread(new ParameterizedThreadStart(UIARegister));
			thread.Name = "Register Listener Thread";
			thread.SetApartmentState(ApartmentState.MTA);
			thread.IsBackground = true;

			RegState reg = RegState.Empty;
			reg.Hwnd = hWnd;
			reg.Events = menuEvents;
			reg.Tree = tree;
			thread.Start(reg);

			if (!async)
				thread.Join(); // blocks calling thread until 'thread' finishes
		}

		/// <summary>
		/// Unregister all of the event handlers that have been registered.
		/// </summary>
		/// <param name="async">If true, this method returns immediately after spawning a thread to carry
		/// out the work. If false, the calling thread waits for this method to completely finish.</param>
		/// <remarks>This method spawns a new background thread to unregister all menu listeners via 
		/// UI Automation. Also, it is important that it uses an MTA apartment state.</remarks>
		/// <see cref="https://docs.microsoft.com/en-us/dotnet/framework/ui-automation/ui-automation-threading-issues"/>
		/// <see cref="https://stackoverflow.com/questions/4154429/apartmentstate-for-dummies"/>
		public virtual void UnregisterListeners(bool async)
		{
			Thread thread = new Thread(new ThreadStart(UIAUnregister));
			thread.Name = "Unregister Listeners Thread";
			thread.SetApartmentState(ApartmentState.MTA);
			thread.IsBackground = true;
			thread.Start();

			if (!async)
				thread.Join(); // blocks calling thread until 'thread' finishes
		}

		/// <summary>
		/// Pure virtual helper method to be implemented by subclasses for registering for a 
		/// menu event lister via UI Automation.
		/// </summary>
		/// <param name="state">The state information necessary to complete the registration.
		/// This is a struct of type RegState.</param>
		/// <see cref="https://docs.microsoft.com/en-us/dotnet/framework/ui-automation/ui-automation-threading-issues"/>
		protected abstract void UIARegister(object state);

		/// <summary>
		/// Pure virtual helper method to be implemented by subclasses for unregistering for
		/// all menu event listeners via UI Automation.
		/// </summary>
		/// <see cref="https://docs.microsoft.com/en-us/dotnet/framework/ui-automation/ui-automation-threading-issues"/>
		protected abstract void UIAUnregister();

		#endregion

		#region Current Listening

		/// <summary>
		/// Gets the number of menu event listeners currently active.
		/// </summary>
		/// <remarks>This getter is not processed on a separate thread, but we do not want
		/// to use 'lock' because it could lock up the main UI thread, or whatever other 
		/// client thread is doing the calling. Thus, this getter should be thought of as 
		/// getting the number of listeners right when it's called. A pending registration
		/// of a new listener could be in progress.</remarks>
		public int Listeners
		{
			get
			{
				return _listeners.Count;
			}
		}

		/// <summary>
		/// Gets a string showing current menu listeners.
		/// </summary>
		/// <remarks>This getter is not processed on a separate thread, but we do not want
		/// to use 'lock' because it could lock up the main UI thread, or whatever other 
		/// client thread is doing the calling. Thus, this getter should be thought of as 
		/// getting the string list of listeners right when it's called. A pending registration
		/// of a new listener could be in progress.</remarks>
		public string NowListening
		{
			get
			{
				StringBuilder sb = new StringBuilder("Listeners: ");
				if (_listeners?.Count > 0)
				{
					foreach (KeyValuePair<IntPtr, uint> kvp in _listeners)
					{
						sb.AppendFormat("<{0}:{1}>; ", MenuEventName(kvp.Value), kvp.Key);
					}
				}
				else // null or zero
				{
					sb.Append("None.");
				}
				return sb.ToString();
			}
		}

		/// <summary>
		/// Gets the menu event name for the given menu event.
		/// </summary>
		/// <param name="menuEvent">The given menu event value.</param>
		/// <returns>The string name representation for this event.</returns>
		protected virtual string MenuEventName(uint menuEvent)
		{
			return MenuEvent.ToString(menuEvent);
		}

		/// <summary>
		/// Tests whether the the current instance is listening to the given window for the given
		/// event. The given event should be a single event, not a composite of MenuEvent and/or
		/// MenuEventEx flags. If it is a composite, false will be returned even if the individual
		/// events are indeed being listened to.
		/// </summary>
		/// <param name="hWnd">The window handle on which we might be currently listening.</param>
		/// <param name="menuEvent">The MenuEvent or MenuEventEx value to inquire about. This value
		/// must be a single event flag, not a composite of multiple flags.</param>
		/// <returns>True if we're listening to this window for this event; false otherwise.</returns>
		public bool ListeningFor(IntPtr hWnd, uint menuEvent)
		{
			foreach (KeyValuePair<IntPtr, uint> kvp in _listeners)
			{
				if (kvp.Key == hWnd && kvp.Value == menuEvent)
					return true;
			}
			return false;
		}

		#endregion

		#region Raising Events

		/// <summary>
		/// Raises the MenuOpened event. Called by subclasses.
		/// </summary>
		/// <param name="e">The arguments for this event, in this case, an IntPtr to a window handle
		/// on which the menu was opened.</param>
		/// <remarks>These events are raised when UI Automation events are fired. Those events are 
		/// fired in their own separate thread. Clients of these events will need to be aware of this
		/// to avoid cross-threaded calls, e.g., to update their UI on this event.</remarks>
		protected virtual void OnMenuOpened(IntPtr e)
		{
			this.MenuOpened?.BeginInvoke(this, e, null, null);
		}

		/// <summary>
		/// Raises the MenuClosed event. Called by subclasses.
		/// </summary>
		/// <param name="e">The arguments for this event, in this case, an IntPtr to a window handle
		/// on which the menu was closed.</param>
		/// <remarks>These events are raised when UI Automation events are fired. Those events are 
		/// fired in their own separate thread. Clients of these events will need to be aware of this
		/// to avoid cross-threaded calls, e.g., to update their UI on this event.</remarks>
		protected virtual void OnMenuClosed(IntPtr e)
		{
			this.MenuClosed?.BeginInvoke(this, e, null, null);
		}

		#endregion
	}
}
