﻿using System;
using System.Diagnostics;
using System.Collections.Generic;
using Interop.UIAutomationCore; // COM Interop DLL, not managed .NET DLL

namespace WobbrockLib.Windows
{
	/// <summary>
	/// A class that encapsulates UI Automation via COM Interop for working with menus. Although the Systems.Windows.Automation 
	/// namespace provides managed .NET methods for doing so, it has problems in terms of bugs and lag. Therefore, we access this 
	/// functionality via COM Interop.
	/// </summary>
	/// <see cref="https://stackoverflow.com/questions/41768046/system-windows-automation-is-extremely-slow"/>
	/// <see cref="https://stackoverflow.com/questions/38217930/how-to-use-iuiautomationelementfromiaccessible-method-in-c"/>
	/// /// <example>
	///		A client should first declare a member instance for UIAMenus:
	///		<code>
	///			private readonly UIAComMenus _uiaMenus;
	///		</code>
	///		
	///		Then it should construct the instance and register for events:
	///		<code>
	///			_uiaMenus = new UIAComMenus();
	///			_uiaMenus.MenuOpened += new EventHandler&lt;IntPtr&gt;(SomeWindow_MenuOpened);
	///			_uiaMenus.MenuClosed += new EventHandler&lt;IntPtr&gt;(SomeWindow_MenuClosed);
	///		</code>
	///		
	///		Then it should register the window or control it wants to listen to for menu open/close events:
	///		<code>
	///			_uiaMenus.RegisterListener(hWnd, UIAMenus.MenuEvent.OpenClose);
	///			Console.WriteLine("No. of listeners is {0}.", _uiaMenus.Listeners);
	///		</code>
	///		
	///		Then it should handle events when they occur:
	///		<code>
	///			private void SomeWindow_MenuOpened(object sender, IntPtr hWnd)
	///			{
	///				// ...
	///			}
	///			private void SomeWindow_MenuClosed(object sender, IntPtr hWnd)
	///			{
	///				// ...
	///			}
	///		</code>
	///		
	///		Finally, it should unregister when done:
	///		<code>
	///			private void MyForm_FormClosed(object sender, EventArgs e)
	///			{
	///				_uiaMenus.UnregisterListeners();
	///			}
	///		</code>
	/// </example>
	public class UIAComMenus : UIAMenus, IUIAutomationEventHandler
	{
		#region Events

		/// <summary>
		/// Event for when a menu is opened on any registered window.
		/// </summary>
		//public event EventHandler<IntPtr> MenuModeStarted;

		/// <summary>
		/// Event for when a menu is closed on any registered window.
		/// </summary>
		//public event EventHandler<IntPtr> MenuModeEnded;

		#endregion

		#region Fields

		// https://docs.microsoft.com/en-us/windows/win32/winauto/uiauto-event-ids
		private const int MenuOpenedEventId = 20003;			// UIA_EventIds.UIA_MenuOpenedEventId
		private const int MenuClosedEventId = 20007;			// UIA_EventIds.UIA_MenuClosedEventId
		private const int MenuModeStartEventId = 20018;			// UIA_EventIds.UIA_MenuModeStartEventId
		private const int MenuModeEndEventId = 20019;           // UIA_EventIds.UIA_MenuModeEndEventId

		// https://docs.microsoft.com/en-us/windows/win32/winauto/uiauto-automation-element-propids
		private const int ControlTypePropertyId = 30003;		// UIA_PropertyIds.UIA_ControlTypePropertyId
		private const int AcceleratorKeyPropertyId = 30006;		// UIA_PropertyIds.UIA_AcceleratorKeyPropertyId
		private const int AccessKeyPropertyId = 30007;			// UIA_PropertyIds.UIA_AccessKeyPropertyId
		private const int NativeWindowHandlePropertyId = 30020; // UIA_PropertyIds.UIA_NativeWindowHandlePropertyId

		// https://docs.microsoft.com/en-us/windows/win32/winauto/uiauto-controltype-ids
		private const int MenuControlTypeId = 50009;			// UIA_ControlTypeIds.UIA_MenuControlTypeId
		private const int MenuBarControlTypeId = 50010;			// UIA_ControlTypeIds.UIA_MenuBarControlTypeId
		private const int MenuItemControlTypeId = 50011;		// UIA_ControlTypeIds.UIA_MenuItemControlTypeId

		/// <summary>
		/// Our instance of a UI Automation object that gives us access to automation 
		/// events, methods, and properties.
		/// </summary>
		private readonly CUIAutomation _uia;

		#endregion

		#region Constructor

		/// <summary>
		/// Default constructor.
		/// </summary>
		public UIAComMenus()
		{
			_uia = new CUIAutomation();
		}

		/// <summary>
		/// Destructor.
		/// </summary>
		~UIAComMenus()
		{
			if (_listeners?.Count > 0)
			{
				_uia?.RemoveAllEventHandlers();
				_listeners.Clear();
			}
		}

		#endregion

		#region UIA Menu Listening

		/// <summary>
		/// Register for MenuOpened and MenuClosed events through UI Automation.
		/// </summary>
		/// <param name="hWnd">A handle to the window or control on which to listen for menu events.</param>
		/// <param name="what">What menu events to listen for: opened, closed, or both.</param>
		/// <param name="where">Where to listen for the menu events in the UI tree.</param>
		/// <returns>True if we successfully register for the events; false otherwise.</returns>
		/// <seealso cref="http://geekswithblogs.net/CalinTatar/archive/2009/05/16/switching-from-win32-to-ui-automation.aspx"/>
		/// <seealso cref="https://docs.microsoft.com/en-us/visualstudio/debugger/just-my-code?view=vs-2019"/>
		/// <seealso cref="https://stackoverflow.com/questions/32347734/ui-automation-events-stop-being-received-after-a-while-monitoring-an-application"/>
		public override bool RegisterListener(IntPtr hWnd, MenuEvent what, UITree where)
		{
			if (what == MenuEvent.None)
				return false;

			bool hasOpen = false;
			bool hasClose = false;
			bool wantOpen = ((what & MenuEvent.Open) != 0);
			bool wantClose = ((what & MenuEvent.Close) != 0);

			foreach (KeyValuePair<IntPtr, MenuEvent> kvp in _listeners)
			{
				if (wantOpen && !hasOpen)
				{
					if (kvp.Key == hWnd && kvp.Value == MenuEvent.Open)
						hasOpen = true;
				}
				if (wantClose && !hasClose)
				{
					if (kvp.Key == hWnd && kvp.Value == MenuEvent.Close)
						hasClose = true;
				}
				if (hasOpen && hasClose)
					break;
			}
			if ((wantOpen && !hasOpen) || (wantClose && !hasClose))
			{
				IUIAutomationElement element = null;
				try
				{
					IUIAutomationCacheRequest cache = _uia.CreateCacheRequest();
					cache.TreeScope = TreeScope.TreeScope_Element;
					cache.AddProperty(NativeWindowHandlePropertyId);
					element = _uia.ElementFromHandleBuildCache(hWnd, cache);
					if (element != null)
					{
						if (wantOpen && !hasOpen)
						{
							_uia.AddAutomationEventHandler(MenuOpenedEventId, element, (TreeScope) where, cache, this); // MenuStrips
							//_uia.AddAutomationEventHandler(MenuModeStartEventId, element, (TreeScope) where, cache, this); // Win32
							_listeners.Add(new KeyValuePair<IntPtr, MenuEvent>(hWnd, MenuEvent.Open));
							Debug.WriteLine(String.Format("Listener++: \"{0}\": <Open:{1}>", WindowEx.GetWindowTitle(hWnd), hWnd));
						}
						if (wantClose && !hasClose)
						{
							_uia.AddAutomationEventHandler(MenuClosedEventId, element, (TreeScope) where, cache, this); // MenuStrips
							//_uia.AddAutomationEventHandler(MenuModeEndEventId, element, (TreeScope) where, cache, this); // Win32
							_listeners.Add(new KeyValuePair<IntPtr, MenuEvent>(hWnd, MenuEvent.Close));
							Debug.WriteLine(String.Format("Listener++: \"{0}\": <Close:{1}>", WindowEx.GetWindowTitle(hWnd), hWnd));
						}
					}
				}
				catch (Exception ex)
				{
					Console.WriteLine(ex);
				}
				return (element != null);
			}
			return false;
		}

		/// <summary>
		/// Unregister all of the event handlers that have been registered.
		/// </summary>
		/// <seealso cref="https://docs.microsoft.com/en-us/visualstudio/debugger/just-my-code?view=vs-2019"/>
		public override void UnregisterListeners()
		{
			try
			{
				if (_listeners?.Count > 0)
				{
					_uia?.RemoveAllEventHandlers();
					_listeners.Clear();
					Debug.WriteLine("Listeners: Cleared.");
				}
			}
			catch (Exception ex)
			{
				Console.WriteLine(ex);
			}
		}

		/// <summary>
		/// Handles when a menu is opened or closed on a registered window.
		/// </summary>
		/// <param name="sender">The element that generated this event.</param>
		/// <param name="eventId">The event identifier for this event.</param>
		/// <seealso cref="https://stackoverflow.com/questions/32347734/ui-automation-events-stop-being-received-after-a-while-monitoring-an-application"/>
		public void HandleAutomationEvent(IUIAutomationElement sender, int eventId)
		{
			IntPtr hWnd = IntPtr.Zero;
			try
			{
				if (sender != null)
				{
					hWnd = sender.CachedNativeWindowHandle;
				}
			}
			catch (Exception ex)
			{
				Console.WriteLine(ex);
			}
			// fire the event that occurred
			switch (eventId)
			{
				case MenuOpenedEventId:
					OnMenuOpened(hWnd);
					break;
				case MenuClosedEventId:
					OnMenuClosed(hWnd);
					break;
				case MenuModeStartEventId:
					OnMenuOpened(hWnd);
					break;
				case MenuModeEndEventId:
					OnMenuClosed(hWnd);
					break;
				default:
					// do nothing
					break;
			}
		}
		
		#endregion
	}
}
