﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Text;
using System.Windows.Automation;

namespace WobbrockLib.Windows
{
	/// <summary>
	/// Class that encapsulates UIAutomation for working with menus in .NET managed code.
	/// </summary>
	/// <example>
	///		A client should first declare a member instance for UIAMenus:
	///		<code>
	///			private readonly UIAMenus _uiaMenus;
	///		</code>
	///		
	///		Then it should construct the instance and register for events:
	///		<code>
	///			_uiaMenus = new UIAMenus();
	///			_uiaMenus.MenuOpened += new EventHandler&lt;IntPtr&gt;(SomeWindow_MenuOpened);
	///			_uiaMenus.MenuClosed += new EventHandler&lt;IntPtr&gt;(SomeWindow_MenuClosed);
	///		</code>
	///		
	///		Then it should register the window or control it wants to listen to for menu open/close events:
	///		<code>
	///			_uiaMenus.RegisterListener(hWnd, UIAMenus.MenuEvent.OpenClose);
	///			Console.WriteLine("No. of listeners is {0}.", _uiaMenus.Listeners);
	///		</code>
	///		
	///		Then it should handle events when they occur:
	///		<code>
	///			private void SomeWindow_MenuOpened(object sender, IntPtr hWnd)
	///			{
	///				// ...
	///			}
	///			private void SomeWindow_MenuClosed(object sender, IntPtr hWnd)
	///			{
	///				// ...
	///			}
	///		</code>
	///		
	///		Finally, it should unregister when done:
	///		<code>
	///			private void MyForm_FormClosed(object sender, EventArgs e)
	///			{
	///				_uiaMenus.UnregisterListeners();
	///			}
	///		</code>
	/// </example>
	public class UIANetMenus : UIAMenus
	{
		#region Constructor

		/// <summary>
		/// Default constructor.
		/// </summary>
		public UIANetMenus()
		{
			// nothing
		}

		/// <summary>
		/// Destructor.
		/// </summary>
		~UIANetMenus()
		{
			if (_listeners?.Count > 0)
			{
				Automation.RemoveAllEventHandlers();
				_listeners.Clear();
			}
		}

		#endregion

		#region UIA Menu Listening

		/// <summary>
		/// Register for MenuOpened and MenuClosed events through UI Automation.
		/// </summary>
		/// <param name="hWnd">A handle to the window or control on which to listen for menu events.</param>
		/// <param name="what">What menu events to listen for: opened, closed, or both.</param>
		/// <param name="where">Where to listen for the menu events in the UI tree.</param>
		/// <returns>True if we successfully register for the events; false otherwise.</returns>
		public override bool RegisterListener(IntPtr hWnd, MenuEvent what, UITree where)
		{
			if (what == MenuEvent.None)
				return false;

			bool hasOpen = false;
			bool hasClose = false;
			bool wantOpen = ((what & MenuEvent.Open) != 0);
			bool wantClose = ((what & MenuEvent.Close) != 0);

			foreach (KeyValuePair<IntPtr, MenuEvent> kvp in _listeners)
			{
				if (wantOpen && !hasOpen)
				{
					if (kvp.Key == hWnd && kvp.Value == MenuEvent.Open)
						hasOpen = true;
				}
				if (wantClose && !hasClose)
				{
					if (kvp.Key == hWnd && kvp.Value == MenuEvent.Close)
						hasClose = true;
				}
				if (hasOpen && hasClose)
					break;
			}
			if ((wantOpen && !hasOpen) || (wantClose && !hasClose))
			{
				AutomationElement element = null;
				try
				{
					CacheRequest cache = new CacheRequest();
					cache.TreeScope = TreeScope.Element;
					cache.Add(AutomationElement.NativeWindowHandleProperty);
					cache.Push(); // activate the request
					element = AutomationElement.FromHandle(hWnd); // will cache the native window handle as per above
					if (element != null)
					{
						if (wantOpen && !hasOpen)
						{
							Automation.AddAutomationEventHandler(AutomationElement.MenuOpenedEvent, element, (TreeScope) where, UIAMenuEventHandler);
							_listeners.Add(new KeyValuePair<IntPtr, MenuEvent>(hWnd, MenuEvent.Open));
							Debug.WriteLine(String.Format("Listener++: \"{0}\": <Open:{1}>", WindowEx.GetWindowTitle(hWnd), hWnd));
						}
						if (wantClose && !hasClose)
						{
							Automation.AddAutomationEventHandler(AutomationElement.MenuClosedEvent, element, (TreeScope) where, UIAMenuEventHandler);
							_listeners.Add(new KeyValuePair<IntPtr, MenuEvent>(hWnd, MenuEvent.Close));
							Debug.WriteLine(String.Format("Listener++: \"{0}\": <Close:{1}>", WindowEx.GetWindowTitle(hWnd), hWnd));
						}
					}
					cache.Pop(); // deactivate the request
				}
				catch (Exception ex)
				{
					Console.WriteLine(ex);
				}
				return (element != null);
			}
			return false;
		}

		/// <summary>
		/// Unregister all of the event handlers that have been registered.
		/// </summary>
		/// <seealso cref="https://docs.microsoft.com/en-us/visualstudio/debugger/just-my-code?view=vs-2019"/>
		public override void UnregisterListeners()
		{
			try
			{
				if (_listeners?.Count > 0)
				{
					Automation.RemoveAllEventHandlers();
					_listeners.Clear();
					Debug.WriteLine("Listeners: Cleared.");
				}
			}
			catch (Exception ex)
			{
				Console.WriteLine(ex);
			}
		}

		/// <summary>
		/// Handles when a menu is opened or closed on a registered window.
		/// </summary>
		/// <param name="sender">The element that generated this event.</param>
		/// <param name="eventId">The event identifier for this event.</param>
		private void UIAMenuEventHandler(object sender, AutomationEventArgs e)
		{
			IntPtr hWnd = IntPtr.Zero;
			try
			{
				if (sender != null)
				{
					AutomationElement element = (AutomationElement) sender;
					hWnd = new IntPtr(element.Cached.NativeWindowHandle); // get the cached native window handle
				}
			}
			catch (Exception ex)
			{
				Console.WriteLine(ex);
			}
			// fire the event that occurred
			if (e.EventId == AutomationElement.MenuOpenedEvent)
				OnMenuOpened(hWnd);
			else if (e.EventId == AutomationElement.MenuClosedEvent)
				OnMenuClosed(hWnd);
		}

		#endregion

		#region Mnemonics & Shortcuts

		/// <summary>
		/// Given a window handle, gets the mnemonic key strings currently visible for that window. These
		/// mnemonic strings include, e.g., "Alt + Space" if there's a system menu; "Alt+F" if there's a 
		/// File menu whose "F" is the mnemonic key ("&File"); and individual letters if a menu is already 
		/// showing and has items on it with mnemonics. If callers only want the mnemonics from a showing menu,
		/// they can remove any mnemonic beginning with "Alt".
		/// </summary>
		/// <param name="hWnd">A window handle.</param>
		/// <param name="markDisabled">If true, the mnemonics for disabled menu items will be marked with a 
		/// string prefix. This can be a character, such as an asterisk (*), minus sign (-), or tilde (~), or
		/// a string such as "[disabled]" or "[d]", or whatever the caller wants. To avoid confusion, callers 
		/// should not use a lone ampersand character (&) because of its role in delineating mnemonics to begin
		/// with. Note that no space is added between the mark and the mnemonic; if callers want a space, they
		/// need to include it in the mark string.</param>
		/// <param name="mark">A string mark that will be prepended to the front of all disabled menu items. This
		/// parameter is ignored if <i>markDisabled</i> is false.</param>
		/// <returns>An array with the number of slots equivalent to the number of currently available menu items, 
		/// including MenuBar menus and the window's System Menu. For each slot in the array, if that item has a 
		/// mnemonic, a string will be present. Otherwise, its slot will have String.Empty as its value. Also, the 
		/// return value will be null if the given window does not support UI Automation, or if its UI tree has no 
		/// identifiable menus or menu items.</returns>
		public static string[] GetVisibleMnemonics(IntPtr hWnd, bool markDisabled, string mark)
		{
			return UIAMenuHelper(
				AutomationElement.FromHandle(hWnd),
				AutomationElement.AccessKeyProperty,
				TreeScope.Subtree,
				ControlType.MenuItem,
				markDisabled,
				mark);
		}

		/// <summary>
		/// Given a window handle, gets the shortcut key strings currently visible for that window. These
		/// strings are formatted, e.g., as "Ctrl+X" for Cut, or "Ctrl+C" for Copy.
		/// </summary>
		/// <param name="hWnd">A window handle.</param>
		/// <param name="markDisabled">If true, the shortcuts for disabled menu items will be marked with a 
		/// string prefix. This can be a character, such as an asterisk (*), minus sign (-), or tilde (~), or
		/// a string such as "[disabled]" or "[d]", or whatever the caller wants. Note that no space is added 
		/// between the mark and the shortcut; if callers want a space, they need to include it in the mark string.
		/// </param>
		/// <param name="mark">A string mark that will be prepended to the front of all disabled menu items. This
		/// parameter is ignored if <i>markDisabled</i> is false.</param>
		/// <returns>An array with the number of slots equivalent to the number of currently available menu items, 
		/// including MenuBar menus and the window's System Menu. For each slot in the array, if that item has a 
		/// shortcut, a string will be present. Otherwise, its slot will have String.Empty as its value. Also, the 
		/// return value will be null if the given window does not support UI Automation, or if its UI tree has no 
		/// identifiable menus or menu items.</returns>
		public static string[] GetVisibleShortcuts(IntPtr hWnd, bool markDisabled, string mark)
		{
			return UIAMenuHelper(
				AutomationElement.FromHandle(hWnd),
				AutomationElement.AcceleratorKeyProperty,
				TreeScope.Subtree,
				ControlType.MenuItem,
				markDisabled,
				mark);
		}

		/// <summary>
		/// Gets the mnemonic key strings currently visible on a popup (i.e., context) menu. These mnemonic strings 
		/// include, e.g., "Alt+C" if there's a Copy menu item whose "C" is the mnemonic key ("&Copy").
		/// </summary>
		/// <param name="markDisabled">If true, the mnemonics for disabled menu items will be marked with a 
		/// string prefix. This can be a character, such as an asterisk (*), minus sign (-), or tilde (~), or
		/// a string such as "[disabled]" or "[d]", or whatever the caller wants. To avoid confusion, callers 
		/// should not use a lone ampersand character (&) because of its role in delineating mnemonics to begin
		/// with. Note that no space is added between the mark and the mnemonic; if callers want a space, they
		/// need to include it in the mark string.</param>
		/// <param name="mark">A string mark that will be prepended to the front of all disabled menu items. This
		/// parameter is ignored if <i>markDisabled</i> is false.</param>
		/// <returns>An array with the number of slots equivalent to the number of currently available popup menu 
		/// items. For each slot in the array, if that item has a mnemonic, a string will be present. Otherwise, its 
		/// slot will have String.Empty as its value. Also, the return value will be null if there is no open popup
		/// menu.</returns>
		public static string[] GetPopupMnemonics(bool markDisabled, string mark)
		{
			return UIAMenuHelper(
				AutomationElement.RootElement,
				AutomationElement.AccessKeyProperty,
				TreeScope.Element | TreeScope.Children,
				ControlType.Menu,
				markDisabled,
				mark);
		}

		/// <summary>
		/// Gets the shortcut key strings currently visible on a popup (i.e., context) menu. These strings are 
		/// formatted, e.g., as "Ctrl+X" for Cut, or "Ctrl+C" for Copy.
		/// </summary>
		/// <param name="markDisabled">If true, the shortcuts for disabled menu items will be marked with a 
		/// string prefix. This can be a character, such as an asterisk (*), minus sign (-), or tilde (~), or
		/// a string such as "[disabled]" or "[d]", or whatever the caller wants. Note that no space is added 
		/// between the mark and the shortcut; if callers want a space, they need to include it in the mark string.
		/// </param>
		/// <param name="mark">A string mark that will be prepended to the front of all disabled menu items. This
		/// parameter is ignored if <i>markDisabled</i> is false.</param>
		/// <returns>An array with the number of slots equivalent to the number of currently available menu items. 
		/// For each slot in the array, if that item has a shortcut, a string will be present. Otherwise, its slot 
		/// will have String.Empty as its value. Also, the return value will be null if there is no open popup menu.
		/// </returns>
		public static string[] GetPopupShortcuts(bool markDisabled, string mark)
		{
			return UIAMenuHelper(
				AutomationElement.RootElement,
				AutomationElement.AcceleratorKeyProperty,
				TreeScope.Element | TreeScope.Children,
				ControlType.Menu,
				markDisabled,
				mark);
		}

		/// <summary>
		/// UI Automation helper function.
		/// </summary>
		/// <param name="element">The top element from which to search the UI tree. For mounted menus, use
		/// AutomationElement.FromHandle(hWnd), where 'hWnd' is a window handle. For popup menus, use
		/// AutomationElement.RootElement, which corresponds to the Desktop window, of which all popup windows are 
		/// children.</param>
		/// <param name="keyProperty">For mnemonics, use AutomationElement.AccessKeyProperty. For shortcuts, use  
		/// AutomationElement.AcceleratorKeyProperty.</param>
		/// <param name="searchTree">For mounted menus, use TreeScope.Subtree. For popup menus, use 
		/// TreeScope.Element | TreeScope.Children.</param>
		/// <param name="controlType">For mounted menus, use ControlType.MenuItem, as we'll search directly for 
		/// menu items. For popup menus, use ControlType.Menu, as we'll search for the menu first, and then access 
		/// its children, which are its menu items.</param>
		/// <param name="markDisabled">If true, the mnemonics or shortcuts for disabled menu items will be marked 
		/// with a  string prefix.</param>
		/// <param name="mark">A string mark that will be prepended to the front of all disabled menu items. This
		/// parameter is ignored if <i>markDisabled</i> is false.</param>
		/// <returns>
		/// An array with the number of slots equivalent to the number of currently available menu items.
		/// </returns>
		/// <remarks>This function uses caching to make only one cross-process call.</remarks>
		/// <see cref="https://www.codeproject.com/Articles/33049/WPF-UI-Automation"/>
		private static string[] UIAMenuHelper(
			AutomationElement element, 
			AutomationProperty keyProperty, 
			TreeScope searchTree, 
			ControlType controlType, 
			bool markDisabled, 
			string mark)
		{
			string[] array = null;
			if (element != null)
			{
				// request to cache certain properties for a certain UI tree scope
				CacheRequest cache = new CacheRequest();
				cache.TreeScope = searchTree; // TreeScope.Subtree or TreeScope.Element | TreeScope.Children
				cache.Add(AutomationElement.ControlTypeProperty);
				cache.Add(keyProperty); // AutomationElement.AccessKeyProperty or AutomationElement.AcceleratorKeyProperty
				cache.Add(AutomationElement.IsEnabledProperty);
				cache.Push(); // activate the request

				PropertyCondition condition = new PropertyCondition(AutomationElement.ControlTypeProperty, controlType);
				AutomationElementCollection items = null;
				if (controlType.LocalizedControlType == "menu item") // for mounted menus
				{
					items = element.FindAll(searchTree, condition); // get all menu items
				}
				else if (controlType.LocalizedControlType == "menu") // for popup menus
				{
					AutomationElement menu = element.FindFirst(searchTree, condition); // get open popup menu
					items = menu?.CachedChildren;
				}
				cache.Pop(); // release the request

				if (items?.Count > 0)
				{
					array = new string[items.Count];
					for (int i = 0; i < items.Count; i++)
					{
						object keyStr = items[i].GetCachedPropertyValue(keyProperty, true);
						if (keyStr == AutomationElement.NotSupported) // doesn't have the key property
						{
							array[i] = String.Empty;
						}
						else if (!markDisabled) // we don't care about marking disabled items
						{
							array[i] = (string) keyStr;
						}
						else // we want disabled items marked
						{
							object isEnabledProp = items[i].GetCachedPropertyValue(AutomationElement.IsEnabledProperty, true);
							if (isEnabledProp == AutomationElement.NotSupported) // doesn't have an enabled property
							{
								array[i] = String.Empty;
							}
							else
							{
								bool isEnabled = (bool) isEnabledProp;
								array[i] = isEnabled ? (string) keyStr : String.Format("{0}{1}", mark, keyStr);
							}
						}
					}
				}
			}
			return array;
		}

		#endregion
	}
}
